Package: twoPhaseGAS
Type: Package
Title: Two-Phase Genetic Association Study Design and Analysis with
        Missing Covariates by Design
Version: 1.2.5
Authors@R: c(
    person("Osvaldo", "Espin-Garcia",
           email = "oespinga@uwo.ca",
           role = c("aut", "cre")),
    person("Apostolos", "Dimitromanolakis",
           email = "apostolis@live.ca",
           role = "aut"),
    person("Radu V", "Craiu",
           email = "radu.craiu@utoronto.ca",
           role = "ths"),
    person("Shelley B", "Bull",
           email = "bull@lunenfeld.ca",
           role = "ths"))
Description: Provides functionality for designing and analysing two-phase 
    genetic association studies. Phase 1 data usually come from
    genome-wide association study (GWAS) results and we assume phase 2  data 
    will be part of a targeted genome sequencing or fine-mapping study. 
    At design stage, the package assists in selecting a subset of individuals 
    that will be sequenced for phase 2 via alternative approaches, including a 
    flexible genetic algorithm (GA) for near-optimal designs. Once phase 2 data 
    have been collected, the package implements methods to analyse phase 1 and 
    phase 2 data together using semi-parametric regression models via the 
    expectation-maximization (EM) algorithm. For more details see 
    Espin-Garcia, Craiu and Bull (2018) <doi:10.1002/gepi.22099> and 
    Espin-Garcia, Craiu and Bull (2021) <doi:10.1002/sim.9211>.
License: GPL (>= 2)
URL: https://github.com/egosv/twoPhaseGAS
BugReports: https://github.com/egosv/twoPhaseGAS/issues
Depends: R (>= 3.5.0)
Imports: data.table, dfoptim, enrichwith, kofnGA (>= 1.2), MASS,
        Matrix, nloptr, parallel, stats, utils
Suggests: gplots, knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-19 16:07:48 UTC; osvespinga
Author: Osvaldo Espin-Garcia [aut, cre],
  Apostolos Dimitromanolakis [aut],
  Radu V Craiu [ths],
  Shelley B Bull [ths]
Maintainer: Osvaldo Espin-Garcia <oespinga@uwo.ca>
Repository: CRAN
Date/Publication: 2026-05-27 10:00:02 UTC
Built: R 4.4.3; ; 2026-05-27 13:52:57 UTC; unix

Package: twoPhaseGAS
Type: Package
Title: Two-Phase Genetic Association Study Design and Analysis with
        Missing Covariates by Design
Version: 1.2.5
Authors@R: c(
    person("Osvaldo", "Espin-Garcia",
           email = "oespinga@uwo.ca",
           role = c("aut", "cre")),
    person("Apostolos", "Dimitromanolakis",
           email = "apostolis@live.ca",
           role = "aut"),
    person("Radu V", "Craiu",
           email = "radu.craiu@utoronto.ca",
           role = "ths"),
    person("Shelley B", "Bull",
           email = "bull@lunenfeld.ca",
           role = "ths"))
Description: Provides functionality for designing and analysing two-phase 
    genetic association studies. Phase 1 data usually come from
    genome-wide association study (GWAS) results and we assume phase 2  data 
    will be part of a targeted genome sequencing or fine-mapping study. 
    At design stage, the package assists in selecting a subset of individuals 
    that will be sequenced for phase 2 via alternative approaches, including a 
    flexible genetic algorithm (GA) for near-optimal designs. Once phase 2 data 
    have been collected, the package implements methods to analyse phase 1 and 
    phase 2 data together using semi-parametric regression models via the 
    expectation-maximization (EM) algorithm. For more details see 
    Espin-Garcia, Craiu and Bull (2018) <doi:10.1002/gepi.22099> and 
    Espin-Garcia, Craiu and Bull (2021) <doi:10.1002/sim.9211>.
License: GPL (>= 2)
URL: https://github.com/egosv/twoPhaseGAS
BugReports: https://github.com/egosv/twoPhaseGAS/issues
Depends: R (>= 3.5.0)
Imports: data.table, dfoptim, enrichwith, kofnGA (>= 1.2), MASS,
        Matrix, nloptr, parallel, stats, utils
Suggests: gplots, knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-19 16:07:48 UTC; osvespinga
Author: Osvaldo Espin-Garcia [aut, cre],
  Apostolos Dimitromanolakis [aut],
  Radu V Craiu [ths],
  Shelley B Bull [ths]
Maintainer: Osvaldo Espin-Garcia <oespinga@uwo.ca>
Repository: CRAN
Date/Publication: 2026-05-27 10:00:02 UTC
Built: R 4.5.2; ; 2026-05-27 11:09:57 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(lrtest,twoPhaseSPML)
S3method(print,lrtest.twoPhaseSPML)
S3method(print,twoPhaseSPML)
export(DataGeneration_TPD)
export(lrtest)
export(optimTP.GA)
export(optimTP.LM)
export(twoPhase)
export(twoPhaseDesign)
export(twoPhaseHeuristic)
export(twoPhaseSPML)
export(twoPhaseSPML.control)
import(data.table)
import(stats)
importFrom(MASS,ginv)
importFrom(Matrix,colSums)
importFrom(Matrix,crossprod)
importFrom(Matrix,fac2sparse)
importFrom(Matrix,t)
importFrom(dfoptim,hjkb)
importFrom(enrichwith,enrich)
importFrom(kofnGA,kofnGA)
importFrom(nloptr,cobyla)
importFrom(parallel,clusterEvalQ)
importFrom(stats,pchisq)
importFrom(utils,combn)
importFrom(utils,stack)
importFrom(utils,tail)

## twoPhaseGAS

Two-Phase Genetic Association Study design and analysis with missing
covariates by design

## Description

Provides functionality for selecting and analyzing individuals in
two-phase genetic association studies. Phase 1 data usually come from
GWAS results and we assume phase 2 genetic data will be part of a
targeted genome sequencing/fine-mapping study. The package assists in
selecting a subset of individuals that will be sequenced for phase 2.
Once phase 2 data have been collected, the package implements methods to
analyze phase 1 and 2 data together using semi-parametric regression
models.

## Installation

You can install the current version of `twoPhaseGAS` from
[GitHub](https://github.com) using (`devtools` required):

``` r
devtools::install_github("egosv/twoPhaseGAS", ref="main")
```

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4
)

## ----simulate-----------------------------------------------------------------
library(twoPhaseGAS)

data.table::setDTthreads(1)

set.seed(42)
data <- DataGeneration_TPD(
  Beta0  = 2,    # intercept
  Beta1  = 0.5,  # genetic effect (G on Y)
  Disp = 1,      # error variance for default gaussian family
  N      = 3000, # phase 1 cohort size
  LD.r   = 0.75, # LD (r) between causal SNP G and GWAS SNP Z
  P_g    = 0.20, # MAF of G
  P_z    = 0.30  # MAF of Z
)
head(data)

## ----subsample----------------------------------------------------------------
set.seed(1)
n2   <- 500                              # phase 2 size
R    <- rep(0L, nrow(data))
R[sample(nrow(data), n2)] <- 1L         # 1 = selected for phase 2

data[R == 0, c("G1")]  <- NA # Make all phase 1 data complement G1 values  missing

cat("Phase 1 complement:", sum(R == 0), "individuals\n")
cat("Phase 2:           ", sum(R == 1), "individuals\n")

## ----null-model---------------------------------------------------------------
res_Ho <- twoPhaseSPML(
  formula = Y ~ Z,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

print(res_Ho)

## ----alt-model----------------------------------------------------------------
res_Ha <- twoPhaseSPML(
  formula = Y ~ Z + G1,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

print(res_Ha)

## ----access-------------------------------------------------------------------
# Regression coefficients
res_Ha$theta

# Log-likelihood
res_Ha$ll

# Estimated joint distribution of G1 and Z
head(res_Ha$qGZ)

# Wald statistic and degrees of freedom
cat("Wobs =", res_Ha$Wobs, "  df =", res_Ha$df, "\n")
cat("p-value =", pchisq(res_Ha$Wobs, df = res_Ha$df, lower.tail = FALSE), "\n")

## ----warm-start, eval = FALSE-------------------------------------------------
#  # Use null-model estimates as starting point for the alternative model
#  res_warm <- twoPhaseSPML(
#    formula     = Y ~ Z + G1,
#    miscov      = ~ G1,
#    auxvar      = ~ Z,
#    data       = data,
#    start.values = list(betas = res_Ho$theta,
#                        q     = res_Ho$qGZ)
#  )

## ----session------------------------------------------------------------------
sessionInfo()

#' Internal function to optimize over a range of maf, LD and betas.
#'
#'
#' @param design_formula Formula for the regression model, default is Y ~ G + Z. Z is the GWAS SNP, G is the sequence variant. Y is outcome.
#' @param beta  Vector of betas (length of 3, corresponding to intercept, effect size for G, effect size for Z).
#' @param maf_G Maf for G (numeric, ranging from 0 to 1).
#' @param LD Correlation r betweek G and Z (numeric, ranging from -1 to 1, r value).
#' @param data Data frame with Y and Z variables.
#' @param n2 Phase 2 sample size
#' @param family Distrubution of the outcome (default: gaussian()  ). Familes available for glm can be used here. See help(stats::family) for examples.
#' @param useGeneticAlgorithm   If TRUE, use genetic algorithm in addition to Lagrange multiplier approach (slower). Default: FALSE
#'
#'
#'
#' @return  sth
#'
#'
#' @examples
#'
#' data = twoPhase()
#'
#' @export
twoPhase <- function(
  beta = c( 1,1,1 ),
  maf_G = 0.1,
  LD = 0.3 ,
  data = NA,
  n2 = NA,
  design_formula  =  Y ~  G  +  Z,
  family = gaussian(),
  useGeneticAlgorithm = FALSE
){

  message("Optimizing for a range of maf and LD")

  # maf_G = c(0.2,0.4)
  #LD = c(0.1,0.3,0.5)


  if( !is.list(beta)) beta = list( beta )


  values = expand.grid(nbeta = seq_along(beta), maf_G = maf_G, LD = LD)


  L = nrow(values)


  res = list()
  nres = 1


  for(l in 1:L ) {

      val_beta = beta[[values$nbeta[l]]]
      val_maf_G = values$maf_G[l]
      val_LD = values$LD[l]

      message("  - running beta=", paste(val_beta,sep=",",collapse=","), " maf_G=", val_maf_G, " LD=", val_LD)

      result <- try( twoPhaseDesign( beta = val_beta , maf_G = val_maf_G, LD = val_LD, data = data, n2 = n2,  design_formula = design_formula,
                                   family = family, design = ifelse(useGeneticAlgorithm,"optGA","optLM") ) )


      res[[nres]] = result
      nres = nres+1


      names(res)[l] = sprintf("beta=%s maf_G=%.3f LD=%.3f", paste(val_beta,collapse=","), val_maf_G , val_LD )

  }


  # res = res[ sapply(res,class) != "try-error" ] ## better do this filtering outside the function to identify which combinations broke

  return( res )
}




#' Compute sample allocations for a two phase study design.
#'
#'
#' @param design_formula Formula for the regression model, default is Y ~ G + Z, where Z is the GWAS SNP, G is the sequence variant. Y is outcome. Rename the variables in your data.frame to match Y and Z, G is the seq-SNP not present in the data.frame.
#' @param beta  Vector of betas (length of 3, corresponding to intercept, effect size for G, effect size for Z).
#' @param maf_G minor allele frequency for G (numeric, ranging from 0 to 1). Numeric or vector of possible values.
#' @param LD Correlation r betweek G and Z (numeric, ranging from -1 to 1, r value). Numeric or vector of possible values.
#' @param data Data frame with Y and Z variables.
#' @param n2 Phase 2 sample size.
#' @param family Distrubution of the outcome. Default: gaussian(). Familes available for glm can be used here. See help(stats::family) for examples.
#' @param S stratification of the outcome Y. Optional. Should be a numeric vector with strata categories (e.g. 1 1 2 2 3 3 ). If present, its length must be equal to the number of rows in data. Default: NULL. Needed when Y does not render itself into strata, e.g. Gaussian, Poisson, Gamma.
#' @param perc_Y vector of percentiles in increasing order for which the outcome Y will be stratified. Default: c(1/5,4/5). Only used when S is NA. Note that setting up S is strongly suggested.
#' @param p_gz data frame with the joint distribution between G and Z. See examples for the right format. Default: NULL. If present, values of maf_G and LD are disregarded in the analysis.
#' @param design string for the design to use for phase 2 sample selection. One of residual-dependent sampling ("RDS"), optimal as defined by Tao, Zheng and Lin (2019) ("TZL"), optimal via Lagrange multipliers ("LM"), optimal via genetic algorithm ("GA"), probability proportional to size ("PPS"), balanced ("BAL") or combined ("COM") allocations. Default: "RDS". See details for a more explanations.
#' @param ndraws integer that determines the number of draws to examine when design is one of "pps", "bal" or "comb" and the design parameter combinations is greater than 1. Default: 10
#' @param optimCriterion string denoting the optimality criterion used during the optimization. One of "Par-spec" (default), "A-opt" or "D-opt". For parameter-specific, A-optimality or D-optimality, respectively.
#' @param overallMethod string denoting the method to select the overall design when multiple design parameters are given. One of "med-max" (default) or "cumm" for median-maximum and cummulative frequencies, respectively.
#'
#' Note that in this version strata are always defined in terms of Z and S, i.e. a joint design, future implementations may relax this by allowing for only S or only Z (marginal designs, outcome- or covariate-dependent, respectively)
#' @return  sth
#'
#'
#' @export
twoPhaseDesign <- function(

      beta,
      maf_G,
      LD,
      data,
      n2,

      design_formula  =  Y ~  G  +  Z,
      family = gaussian(),
      S,
      perc_Y =  c(1/5,4/5),
      p_gz,

      #useGeneticAlgorithm = FALSE,
      design = c("RDS","LM","GA","PPS","BAL","COM","TZL"),
      ndraws = 10,
      optimCriterion = c("Par-spec","A-opt","D-opt"),
      overallMethod = c("med-max","cumm")

)
{

  #### Perform some checkings

  if( missing(beta) | missing(maf_G) | missing(LD) | missing(data) | missing(n2)) stop("none of beta, maf_G, LD, data or n2 can be missing.")

  if( !is.data.frame(data) ) stop("data should be a data.frame")

  terms1 = all.vars(design_formula)
  terms2 = setdiff(terms1,c("G","Z","Y") )

  if( ! ( "G" %in% terms1 ) ) { stop("G should be included in the design_formula") }
  if( ! ( "Z" %in% terms1 ) ) { stop("Z should be included in the design_formula") }
  if( ! ( "Y" %in% terms1 ) ) { stop("Y should be included in the design_formula") }


  if( sum(terms2 %in% colnames(data) ) != length(terms2)  ) {

    s = which( ! (terms2 %in% colnames(data) ) ) [1]

    stop(sprintf("The variable %s is missing from the data.frame data", terms2[s]) )

  }


  data_tmp  = data[1:2,]; if(  !( "G" %in% colnames(data_tmp) ) )  data_tmp$G = 0.1;
  beta_names = colnames( model.matrix(design_formula,data_tmp) )

  if ( is.list(beta) ) {

    tmp = sapply(beta,length)
    beta_test = beta[[1]]

    if( min(tmp) != max(tmp) ) stop("Length of betas is not equal for all elements of list\n");

  } else  beta_test = beta

  if( length(beta_names) != length(beta_test) ) {

    stop( sprintf("Number of terms in design formula and length of beta do not match: %d terms in beta, %d terms in formula\n",
                  length(beta_test), length(beta_names) ) )

  }


  N = nrow(data) ## phase 1 sample size

  Z = data$Z

  maf_Z = mean(Z, na.rm=TRUE)/2

  Y = data$Y

  if ( missing(S) ){

    if( !is.numeric(perc_Y) || !all(perc_Y<1,perc_Y>0) || is.unsorted(perc_Y) ) stop("perc_Y must all be numeric values between 0 and 1 and in increasing order." )
    S = as.numeric( cut(Y, breaks=quantile(Y, probs=c(0,perc_Y,1), na.rm=TRUE),
            include.lowest=TRUE) )

  } else {
    S_uniq = unique(S)
    S_uniq_sort = S_uniq[order(S_uniq)]
    if( !is.numeric(S_uniq_sort) | any(diff(S_uniq_sort)!=1) | S_uniq_sort[1]!=1 ) stop("S must contain only consecutive integers starting from 1")
  }

  data$S = S

  design <- match.arg(design)

  # These parameters are used in the functions
  strataformula <- ~ Z+S ## formula that determines the stratification factors, i.e. how the strata is defined (other formulas may be used, for example strataformula = ~S or strataformmula= ~Z). These must be discrete numeric variables in the phase 1 data

  ## optimality criterion (so far only 3 implemented)
  optMethod <- match.arg(optimCriterion)
  overallMethod <- match.arg(overallMethod)

  if( optMethod=="Par-spec" ){
    Kind <- which(beta_names=="G")
  } else Kind <- NULL

  if( !missing(p_gz) ){
    message("Since 'p_gz' is specified, values of maf_G and LD are disregarded from the analysis.")
    maf_G <- NULL
    LD <- NULL
  }

  # Optimize over a series of parameters

  if( length(maf_G) > 1 || length(LD) > 1 || is.list( beta ) ) {

    if( design %in% c("optLM","optGA") ){
      result = twoPhase( beta = beta ,
                                 maf_G = maf_G,
                                 LD = LD,
                                 data = data,
                                 n2 = n2,
                                 design_formula = design_formula,
                                 family = family,
                                 useGeneticAlgorithm = design=="optGA" )

      indx_noerror <- sapply(result,class) != "try-error"
      result = result[ indx_noerror ]
      nams = names(result)
      L = sum(indx_noerror)
      result_tmp = matrix( NA, nrow = N, ncol = L )
      for( l in 1:L ) {
        result_tmp[,l] = if( design=="optGA" ){
                      result[[l]]$R.opt.GA
                    }else result[[l]]$R.opt.LM
      }
      result = result_tmp
      colnames(result) = nams
      rm(result_tmp)

      if( !is.list(beta) ) beta = list( beta )

      values = expand.grid(nbeta = seq_along(beta), maf_G = maf_G, LD = LD)

    } else{

      result = twoPhaseHeuristic(
        design = design,
        ndraws = ndraws,
        data = data,
        n2 = n2,
        family = family,
        S = data$S
      )

      L = ndraws


      if( !is.list(beta) ) beta = list( beta )

      values = expand.grid(nbeta = seq_along(beta), maf_G = maf_G, LD = LD)

      indx_noerror = sapply(seq_len(nrow(values)), function(i){
        maf_G = values[i,"maf_G"]
        LD = values[i,"LD"]
        res = try(p_gz_func(maf_Z, maf_G, LD, "G", "Z"))
        if( inherits(res, "try-error")) cat("  - skipping this MAFs-LD combination: maf_Z=", maf_Z, " maf_G=", maf_G, " LD=", LD,"\n", sep="")

        !inherits(res, "try-error")
      })
    }

    values = values[ indx_noerror, ,drop=FALSE]
    L2 = nrow(values)

    # To select a single sample across all design values, we propose two methods (but there may be more)

    ### method 1 calculate the fitness and perform med-max
    if ( overallMethod == "med-max" ){

      FitnessVals <- matrix( NA, nrow = L, ncol = L2 )

      for( k in 1:L2 ){ # k <- 2
        val_beta_b = beta[[values$nbeta[k]]]
        val_maf_G_b = values$maf_G[k]
        val_LD_b = values$LD[k]
        p_gz0_b = p_gz_func(maf_Z, val_maf_G_b, val_LD_b, "G", "Z")

        obsIM.Rk <- obsIM.R(formula = design_formula,
                            miscov = ~G,
                            auxvar = ~Z,
                            family= family,
                            data = data,
                            beta = val_beta_b,
                            p_gz = p_gz0_b,
                            disp=NULL)

        for( l in 1:L ) {
          Rl = result[,l]
          FitnessVals[l,k] <- fitnessTP(obsIM.R = obsIM.Rk,
                                        Rj = Rl,
                                        optimMeasure = optMethod, K.idx = Kind)
        }
      }

      x <- apply(FitnessVals,1,median)
      x1 <- which(x==min(x))
      if( length(x1)>1 ) x1 <- x1[1]
      indx <- x1

      R.overall <- result[,indx]

    ### method 2, group the indicators to inform a new sampling
    } else if ( overallMethod == "cumm" ){

      x <- NA
      Freqs <- rowSums(result)
      indx <- sample(1:N,n2,prob=Freqs)
      R.overall <- rep(0, N)
      R.overall[indx] <- 1

    } else stop("Method for overall design not identified. Select one of med-max or cumm.")

    ### TODO: implement the sequential approach discussed with Radu on July 8, 2019

    data$R = R.overall
    R.name = switch(design,
                    optLM="R.opt.LM",
                    optGA="R.opt.GA",
                    pps="R.pps",
                    bal="R.bal",
                    comb="R.comb")
    names(data)[which(names(data)=="R")] = R.name

    return(list(overall_design=data,indiv_design=result,medianFitness=x))
  }


  # example data
  if(0) {

    maf_G = 0.1
    LD = 0.3
    data = data.frame(  Y = rnorm(1000),  Z = rbinom(1000, 2, 0.2) )
    beta = c( 1,1,1 )
    n2 = 200
    family = gaussian
    design_formula =  Y ~  G  +  Z
  }

  #** Single draw from heuristic design ----
  if( design %in% c("pps","bal","comb")){

    data.ret = twoPhaseHeuristic(design = design,
                                 ndraws = 1,
                                 data = data,
                                 n2 = n2,
                                 family = family,
                                 S = data$S)

    data.ret = cbind(data,R=data.ret)
    R.name = switch(design,pps="R.pps",bal="R.bal",comb="R.comb")
    names(data.ret)[which(names(data.ret)=="R")] = R.name
    return ( data.ret )
  }
  # TODO: Check data for validity and data types.


  ## determine "design" quantities to select the phase 2 data

  p_Z <- data.frame( xtabs(~Z,data)/nrow(data) )  ## calculates genotype distribution of Z from data

  LD.r <- LD ## specify LD between GWAS SNP, Z, and potentially "causal" locus

  ### if the joint distribution between g and z is provided, do not calculate it
  if( !missing(p_gz) ){

    p_gz0 = p_gz
  ### else calculate p_gz, i.e. the joint distribution of G and Z using maf_G, LD.r and available data from Z using function p_gz_func() in optimJTC_v1.R. This procedure sometimes fails, if that's the case we may cheat a bit and use the complete data.
  } else p_gz0 <-  p_gz_func(p_Z, maf_G, LD.r, "G", "Z")


  #** Phase 2 sample selection via optimal designs ----
  ## R1-R4 essentially contain a vector of size N with zeros (unselected) and ones (selected)

  ### check number of strata
  strat_df = data.frame(xtabs(strataformula,data=data))
  n_strat = nrow(strat_df)
  S_unique = unique(data$S)

  sel.prob.bal <- rep(1/n_strat,n_strat) ## selection probabilities for a balanced design

  ### To generalize the concept of a combined design for more than 3 groups for Y, we'll always set to selection probability for the median value(s) to 0

  ### Find the strata that will have zero probability of being selected (the median)
  if(  (length(S_unique)%%2) == 1  ){ ### calculate the median depending on whether the number of strata for S is even or odd
    med_stra <- median(unique(data$S))
  }else {
    med_stra <- as.numeric(quantile(S_unique,0.5,type=1))
    med_stra <- c(med_stra,med_stra+1)
  }
  indx_com <- which(!strat_df$S%in%med_stra)
  n_strat_com <- length(indx_com)


  sel.prob.com <- rep(0,n_strat) ## selection probabilities for a combined design
  sel.prob.com[indx_com] <- 1/n_strat_com


  # Basic stratified sampling for the "intuitive allocations"
  # R1 <- BSS(samp.fracs=sel.prob.bal,n2,data,strataformula)$R ## indicators for balanced design
  # R2 <- BSS(samp.fracs=sel.prob.com,n2,data,strataformula)$R ## indicators for combined design




  if(1)

    opt.prop <- optimTP.LM(  formula = design_formula,
                               miscov = ~ G,
                               auxvar = ~ Z,

                               strata = strataformula,

                               family = family,

                               n = n2,

                               data = data,

                               beta = beta,

                               p_gz = p_gz0,


                               disp = NULL,

                               optimMeasure  =  optMethod,

                               K.idx = Kind,
                               min.nk = NULL
                            )




  R3 <- BSS(samp.fracs = opt.prop$prR_cond_optim, n2,  data, strataformula) ## indicators for LM design



  ### TODO::::::::::::::  Add fitness of the LM results for testing



  if( design  == "optGA" )  {

  ### TODO:::::::::::::: Add more control to the GA parameters, e.g. whether initial sample is provided or not. If provided we should give means to let the user initialize thm

    ## For the genetic algorithm (GA), instead of providing a random initialization, I give an informed initialization based on combined, balanced and LM designs (a third each).


    pop1 <- sapply(1:20,function(x)  {


      sa <- which(BSS(samp.fracs=opt.prop$prR_cond_optim, n2, data, strataformula)==1)
      len <- length(sa)
      if( len==n2 ){ return(sa)
      }else if( len>n2 ){ return( sa[-sample(len, len-n2)] )
      }else return( c(sa,sample((1:N)[-sa],n2-len)) )


    })



    ## combined


    pop2 <- sapply(1:20,function(x)  {

      sa <- which(BSS(samp.fracs=sel.prob.com,n2,data,strataformula)$R==1)
      len <- length(sa)
      if( len==n2 ){ return(sa)
      }else if( len>n2 ){ return( sa[-sample(len, len-n2)] )
      }else return( c(sa,sample((1:N)[-sa],n2-len)) )

    })



    ## balanced


    pop3 <- sapply(1:20,function(x)  {

      sa <- which(BSS(samp.fracs=sel.prob.bal,n2,data,strataformula)$R==1)
      len <- length(sa)
      if( len==n2 ){ return(sa)
      }else if( len>n2 ){ return( sa[-sample(len, len-n2)] )
      }else return( c(sa,sample((1:N)[-sa],n2-len)) )


    } )


    ## Genetic Algorithm approach. This function may take a bit to run (in the order of a few minutes).
    GA.sol <- optimTP.GA(
                            formula=design_formula,
                            miscov=~G,
                            auxvar=~Z,
                            family=family,
                            n=n2,
                            data,
                            beta=beta,
                            p_gz=p_gz0,
                            disp=NULL,

                ### TODO:::: ideally these parameters could be specified by the user
                            ga.popsize = 60,
                            ga.propelit = 0.8,
                            ga.proptourney = 0.8,
                            ga.ngen = 300,
                            ga.mutrate = 0.001,
                            ga.initpop = t(cbind(pop1,pop2,pop3)),

                            optimMeasure = optMethod,
                            K.idx = Kind,
                            seed = 1

              )


    R4 <- rep(0, N ); R4[GA.sol$bestsol] <- 1  ## indicators for GA design

    data$R.opt.GA = R4

  }



  if( design  == "optLM" ) {

     data$R.opt.LM = R3
    # print(opt.prop)
  }


  return ( data )


}

#' Select samples for phase 2 under heuristic designs.
#'
#' @param design Heuristic design to use for phase 2 sample selection. One of probability proportional to size ("pps"), balanced ("bal") or combined ("comb") allocations. Default: "pps".
#' @param ndraws Number of draws of the heuristic design to generate Default: 1.
#' @param data Data frame with Y and Z variables.
#' @param n2 Phase 2 sample size.
#' @param family Distrubution of the outcome. Default: gaussian(). Familes available for glm can be used here. See help(stats::family) for examples.
#' @param S stratification of the outcome Y. Optional. Should be a numeric vector with strata categories (e.g. 1 1 2 2 3 3 ). If present, its length must be equal to the number of rows in data. Default: NULL. Needed when Y does not render itself into strata, e.g. Gaussian, Poisson, Gamma.
#' @param perc_Y vector of percentiles in increasing order for which the outcome Y will be stratified. Default: c(1/5,4/5). Only used when S is NA. Note that setting up S is strongly suggested.
#'
#' Note that in this version strata are always defined in terms of Z and S, i.e. a joint design, future implementations may relax this by allowing for only S or only Z (marginal designs, outcome- or covariate-dependent, respectively)
#' @return  sth
#'
#'
#' @export
twoPhaseHeuristic <- function(
  design = c("pps","bal","comb"),
  ndraws = 1,
  data = NA,
  n2 = NA,

  family = gaussian(),
  S = NULL,
  perc_Y =  c(1/5,4/5)
)
{

  #### Perform some checkings

  if( !is.data.frame(data) ) stop("data should be a data.frame")

  terms1 = all.vars(as.formula("Y~Z"))
  terms2 = setdiff(terms1,c("Z","Y") )

  if( ! ( "Z" %in% terms1 ) ) { stop("Z should be included in the design_formula") }
  if( ! ( "Y" %in% terms1 ) ) { stop("Y should be included in the design_formula") }


  if( sum(terms2 %in% colnames(data) ) != length(terms2)  ) {

    s = which( ! (terms2 %in% colnames(data) ) ) [1]

    stop(sprintf("The variable %s is missing from the data.frame data", terms2[s]) )

  }

  N = nrow(data) ## phase 1 sample size

  Y = data$Y

  if ( is.null(S) ){

    if( !all(perc_Y<1,perc_Y>0) || is.unsorted(perc_Y) ) stop("perc_Y must all be values between 0 and 1 and in increasing order." )
    S = as.numeric( cut(Y, breaks=quantile(Y, probs=c(0,perc_Y,1), na.rm=TRUE),
                        include.lowest=TRUE) )

  } else {
    S_uniq = unique(S)
    S_uniq_sort = S_uniq[order(S_uniq)]
    if( !is.numeric(S_uniq_sort) | any(diff(S_uniq_sort)!=1) | S_uniq_sort[1]!=1 ) stop("S must contain only consecutive integers starting from 1")
  }

  data$S = S

  design <- match.arg(design)

  ### TODO: check that S and Z are discrete


  # These parameters are used in the functions
  strataformula <- ~ Z+S ## formula that determines the stratification factors, i.e. how the strata is defined (other formulas may be used, for example strataformula = ~S or strataformmula= ~Z). These must be discrete numeric variables in the phase 1 data

  ### check number of strata
  strat_df = data.frame(xtabs(strataformula,data=data))
  n_strat = nrow(strat_df)
  S_unique = unique(data$S)

  ### draw the phase 2 sample indicators accroding to the selected design
  if( design == "pps" ){
    sel.prob.pps = rep(n2/N,n_strat) ## selection probabilities for a design with selection probabilities proportional to size

    Rs = sapply(1:ndraws,function(i)BSS(samp.fracs=sel.prob.pps,n2,data,strataformula) )

  } else if ( design == "bal" ){
    sel.prob.bal <- rep(1/n_strat,n_strat) ## selection probabilities for a balanced design

    Rs = sapply(1:ndraws, function(x)BSS(samp.fracs=sel.prob.bal,n2,data,strataformula)$R)

  } else if (design == "comb" ){

    ### To generalize the concept of a combined design for more than 3 groups for Y, we'll always set to selection probability for the median value(s) to 0

    ### Find the strata that will have zero probability of being selected (the median)
    if(  (length(S_unique)%%2) == 1  ){ ### calculate the median depending on whether the number of strata for S is even or odd
      med_stra <- median(unique(data$S))
    }else {
      med_stra <- as.numeric(quantile(S_unique,0.5,type=1))
      med_stra <- c(med_stra,med_stra+1)
    }
    indx_com <- which(!strat_df$S%in%med_stra)
    n_strat_com <- length(indx_com)

    sel.prob.com <- rep(0,n_strat) ## selection probabilities for a combined design
    sel.prob.com[indx_com] <- 1/n_strat_com

    Rs = sapply(1:ndraws, function(x)BSS(samp.fracs=sel.prob.com,n2,data,strataformula)$R)
  }

  ### Rs above should correspond to a matrix of N rows and ndraws columns with zeros and ones
  return ( Rs[,,drop=FALSE] )
}

### These new functions do NOT consider the phi (dispersion parameter) as a quantity of inference
fnc_wgts <- function(formula,family,dat0,theta,q,disp,G_,Z_,uniqZ){
  N <- wg0 <- id_ <- NULL # suggested by https://www.r-bloggers.com/2019/08/no-visible-binding-for-global-variable/
  resp<-as.character(formula[[2]]); terms_q<-c(G_,Z_)
  pG_<-aggregate(as.formula(paste0("q~",paste(G_,collapse="+"))),q,FUN=sum); names(pG_)[ncol(pG_)]<-"q_g"
  dat0$ord <- seq_len(nrow(dat0))
  dat0bis<-merge(as.data.table(dat0),as.data.table(q[,c(terms_q,"q")]),by=terms_q,all.x=TRUE,sort = FALSE)
  stopifnot("First merge failed, q may contain duplicated keys." = nrow(dat0bis) == nrow(dat0))
  dat0bis<-merge(dat0bis,as.data.table(pG_),by=G_,all.x=TRUE,sort = FALSE)
  stopifnot("Second merge failed, pG_ may contain duplicated keys." =nrow(dat0bis) == nrow(dat0))

  dat0bis$ind<-ifelse(dat0bis[[Z_]]%in%uniqZ,1,0)
  dat0bis$q<-dat0bis$ind*dat0bis$q + (1-dat0bis$ind)*dat0bis$q_g

  X0 = model.matrix(formula,dat0bis)
  eta = as.vector(if (NCOL(X0) == 1L){ X0 * theta }else X0 %*% theta)
  mu <- family$linkinv(eta)

  if(family$family=="gaussian"){
    dat0bis$wg0 <- dnorm(x=dat0bis[[resp]],mean=mu,sd=sqrt(disp))*dat0bis[["q"]]
  }else if(family$family=="binomial"){
    dat0bis$wg0 <- dbinom(x=dat0bis[[resp]],size=1,prob=mu)*dat0bis[["q"]]
  }else if(family$family=="poisson"){
    dat0bis$wg0 <- dpois(x=dat0bis[[resp]],lambda=mu)*dat0bis[["q"]]
  }else if(family$family=="Gamma"){
    dat0bis$wg0 <- dgamma(x=dat0bis[[resp]],shape=1/disp,scale=mu*disp)*dat0bis[["q"]]
  }else stop("family not one of: gaussian, binomial, Gamma or poisson.")

  dat0bis[, N:=sum(wg0), by = id_]

  # dat0bis1 <- dat0bis

  # if(family$family=="gaussian"){
  #   dat0bis$lwg0 <- dnorm(x=dat0bis[[resp]],mean=mu,sd=sqrt(disp),log=TRUE) + log(dat0bis[["q"]])
  # }else if(family$family=="binomial"){
  #   dat0bis$lwg0 <- dbinom(x=dat0bis[[resp]],size=1,prob=mu,log=TRUE) + log(dat0bis[["q"]])
  # }else if(family$family=="poisson"){
  #   dat0bis$lwg0 <- dpois(x=dat0bis[[resp]],lambda=mu,log=TRUE) + log(dat0bis[["q"]])
  # }else if(family$family=="Gamma"){
  #   dat0bis$lwg0<-dgamma(x=dat0bis[[resp]],shape=1/disp,scale=mu,log=TRUE) + log(dat0bis[["q"]])
  # }else stop("family not one of: gaussian, binomial, Gamma or poisson.")
  #
  # dat0bis$wg0 <- exp(dat0bis$lwg0)
  #
  # dat0bis[, N:=sum(wg0), by = id_] ## fast way to add a column (N) with the sum of the weigths by id_
  #
  # #    ## Procedure to normalize probabilities under small values taken from http://stats.stackexchange.com/questions/66616/converting-normalizing-very-small-likelihood-values-to-probability
  # if( any(dat0bis$N==0) ){
  #   zerindx <- which(dat0bis$N==0)
  #   dfwg0 <- dat0bis[zerindx,c("id_","ord","lwg0"), with=FALSE]
  #   dfwg0[, max:=max(lwg0),by=id_]
  #   if( !all(dfwg0$ord==dat0bis$ord[zerindx]) ) stop("Something went wrong in the probability normalization process.")
  #   dat0bis$wg0[zerindx] <- exp(dfwg0$lwg0-dfwg0$max)
  #   dat0bis[zerindx, N:=sum(wg0), by = id_]
  # }
  #
  # dat0bis1$wg <- dat0bis1$wg0/dat0bis1$N
  # dat0bis1 <- dat0bis1[order(dat0bis1$ord),]

  dat0bis$wg <- dat0bis$wg0/dat0bis$N
  dat0bis <- dat0bis[order(dat0bis$ord),]

  return(dat0bis$wg)
  #return(cbind(dat0bis$wg,dat0bis1$wg,dat0bis$N,dat0bis1$N,dat0bis$wg0,dat0bis1$wg0))
}

.loglik<-function(theta,q,disp,formula,Y_,G_,Z_,dat,family){
  datbis=merge(dat,q,by=c(G_,Z_))
  In<-t(fac2sparse(datbis$id_))#class.ind2.Mat(datbis$id_)
  mu<-family$linkinv(model.matrix(formula,datbis)%*%theta)

  if(family$family=="gaussian"){
    sig<-sqrt(disp)
    ll <- sum( log( colSums( dnorm(datbis[,Y_], mu, sig)*datbis$q*In ) ) )
  }else if(family$family=="poisson"){
    ll <- sum( log( colSums( dpois(datbis[,Y_], lambda=mu)*datbis$q*In ) ) )
  }else if(family$family=="Gamma"){
    ll <- sum( log( colSums( dgamma(datbis[,Y_],shape=1/disp, scale=mu*disp)*datbis$q*In ) ) )
  }else if(family$family=="binomial"){
    ll <- sum( log( colSums( dbinom(datbis[,Y_],size=1,prob=mu)*datbis$q*In ) ) )
  }else stop("family not one of: gaussian, binomial, Gamma, inverse.gaussian or poisson.")

  return(ll)
}

score_glm<-function(theta,disp,q,formula,Y_,data,family){
  q_j<-q[,"q"]; nq<-nrow(q)
  y <- data[,Y_]
  x <- as.matrix(model.matrix(formula,data))

  nobs <- NROW(y); nvars <- NCOL(x)

  variance <- family$variance
  linkinv <- family$linkinv
  mu.eta <- family$mu.eta

  eta <- as.vector(if (NCOL(x) == 1L){ x * theta }else x %*% theta)
  mu <- linkinv(eta)
  varmu <- variance(mu)
  mu.eta.val <- mu.eta(eta)
  if (any(is.na(varmu))) #anyNA(varmu)
    stop("NAs in V(mu)")
  if (any(varmu == 0))
    stop("0s in V(mu)")
  if (any(is.na(mu.eta.val)))
    stop("NAs in d(mu)/d(eta)")

  w <- (mu.eta.val^2)/varmu
  S <- x[, , drop = FALSE] * w * (y - mu)/(disp*mu.eta.val)

  #return(list(S=S,X=x[, , drop = FALSE],Wx2=w^2*varmu/(disp*mu.eta.val^2)))
  return(list(S=S,X=x[, , drop = FALSE],varmu=varmu,Y=y,mu=mu,w=w,mu.eta.val=mu.eta.val))
}

FIM.glm <- function(theta,qG,disp,formula,Y_,dat,family,glmfit){
  nq<-nrow(qG); nbetas<-length(theta)
  phi<-c(theta,log(qG[,"q"])) #send the log(q_j) here to deal with the numerical issue on the hessian computation
  n_phi<-length(phi)

  Qd1=sapply(dat$k_,function(k){1/qG[qG[,"k_"]==k,"q"]})
  Qd2=sapply(dat$k_,function(k){-1/qG[qG[,"k_"]==k,"q"]^2})
  Qmat = as.matrix(t(fac2sparse(dat$k_)))#class.ind2(dat$k_)
  Qmat.d1=Qmat*Qd1
  Psi=dat$wg_

  Sc = score_glm(theta,disp,qG,formula,Y_,dat,family)
  U <- colSums(Psi*cbind(Sc$S,Qmat.d1)) ##cbind(ptheta,pq_j)
  X = Sc$X
  
  ## second derivatives (i.e. hessian of log P(Y_i|g_j) + log q_j or jacobian of the gradient)
  # I1_0<-crossprod(qr.R(glmfit$qr))/disp # upper left matrix with the fisher information matrix for betas
  mu_fit    <- family$linkinv(as.vector(X %*% theta))
  W_fisher  <- Psi * (family$mu.eta(family$linkfun(mu_fit))^2 /
                        family$variance(mu_fit))
  I1_0      <- crossprod(X, W_fisher * X) / disp
  
  I2_0 <-diag(colSums(Psi * Qmat * Qd2)) #second derivatives for an i,jth observation

  I1<-rbind(cbind(I1_0, matrix(0, nrow=nrow(I1_0), ncol=ncol(I2_0))),
            cbind(matrix(0, nrow=nrow(I2_0), ncol=ncol(I1_0)), -I2_0))

  ## first derivatives (i.e. gradient of log P(Y_i|g_j,z_j) + log q_j)

  ## calculate second matrix, i.e. expected value of cross score functions

  Wx2<- (Sc$w * (Sc$Y - Sc$mu)/(disp*Sc$mu.eta.val))^2
  Mx2<- crossprod( X,  Psi*Wx2*X )

  Wxq2<- Sc$w * (Sc$Y - Sc$mu)/(disp*Sc$mu.eta.val)
  Mxq2<-crossprod( X,  Psi*Wxq2*Qmat.d1)

  Mq2<- crossprod( Qmat.d1,  Psi*Qmat.d1 )

  obsI2<-rbind(cbind(Mx2, Mxq2),
               cbind(t(Mxq2),Mq2))

  ## Third matrix, i.e. outer product of expected score. Note this uses the W's from the second matrix and weight them accordingly
  IIt <- crossprod(fac2sparse(dat$id_))#tcrossprod(class.ind2.Mat(dat$id_)) #IIt <- tcrossprod(class.ind2.Mat(dat$id_)) #IIt<-tcrossprod(class.ind2(dat$id_)) ## much slower, specially for larga matrices

  #Mx3<- crossprod( sign(Wx2)*sqrt(abs(Wx2))*X*Psi, IIt %*% (sqrt(abs(Wx2))*X*Psi) )
  #Mxq3<-crossprod( sign(Wxq2)*sqrt(abs(Wxq2))*X*Psi, IIt %*% (sqrt(abs(Wxq2))*Qmat.d1*Psi) )
  
  Mx3 <- as.matrix(crossprod( Wxq2*X*Psi, IIt %*% (Wxq2*X*Psi) ))
  Mxq3<- as.matrix(crossprod( Wxq2*X*Psi, IIt %*% (Qmat.d1*Psi) ))
  Mq3<- as.matrix(crossprod( Qmat.d1*Psi, IIt %*% (Qmat.d1*Psi) ))

  obsI3<-as.matrix(rbind(cbind(Mx3, Mxq3),
                         cbind(t(Mxq3),Mq3)))

  obsFIM=I1-obsI2+obsI3
  return(list(FIM=obsFIM,I1=I1,I2=obsI2,I3=obsI3,U=U))
}

expFIM.glm<-function(theta,qG,disp,formula,Y_,dat,family,glmfit){ ## for this case Psi is the vector of weights (not the matrix)
  nq<-nrow(qG); nbetas<-length(theta)
  phi<-c(theta,log(qG[,"q"]))
  n_phi<-length(phi) # n<-nrow(Psi);

  Qd1=sapply(dat$k_,function(k){1/qG[qG[,"k_"]==k,"q"]})
  Qd2=sapply(dat$k_,function(k){-1/qG[qG[,"k_"]==k,"q"]^2})
  Qmat=as.matrix(t(fac2sparse(dat$k_)))#class.ind2(dat$k_)
  Qmat.d1=Qmat*Qd1
  Qmat.d2=Qmat*Qd2
  Psi=dat$wg_

  Sc <- score_glm(theta,disp,qG,formula,Y_,dat,family)
  X = Sc$X
  
  ## second derivatives (i.e. hessian of log P(Y_i|g_j) + log q_j or jacobian of the gradient)
  # I1_0<-crossprod(qr.R(glmfit$qr))/disp # upper left matrix with the (expected) Fisher information matrix for betas
  mu_fit    <- family$linkinv(as.vector(X %*% theta))
  W_fisher  <- Psi * (family$mu.eta(family$linkfun(mu_fit))^2 /
                        family$variance(mu_fit))
  I1_0      <- crossprod(X, W_fisher * X) / disp
  
  I2_0 <- colSums(Psi * Qmat.d2)
  I2_0 <- diag(I2_0,NROW(I2_0))
  #I2_0 <-diag(colSums(Psi * Qmat.d2)) ##old implementation (note that this brings a problem when the length  inside diag is equal to 1, i.e. one estimate for qG )

  expI1<-rbind(cbind(I1_0, matrix(0, nrow=nrow(I1_0), ncol=ncol(I2_0))),
               cbind(matrix(0, nrow=nrow(I2_0), ncol=ncol(I1_0)), -I2_0))

  ## first derivatives (i.e. gradient of log P(Y_i|g_j,z_j) + log q_j)
 
  lst_indx <- nq+nbetas+1
  I2_0_ <- cbind(Sc$S, Qmat.d1, Psi)
  U <- colSums(I2_0_[,lst_indx]*I2_0_[,-lst_indx])

  ## calculate second matrix, i.e. expected value of cross score functions
  Wx2<- Sc$w^2*Sc$varmu/(disp*Sc$mu.eta.val^2)
  Mx2<- crossprod( X,  Psi*Wx2*X )

  Wxq2<- 0
  Mxq2<-crossprod( X,  Psi*Wxq2*Qmat.d1)

  Mq2<- crossprod( Qmat.d1,  Psi*Qmat.d1 )

  expI2<-rbind(cbind(Mx2, Mxq2),
               cbind(t(Mxq2),Mq2))

  ## Third matrix, i.e. outer product of expected score. Note this uses the W's from the second matrix and weight them accordingly
  IIt <- crossprod(fac2sparse(dat$id_)) #tcrossprod(class.ind2.Mat(dat$id_)) #IIt<-tcrossprod(class.ind2(dat$id_)) ## much slower, specially for larga matrices

  # Mx3<- as.matrix(crossprod( sign(Wx2)*sqrt(abs(Wx2))*X*Psi, IIt %*% (sqrt(abs(Wx2))*X*Psi) ))
  # Mxq3<-as.matrix(crossprod( sign(Wx2)*sqrt(abs(Wxq2))*X*Psi, IIt %*% (sqrt(abs(Wxq2))*Qmat.d1*Psi) ))
  # Mq3<- as.matrix(crossprod( Qmat.d1*Psi, IIt %*% (Qmat.d1*Psi) ))

  # Compute the signed residual score component (same as in FIM.glm):
  Wxq2_emp <- Sc$w * (Sc$Y - Sc$mu) / (disp * Sc$mu.eta.val)
  
  Mx3  <- as.matrix(crossprod( Wxq2_emp*X*Psi, IIt %*% (Wxq2_emp*X*Psi) ))
  Mxq3 <- as.matrix(crossprod( Wxq2_emp*X*Psi, IIt %*% (Qmat.d1*Psi) ))
  Mq3  <- as.matrix(crossprod( Qmat.d1*Psi,    IIt %*% (Qmat.d1*Psi) ))
  
  expI3<-as.matrix(rbind(cbind(Mx3, Mxq3),
                         cbind(t(Mxq3),Mq3)))

  expFIM=expI1-expI2+expI3

  return(list(FIM=expFIM,I1=expI1,I2=expI2,I3=expI3,U=U))
}

Testing_EM_joint0 <- function(theta,q,Betas_ids=NULL,G_,FIM.obj){
  if(is.null(Betas_ids)) Betas_ids<- which(names(theta)%in%G_)
  nbetas<-length(theta)
  phi<-c(theta,q[,"q"])
  n_phi<-length(phi)

  Q1<-FIM.obj$FIM
  #Vcov matrix
  D<-rbind(diag(n_phi-1),0); F1 <-  t(D) %*% Q1 %*% D
  Omega1 <- tryCatch(solve( F1 ), error=function(e){ ginv(F1) })

  ###  Score test (following Lin's paper)
  IndU<-rep(FALSE,n_phi); IndU[Betas_ids]<-TRUE
  U<-FIM.obj$U
  U1<-U[IndU]

  Ind<-rep(FALSE,n_phi-1); Ind[Betas_ids]<-TRUE
  invF1_noInd<-tryCatch(solve(F1[!Ind,!Ind]), error=function(e){ ginv(F1[!Ind,!Ind]) })
  V1<-F1[Ind,Ind] - F1[Ind,!Ind] %*% invF1_noInd %*% F1[!Ind,Ind]

  invOmega1_Ind<- tryCatch(solve(Omega1[Ind,Ind]), error=function(e){ ginv(Omega1[Ind,Ind]) })
  W<-as.numeric(t(phi[IndU]) %*% invOmega1_Ind %*% phi[IndU])
  invV1<- tryCatch(solve( V1 ), error=function(e){ ginv(V1) })
  S<-as.numeric(t(U1) %*% invV1 %*% U1)

  return(list(Var=diag(Omega1)[1:nbetas],W=W,S=S,df=length(Betas_ids)))
}

Testing_EM_joint<-function(...){
  tryCatch(Testing_EM_joint0(...),error = function(e){list(Var=NA,W=NA,S=NA,df=NA)} )
}

# svdsubsel from http://bwlewis.github.io/GLM/svdss.html
#
# Input m*p matrix A, m >= p.
# Number of output columns k<=p.
# Returns an index subset of columns of A that *estimates* the k most linearly
# independent columns of A.

svdsubsel <- function(A,k=ncol(A))
{
  S <- svd(scale(A,center=FALSE,scale=TRUE), k)
  n <- which(svd(A)$d < 2*.Machine$double.eps)[1]
  if(!is.na(n) && k>=n)
  {
    k <- n - 1
    warning("k was reduced to match the rank of A")
  }
  Q <- qr( t(S$v[,1:k]) ,LAPACK=TRUE)
  sort(Q$pivot[1:k],decreasing=FALSE)
}

### taken from https://bwlewis.github.io/GLM/ and solves some instability when zero weights are present and other edge cases in glm estimation (note this implementation does not handle offsets yet)
irls_svdnewton = function(A, b, family=binomial, maxit=25, tol=1e-08, weights=rep(1,nrow(A)), rank_deficiency=c("select columns","minimum norm","error")) {
  rank_deficiency = match.arg(rank_deficiency)
  m = nrow(A)
  n = ncol(A)
  select = 1:n
  zw = weights==0
  if(any(zw)) A[zw,]=0
  S = svd(A)
  tiny_singular_values = S$d/S$d[1] < tol
  k = sum(tiny_singular_values)
  if(k>0)
  {
    if(rank_deficiency=="select columns")
    {
      warning("Numerically rank-deficient model matrix")
      # NB This is a different selection method than R's default glm.fit uses.
      # See https://bwlewis.github.io/GLM and https://bwlewis/github.io/GLM/svdss.html
      select = svdsubsel(A,n-k)
      S = svd(A[,select,drop=FALSE])
    } else if(rank_deficiency=="error")
    {
      stop("Near rank-deficient model matrix")
    }
  }
  t = rep(1e-8,m) ## change to a small value to avoid dividing by 0 in with an inverse link.
  s = rep(0,length(select))
  good = weights > 0
  for(j in 1:maxit)
  {
    g       = family$linkinv(t[good])
    varg    = family$variance(g)
    if(any(is.na(varg))) stop("NAs in variance of the inverse link function")
    if(any(varg==0)) stop("Zero value in variance of the inverse link function")
    gprime  = family$mu.eta(t[good])
    if(any(is.na(gprime))) stop("NAs in the inverse link function derivative")
    z       = rep(0,m)
    W       = rep(0,m)
    z[good] = t[good] + (b[good] - g) / gprime ### from glm.fit (may help with offsets z <- (eta - offset)[good] + (y - mu)[good]/mu.eta.val[good])
    W[good] = weights[good] * as.vector(gprime^2 / varg)

    good   = W > .Machine$double.eps*2
    if(sum(good)<m) warning("Tiny weights encountered")
    s_old   = s
    C   = chol(crossprod(S$u[good,,drop=FALSE], W[good]*S$u[good,,drop=FALSE]))
    s   = forwardsolve(t(C), crossprod(S$u[good,,drop=FALSE],W[good]*z[good]))
    s   = backsolve(C,s)
    t   = rep(0,m)
    t[good] = S$u[good,,drop=FALSE] %*% s
    if(sqrt(crossprod(s - s_old)) < tol) break
  }
  if( j>=maxit ) warning("Maximum number of irls iterations reached.")
  x = rep(NA, n)
  if(rank_deficiency=="minimum norm") S$d[tiny_singular_values] = Inf
  x[select] = S$v %*% ((1/S$d) * crossprod(S$u[good,],t[good]))
  names(x)=colnames(A)
  eta = as.vector(if (NCOL(A) == 1L){ A * x }else A %*% x)
  mu = family$linkinv(eta)
  residuals = (b - mu)/family$mu.eta(eta)
  qr = qr(A*as.numeric(sqrt(W)))
  list(coefficients=x,iterations=j,residuals=residuals,qr=qr)
}

#' Lagrange-multiplier optimal phase 2 design for a two-phase genetic association study.
#'
#' Computes stratum-specific sampling fractions that maximise the anticipated
#' Fisher information for the regression parameters of interest subject to a
#' fixed total phase 2 sample size.  The optimisation is performed by the
#' COBYLA algorithm (\code{nloptr}) with a derivative-free fallback
#' (\code{dfoptim::hjkb}) if COBYLA fails to converge.
#'
#' @param formula A two-sided \code{\link{formula}} specifying the regression
#'   model of interest, e.g.\code{Y ~ G + Z}.  All variables except
#'   \code{miscov} must be present in \code{data}.
#' @param miscov Right-hand-side formula identifying the missing-by-design
#'   covariate(s) (the sequencing variant), e.g.\code{~ G}.
#' @param auxvar Right-hand-side formula identifying the auxiliary variable
#'   (the GWAS tag SNP), e.g.\code{~ Z}.  Must be present in \code{data}.
#' @param strata Right-hand-side formula defining the stratification variables
#'   used to cross-tabulate \code{data}, e.g.\code{~ Z + S}.  All variables
#'   must be present in \code{data}.
#' @param family GLM family object (see \code{\link{family}}).  Passed
#'   directly to the information-matrix calculations.  Default:
#'   \code{gaussian()}.
#' @param n Integer.  Target phase 2 sample size (number of individuals to
#'   be sequenced).
#' @param data A \code{data.frame} containing the phase 1 data.  Must include
#'   all variables in \code{formula} (except those in \code{miscov}),
#'   \code{auxvar}, and \code{strata}.
#' @param beta Named or unnamed numeric vector of assumed regression
#'   coefficients, with one element per term in \code{formula} (including
#'   the intercept).  These define the operating point at which the
#'   Fisher information is evaluated.  Setting \code{beta[G] = 0} triggers
#'   the score-test (null-model) criterion; any non-zero value triggers the
#'   Wald (alternative-model) criterion.
#' @param p_gz A \code{data.frame} with the assumed joint distribution of
#'   (\code{miscov}, \code{auxvar}).  Must contain one column per variable
#'   in \code{miscov} and \code{auxvar} plus a column \code{q} holding
#'   non-negative probability masses that sum to 1.  Rows with
#'   \code{q <= 0} are silently removed.  Typically generated by
#'   \code{\link{twoPhaseDesign}} from \code{maf_G} and \code{LD}, or
#'   supplied directly by the user.
#' @param disp Numeric dispersion parameter.  For Gaussian and Gamma
#'   families this is the variance (\eqn{\sigma^2}) or dispersion
#'   (\eqn{\phi}), respectively; for binomial and Poisson it is fixed at
#'   1.  If \code{NULL} (default) it is estimated as \code{var(Y)}.
#' @param optimMeasure Character string specifying the optimality criterion.
#'   One of \code{"Par-spec"} (parameter-specific: minimise the asymptotic
#'   variance of \code{beta[K.idx]}), \code{"A-opt"} (A-optimality:
#'   minimise \code{trace(FIM^{-1})}), or \code{"D-opt"} (D-optimality:
#'   maximise \code{log det(FIM)}).
#' @param K.idx Integer index (or vector of indices) into \code{beta}
#'   identifying the parameter(s) of primary interest for
#'   \code{optimMeasure = "Par-spec"}.  Ignored for A- and D-optimality.
#'   Required when \code{optimMeasure = "Par-spec"}.
#' @param min.nk Optional integer scalar or vector of length equal to the
#'   number of strata.  Specifies the minimum number of individuals to
#'   sample from each stratum.  If a scalar, the same minimum is applied to
#'   all strata.  Default: \code{NULL} (no minimum, i.e. zero).
#' @param logical.sub Optional logical vector of length \code{nrow(data)}.
#'   If supplied, only the rows where \code{logical.sub = TRUE} are used in
#'   the information calculations (the full \code{data} is still returned).
#'   Useful for sensitivity analyses on subsets.  Default: \code{NULL}.
#'
#' @details
#' The function computes, for each stratum defined by \code{strata}, the
#' stratum-specific sampling fraction \eqn{\pi_k = n_k / N_k} that maximises
#' the chosen information criterion.  The optimisation is constrained so that
#' (i) the total selected sample equals \code{n}, (ii) no stratum is
#' over-sampled (\eqn{n_k \le N_k}), and (iii) the \code{min.nk} lower
#' bounds are respected.
#'
#' The baseline Fisher information uses the empirical score outer-product
#' estimator \eqn{I_3/N} (see \code{\link{optimTP.GA}} for details).
#'
#' @return A \code{data.frame} with one row per stratum (as defined by
#'   \code{strata}) and the following columns from
#'   \code{xtabs(strata, data)}:
#'   \describe{
#'     \item{(strata variables)}{The stratification variable values.}
#'     \item{Freq}{Stratum size \eqn{N_k}.}
#'     \item{prR_cond_optim}{Optimal conditional sampling probability
#'       \eqn{n_k / n} (proportional allocation weight).}
#'     \item{convergence}{Convergence code from the solver (\code{0} = success;
#'       codes from \code{nloptr::cobyla} or \code{dfoptim::hjkb}).}
#'   }
#'
#' @seealso \code{\link{twoPhaseDesign}}, \code{\link{optimTP.GA}}
#'
#' @examples
#' \donttest{
#' data <- DataGeneration_TPD(N = 1000)
#' data <- data[, c("Y", "Z")]  # phase-1 data: no G column
#' p_gz <- data.frame(Z = c(0, 0, 0, 1, 1, 1, 2, 2, 2), # theoretical joint distribution
#'         G1 = c(0, 1, 2, 0, 1, 2, 0, 1, 2), # between G and Z (LD = 0.75, MAF_G = 0.2, MAF_Z = 0.3)
#'         q = c(0.486, 0.004, 0, 0.143, 0.276, 0.001, 0.011, 0.04, 0.039)) 
#'         
#' data$S <- as.numeric(cut(data$Y,
#'             breaks = quantile(data$Y, probs = c(0, 0.4, 0.8, 1)),
#'             include.lowest = TRUE))
#'
#' res_lm <- optimTP.LM(
#'   formula     = Y ~ Z + G1,
#'   miscov      = ~ G1,
#'   auxvar      = ~ Z,
#'   strata      = ~ Z + S,
#'   family      = gaussian(),
#'   n           = 200,
#'   data        = data,
#'   beta        = c(2, 0.5, 0.1),
#'   p_gz        = p_gz,
#'   optimMeasure = "Par-spec",
#'   K.idx       = 2L
#' )
#' head(res_lm)
#' }
#'
#' @export
#' @importFrom nloptr cobyla
#' @importFrom dfoptim hjkb
#' @importFrom utils combn stack tail
#' @importFrom enrichwith enrich
optimTP.LM  <-  function(formula,miscov,auxvar,strata,family,n,data,beta,p_gz,disp=NULL,optimMeasure,K.idx=NULL,min.nk=NULL,logical.sub=NULL){

  if (is.character(family))
    family <- get(family, mode = "function", envir = parent.frame())
  if (is.function(family))
    family <- family()
  if (is.null(family$family)) {
    message(deparse(family))
    stop("'family' not recognized")
  }

  ### TODO: check if all unique values in data[,Z_] are in p_gz

  terms_dat <- unique( c(all.vars(formula), attr(terms(miscov), "term.labels") , attr(terms(auxvar), "term.labels"), attr(terms(strata), "term.labels") , attr(terms(auxvar), "term.labels")) )

  Y_ <- as.character(formula[[2]])
  G_ <- all.vars(miscov)
  Z_ <- all.vars(auxvar)
  S_ <- all.vars(strata)

  if( is.null(disp) ){
    if( family$family %in% c("gaussian","Gamma") ){
      disp = var(data[,Y_])
    }else disp=1
  }

  ## remove enties with zero or negative probabilities
  p_gz <- p_gz[p_gz$q>0,]

  N=NROW(data)
  rho=n/N
  if( rho<0.1 )warning("The phase 2 sample size (n2) is smaller than recommended. Please proceed with caution.")

  stratadf <- data.frame(xtabs(strata,data=data))
  #Xs <- model.matrix(strata, stratadf)
  stratadf[S_] <- lapply(stratadf[S_], function(x)as.numeric(as.character(x))) ## may need it for data.table
  g.vals <- unique(p_gz[,G_])
  z.vals <- unique(p_gz[,Z_])

  if( !is.null(logical.sub) ){
    data <- data[logical.sub,]
    N <- NROW(data)
  }

  dat_ <- cbind(id_=rep(seq(1,N,by=1), each=NROW(g.vals)),data[rep(seq_len(N), each=NROW(g.vals)), ],matrix(rep(t(g.vals),N), ncol=NCOL(g.vals), byrow=TRUE))
  names(dat_)[(ncol(dat_)-NCOL(g.vals)+1):ncol(dat_)] <- G_
  dat_ <- droplevels(dat_)
  dat_$ord_ <- seq_len(nrow(dat_))

  wgs <- omega1_g(formula,family,dat_,beta,p_gz,disp,G_,Z_) # P(G|Y,Z) and P(Y,Z)

  dat_$wg_ <- wgs$wg
  dat_ <- dat_[!is.na(dat_$wg_),]

  ## Baseline FIM: I3/N (empirical score outer product; valid large-N approximation)
  obsIM_all_lm <- obsIMf1(formula,family,dat_,beta,p_gz,disp,G_,Z_,by.id = FALSE)
  M1 <- obsIM_all_lm$I3 / N

  ## Per-stratum I3 contributions (phase-2 improvement, by stratum)
  IM3_subs <- do.call(rbind,by(dat_, INDICES = dat_[,S_],function(df.by){
    Nk <- N
    IM <- obsIMf1(formula,family,df.by,beta,p_gz,disp,G_,Z_)
    data.frame(df.by[1,S_,drop=FALSE],t(as.vector(IM$I3)/Nk))
  }))

  ## Per-stratum I2 contributions
  IM2_subs <- do.call(rbind,by(dat_, INDICES = dat_[,S_],function(df.by){
    Nk <- N
    IM <- obsIMf1(formula,family,df.by,beta,p_gz,disp,G_,Z_)
    data.frame(df.by[1,S_,drop=FALSE],t(as.vector(IM$I2)/Nk))
  }))

  lenS_ <- length(S_)
  npars <- NCOL(M1)

  ### Indicator for the beta parameters of interest
  Ind <- rep(FALSE,npars-1)
  Betas_ids <- which(grepl(paste0(paste0("^",G_),collapse="|"),colnames(model.matrix(formula,dat_[1,]))))
  Ind[Betas_ids] <- TRUE

  if( optimMeasure=="Par-spec" & is.null(K.idx) ) stop("For a parameter-specific criterion K.idx must be provided.")
  ## asign these functions to the execution environment
  optMsuref=.optMsure(optimMeasure);  environment(optMsuref)=environment()

  ### parameters for the constained optimization
  # if( !all.equal(pS_[S_],stratadf[,S_],attributes=FALSE,check.attributes = FALSE) ) stop("Strata groups in pS and stratadf don't match!")
  # ais <- (stratadf$Freq)#/sum(stratadf$Freq)
  # ais <- N*pS_$pS_#/sum(stratadf$Freq)
  npis <- nrow(stratadf)-1
  rho <- n
  upperLM <- as.numeric(stratadf$Freq) #rep(1,npis+1)
  # M1inv <- MASS::ginv(M1)
  # In <- diag(npars)
  D <- rbind(diag(npars-1),0)
  # Dinv <- MASS::ginv(D)
  # Dtinv  <- MASS::ginv((t(D)))
  ### result on how to calcuate the inverse of A+B
  # https://math.stackexchange.com/questions/17776/inverse-of-the-sum-of-matrices
  # (A+B)^{-1} = A^{-1}(I-(I+BA^{-1})^{-1}BA^{-1})


  ### The LM approach is going to get optimal values for nk/P(R=1,Z,K) as opposed to pr(R=1|Y,Z). In additio the  objective function is determined according to whether the design regression parameters are under the null or alternative

  if( all(beta[Ind]==0) ){
    ObjFun <- function(pis){ # pis = (upperLM*n/N)[-1]
      pis1 <- c(0,pis)
      pis1[1] <- (rho-sum(pis1))
      stratadf$prReq1YZ <- (pis1/N)/stratadf$Freq

      IM2b_subs <- as.data.frame(merge(as.data.table(IM2_subs), as.data.table(stratadf[,c(S_,"prReq1YZ")]), by=S_, sort = FALSE))
      IM3b_subs <- as.data.frame(merge(as.data.table(IM3_subs), as.data.table(stratadf[,c(S_,"prReq1YZ")]), by=S_, sort = FALSE))

      FIM <- M1 + matrix(colSums(IM2b_subs[,(lenS_+1):(npars^2+lenS_)]*IM2b_subs$prReq1YZ),ncol=npars) - matrix(colSums(IM3b_subs[,(lenS_+1):(npars^2+lenS_)]*IM3b_subs$prReq1YZ),ncol=npars)

      ## to deal with the constraint on the p_g's
      F1 <-  crossprod(D, FIM %*% D)

      ###  Score test variance
      invF1_noInd <- ginv(F1[!Ind,!Ind])
      V1 <- F1[Ind,Ind] - ( F1[Ind,!Ind] %*% invF1_noInd %*% F1[!Ind,Ind] )
      invV1 <- solve( V1 )

      return( optMsuref(invV1,K.idx) )
    }
  }else {
    ObjFun <- function(pis){ # pis = (upperLM*n/N)[-1]
      pis1 <- c(0,pis)
      pis1[1] <- (rho-sum(pis1))
      stratadf$prReq1YZ <- (pis1/N)/stratadf$Freq

      IM2b_subs <- as.data.frame(merge(as.data.table(IM2_subs), as.data.table(stratadf[,c(S_,"prReq1YZ")]), by=S_, sort = FALSE))
      IM3b_subs <- as.data.frame(merge(as.data.table(IM3_subs), as.data.table(stratadf[,c(S_,"prReq1YZ")]), by=S_, sort = FALSE))

      FIM <- M1 + matrix(colSums(IM2b_subs[,(lenS_+1):(npars^2+lenS_)]*IM2b_subs$prReq1YZ),ncol=npars) - matrix(colSums(IM3b_subs[,(lenS_+1):(npars^2+lenS_)]*IM3b_subs$prReq1YZ),ncol=npars)

      ## to deal with the constraint on the p_g's
      F1 <-  crossprod(D, FIM %*% D)

      # B <- matrix(colSums(IM2b_subs[,(lenS_+1):(npars^2+lenS_)]*IM2b_subs$prReq1YZ),ncol=npars) - matrix(colSums(IM3b_subs[,(lenS_+1):(npars^2+lenS_)]*IM3b_subs$prReq1YZ),ncol=npars)
      #
      #
      # FIMinv <- M1inv %*% (In - solve(In+B%*%M1inv)) %*% (B%*%M1inv)
      #
      # vcov = Dinv %*% FIMinv %*% Dtinv

      vcov = solve(F1)

      return( optMsuref(vcov,K.idx) )

    }
  }


  if( is.null(min.nk) ){
    minLB.val <- 0 #ifelse(rho<0.05,rho*0.01,0.05)
    minLB <- rep(minLB.val,npis+1)
  }else{
    if( !length(min.nk) %in% c(1,npis+1) ) stop("The number of elements of min.nk is either 1 or ",npis+1)
    if( any(min.nk<0) )stop("The minimum stratum size must greater or equal than zero")
    if( any(min.nk>stratadf$Freq) )stop("The minimum stratum size must not be greater than the strata sample sizes determined by strata")
    minLB <- min.nk
  }
  initX0 <- (upperLM*n/N)[-1] #rep(rho,npis)
  initX1 <- (upperLM*n/N)  #rep(rho,npis+1)

  ObjFun_in <- function(pis){
    ui <- rbind(rep(-1,npis),rep(1,npis))
    ci <- c(-rho,0)
    return(as.numeric(ui%*%pis-ci))
  }

  ObjFun_eq <- function(pis1){
    return(sum(pis1)-rho)
  }

  ### Mind the warning: "For consistency with the rest of the package the inequality sign may be switched from >= to <= in a future nloptr version."

  ## To disable this warning and avoid printing do the following. However, new versions of the package may return an error when  this is enforced.
  options(nloptr.show.inequality.warning = FALSE)

  ### Note that the function can be changed when the change is implemented using
  # hin <- function(x) (-1)*f2(x, ...)  # NLOPT expects hin <= 0

  ### use this one for base, if an error then use the rest
  sol <- cobyla(initX0, fn = ObjFun, lower = minLB[-1], upper = upperLM[-1], hin = ObjFun_in, nl.info = FALSE, control = list(xtol_rel = 1e-6, maxeval = 10000))

  # return the option to the original state
  options(nloptr.show.inequality.warning = TRUE)

  if( !(sol$convergence>0 & sol$convergence<5) ){
    if( sol$convergence==5 ) message("The maximum number of evaluations in nloptr:::cobyla() has been reached. Trying dfoptim:::hjkb() now. You could also increase the number of evaluations via maxeval (although it is fairly large already.)")
    if( sol$convergence<0 ) message(paste0("Convergence in nloptr:::cobyla() was not achieved; error code ",sol$convergence,". Trying dfoptim:::hjkb() now."))

    ObjFunb <- function(pisA, k){ # pisA <- (upperLM*n/N)
      nstar <- sum(pisA)
      return( ObjFun(pisA[-1]) + k*(nstar-rho)^2 )
    }

    sol1 <- hjkb(initX1, fn=ObjFunb, lower=minLB, upper=upperLM, k=0.01, control=list(maxfeval=10000))
    sol1 <- tryCatch(hjkb(sol1$par, fn=ObjFunb, lower=minLB, upper=upperLM, k=100, control=list(maxfeval=10000)), error=function(e){ NULL })
    if( is.null(sol1) ){
      solB <- c(0,sol$par)
      solB[1] <- (rho-sum(solB))

      sol1 <- hjkb(solB, fn=ObjFunb, lower=minLB, upper=upperLM, k=100, control=list(maxfeval=10000))
    }

    if( sol1$convergence!=0 ){
      warning(paste0("Convergence in dfoptim:::hjkb() was not achieved. Error code ",sol1$convergence,". Returning last par value."))
    }

    stratadf$convergence <- c(sol1$convergence,rep(NA,nrow(stratadf)-1))
    pisopt <- sol1$par

  } else {
    stratadf$convergence <- c(sol$convergence,rep(NA,nrow(stratadf)-1))
    pisopt <- c(0,sol$par)
    pisopt[1] <- (rho-sum(pisopt))
  }

  stratadf$prR_cond_optim <- pisopt/sum(pisopt)

  return(stratadf)
}


#' Genetic-algorithm optimal phase 2 design for a two-phase genetic association study.
#'
#' Uses a genetic algorithm (GA) to search over individual-level phase 2
#' selection indicators \eqn{R_i \in \{0,1\}} for the subset of size
#' \code{n} that maximises the anticipated Fisher information for the
#' regression parameters of interest.  Wraps \code{\link[kofnGA]{kofnGA}}.
#' When a parallel cluster is supplied, multiple independent GA runs are
#' executed in parallel and the best solution across all runs is returned.
#'
#' @param formula A two-sided \code{\link{formula}} specifying the regression
#'   model of interest, e.g. \code{Y ~ G + Z}.  All variables except those in
#'   \code{miscov} must be present in \code{data}.
#' @param miscov Right-hand-side formula identifying the missing-by-design
#'   covariate(s) (the sequencing variant), e.g. \code{~ G}.
#' @param auxvar Right-hand-side formula identifying the auxiliary variable
#'   (the GWAS tag SNP), e.g. \code{~ Z}.  Must be present in \code{data}.
#' @param family GLM family object (see \code{\link{family}}).  Default:
#'   \code{gaussian()}.
#' @param n Integer.  Target phase 2 sample size.
#' @param data A \code{data.frame} containing the phase 1 data.  Must include
#'   all variables in \code{formula} (except those in \code{miscov}) and
#'   \code{auxvar}.
#' @param beta Numeric vector of assumed regression coefficients, one per
#'   term in \code{formula} (including the intercept).  Setting
#'   \code{beta[G] = 0} triggers the score-test (null-model) criterion; any
#'   non-zero value triggers the Wald (alternative-model) criterion.
#' @param p_gz A \code{data.frame} with the assumed joint distribution of
#'   (\code{miscov}, \code{auxvar}).  Must contain columns for each variable
#'   in \code{miscov} and \code{auxvar} and a column \code{q} of
#'   non-negative probability masses summing to 1.  Rows with \code{q <= 0}
#'   are silently removed.
#' @param disp Numeric dispersion parameter (variance for Gaussian/Gamma;
#'   fixed at 1 for binomial/Poisson).  If \code{NULL} (default),
#'   estimated as \code{var(Y)}.
#' @param ga.popsize Integer.  GA population size (number of candidate
#'   solutions maintained per generation).
#' @param ga.propelit Numeric in \eqn{(0, 1)}.  Proportion of the population
#'   that is elite (carried forward unchanged to the next generation).
#'   The elite count is \code{ceiling(ga.popsize * ga.propelit)}.
#' @param ga.proptourney Numeric in \eqn{(0, 1)}.  Proportion of the
#'   population used in each tournament selection.  The tournament size is
#'   \code{max(ceiling(ga.popsize * ga.proptourney), 2)}.
#' @param ga.ngen Integer.  Number of GA generations to run.
#' @param ga.mutrate Numeric in \eqn{[0, 1]}.  Per-index mutation probability.
#'   At each mutation step, each of the \code{n} selected indices is
#'   independently replaced by a randomly chosen non-selected index with
#'   probability \code{ga.mutrate}.
#' @param ga.initpop Optional \code{ga.popsize}-by-\code{n} integer matrix of
#'   starting solutions (each row is a candidate subset of size \code{n}).
#'   If \code{NULL} (default), the initial population is generated randomly.
#'   When \code{cluster} is provided, the rows are split evenly across
#'   workers.
#' @param optimMeasure Character string specifying the optimality criterion.
#'   One of \code{"Par-spec"} (parameter-specific), \code{"A-opt"}
#'   (A-optimality), or \code{"D-opt"} (D-optimality).
#' @param K.idx Integer index (or vector of indices) identifying the
#'   parameter(s) of primary interest.  Required when
#'   \code{optimMeasure = "Par-spec"}.
#' @param seed Integer random seed, set before GA initialisation.  Default: 1.
#' @param verbose Integer.  If \code{> 0}, the best objective value is printed
#'   every \code{verbose} generations.  Default: \code{0} (silent).
#' @param cluster An optional parallel cluster object created by
#'   \code{\link[parallel]{makeCluster}}.  When provided,
#'   \code{length(cluster)} independent GA searches are run in parallel
#'   (each worker receives an equal share of \code{ga.initpop} rows, or a
#'   random initial population if \code{ga.initpop = NULL}), and the best
#'   solution across all workers is returned.  The caller is responsible for
#'   starting and stopping the cluster.  Default: \code{NULL} (sequential
#'   single-run GA).
#'
#' @details
#' The fitness function evaluates the anticipated Fisher information
#' matrix (FIM) for a candidate selection indicator \eqn{R}:
#' \deqn{\mathrm{FIM}(R) = I_3/N + \sum_{i: R_i=1}(I_{2,i} - I_{3,i})/N,}
#' where \eqn{I_3} is the empirical score outer-product estimator (the
#' numerically stable approximation to the full Louis (1982) observed FIM
#' valid for large \eqn{N}, since \eqn{I_1 \approx I_2} in the genotype
#' expansion context), and \eqn{I_{2,i}}, \eqn{I_{3,i}} are the
#' per-individual complete-data score outer product and observed-data
#' score outer product, respectively.  Under \eqn{H_0} (\code{beta[G] = 0})
#' the score-test Schur complement is optimised; under \eqn{H_a} the full
#' inverse FIM is optimised.
#'
#' @return An object of class \code{"kofnGA"} (see
#'   \code{\link[kofnGA]{kofnGA}}) with the following elements:
#'   \describe{
#'     \item{bestsol}{Integer vector of length \code{n} giving the row indices
#'       (in \code{data}) of the selected phase 2 individuals.}
#'     \item{bestobj}{Numeric value of the optimality criterion at
#'       \code{bestsol}.}
#'     \item{pop}{Final population matrix (\code{ga.popsize} rows,
#'       \code{n} columns), sorted by fitness.}
#'     \item{obj}{Objective values for each row of \code{pop}.}
#'     \item{old}{List with per-generation \code{best}, \code{obj}, and
#'       \code{avg} records.}
#'     \item{time.elapsed}{Named numeric vector with the pre-processing
#'       (\code{preproc}) and GA (\code{GA}) times in seconds.}
#'   }
#'
#' @seealso \code{\link[kofnGA]{kofnGA}}, \code{\link{twoPhaseDesign}},
#'   \code{\link{optimTP.LM}}
#'
#' @examples
#' \donttest{
#' data <- DataGeneration_TPD(N = 500)
#' data <- data[, c("Y", "Z")]  # phase-1 data: no G column
#' p_gz <- data.frame(Z = c(0, 0, 0, 1, 1, 1, 2, 2, 2), # theoretical joint distribution
#'         G1 = c(0, 1, 2, 0, 1, 2, 0, 1, 2), # between G and Z (LD = 0.75, MAF_G = 0.2, MAF_Z = 0.3)
#'         q = c(0.486, 0.004, 0, 0.143, 0.276, 0.001, 0.011, 0.04, 0.039)) 
#'
#' res_ga <- optimTP.GA(
#'   formula      = Y ~ Z + G1,
#'   miscov       = ~ G1,
#'   auxvar       = ~ Z,
#'   family       = gaussian(),
#'   n            = 100,
#'   data         = data,
#'   beta         = c(2, 0.5, 0.1),
#'   p_gz         = p_gz,
#'   ga.popsize   = 40,
#'   ga.propelit  = 0.1,
#'   ga.proptourney = 0.1,
#'   ga.ngen      = 50,
#'   ga.mutrate   = 0.01,
#'   optimMeasure = "Par-spec",
#'   K.idx        = 2L
#' )
#' cat("Best objective:", res_ga$bestobj, "\n")
#' cat("Selected indices (first 10):", head(res_ga$bestsol, 10), "\n")
#'
#' ## Parallel processing (2 workers)
#' cl <- parallel::makeCluster(2L)
#' res_par <- optimTP.GA(
#'   formula      = Y ~ Z + G1,
#'   miscov       = ~ G1,
#'   auxvar       = ~ Z,
#'   family       = gaussian(),
#'   n            = 100,
#'   data         = data,
#'   beta         = c(2, 0.5, 0.1),
#'   p_gz         = p_gz,
#'   ga.popsize   = 40,
#'   ga.propelit  = 0.1,
#'   ga.proptourney = 0.1,
#'   ga.ngen      = 50,
#'   ga.mutrate   = 0.01,
#'   optimMeasure = "Par-spec",
#'   K.idx        = 2L,
#'   cluster      = cl
#' )
#' parallel::stopCluster(cl)
#' cat("Best objective (parallel):", res_par$bestobj, "\n")
#' }
#'
#' @importFrom kofnGA kofnGA
#' @importFrom parallel clusterEvalQ
#' @export
optimTP.GA  <-  function(formula,miscov,auxvar,family,n,data,beta,p_gz,disp=NULL,ga.popsize,ga.propelit,ga.proptourney,ga.ngen,ga.mutrate,ga.initpop=NULL,optimMeasure,K.idx=NULL,seed=1,verbose=0,cluster=NULL){
  ptm0 <- Sys.time()
  if (is.character(family))
    family <- get(family, mode = "function", envir = parent.frame())
  if (is.function(family))
    family <- family()
  if (is.null(family$family)) {
    message(deparse(family))
    stop("'family' not recognized")
  }
  terms_dat<-unique( c(all.vars(formula), attr(terms(miscov), "term.labels") , attr(terms(auxvar), "term.labels")) )
  Y_<-as.character(formula[[2]]); G_<-all.vars(miscov); Z_<-all.vars(auxvar)

  ## remove entries with zero or negative probabilities
  p_gz <- p_gz[p_gz$q>0,]

  ### check if names in G_ and Z_ are in the joint ditribution data.frame p_gz
  g.vals<-unique(p_gz[,G_]);
  N <- NROW(data)

  if( is.null(disp) ){
    if( family$family %in% c("gaussian","Gamma") ){
      disp = var(data[,Y_])
    }else disp=1
  }
  ### Function that determines the optimality measure (D-, A- or parameter-specfic for now)
  ### if optimMeasure=="Par-spec" then K.idx MUST be not null
  if( optimMeasure=="Par-spec" & is.null(K.idx) ) stop("For a parameter-specific criterion K.idx must be provided.")

  ## assign these functions to the execution environment
  omega1_g=omega1_g; optMsuref=.optMsure(optimMeasure)#; class.ind2=class.ind2; class.ind2.Mat=class.ind2.Mat

  # envup=environment(omega1_g)=environment(class.ind2)=environment(class.ind2.Mat)=environment(optMsuref)=environment()
  environment(optMsuref) <- environment()

  uniqZ<-unique(p_gz[,Z_]); uniqZ<-uniqZ[order(uniqZ)]

  ### calculate FIMs for all samples (hopefully, a faster way than before)
  dat_ <- cbind(id_=rep(seq(1,N,by=1), each=NROW(g.vals)),data[rep(seq_len(N), each=NROW(g.vals)), ],matrix(rep(t(g.vals),N), ncol=NCOL(g.vals), byrow=TRUE))
  names(dat_)[(ncol(dat_)-NCOL(g.vals)+1):ncol(dat_)] <- G_
  dat_ <- droplevels(dat_)

  wgs <- omega1_g(formula,family,dat_,beta,p_gz,disp,G_,Z_)

  dat_$wg_ <- wgs$wg
  dat_ <- dat_[!is.na(dat_$wg_),]

  ## Baseline FIM: I3/N (empirical score outer product; valid large-N approximation
  ## to the full Louis FIM since I1 ≈ I2 for large N in the expansion context)
  obsIM_all  <- obsIMf1(formula,family,dat_,beta,p_gz,disp,G_,Z_,by.id = FALSE)
  IM1        <- obsIM_all$I3 / N

  ## Per-individual I2 and I3 for phase-2 improvement
  obsIM23_id <- obsIMf1(formula,family,dat_,beta,p_gz,disp,G_,Z_,by.id = TRUE)

  npars <- NCOL(IM1)

  IM2_id <- obsIM23_id$I2 / N
  IM3_id <- obsIM23_id$I3 / N

  Ind <- rep(FALSE,npars-1)
  Betas_ids <- which(grepl(paste0(paste0("^",G_),collapse="|"),colnames(model.matrix(formula,dat_[1,]))))
  Ind[Betas_ids] <- TRUE
  D <- rbind(diag(npars-1),0)

  if( all(beta[Ind]==0) ){
    Assess_fitness <- function(R){

      FIM <- IM1 + matrix(colSums(IM2_id*R),ncol=npars) - matrix(colSums(IM3_id*R),ncol=npars)
      F1 <-  t(D) %*% FIM %*% D

      invF1_noInd <- ginv(F1[!Ind,!Ind])
      V1 <- F1[Ind,Ind] - F1[Ind,!Ind] %*% invF1_noInd %*% F1[!Ind,Ind]
      invV1 <- solve( V1 )

      return( optMsuref(invV1,K.idx) )
    }


  } else {
    Assess_fitness <- function(R){

      FIM <- IM1 + matrix(colSums(IM2_id*R),ncol=npars) - matrix(colSums(IM3_id*R),ncol=npars)
      F1 <-  t(D) %*% FIM %*% D

      vcov = solve(F1)
      return(optMsuref(vcov,K.idx))

    }
  }

  #### R<-rep(0,N); R[sample(N,n)] <- 1
  R0s <- rep(0,N)
  ObjFun <- function(v){
    R <- R0s; R[v]<-1
    Assess_fitness(R)
  }

  time.elapsed0 <- as.numeric(difftime(Sys.time(), ptm0, units="secs"))

  ptm <- Sys.time()

  ## When cluster is non-NULL, kofnGA distributes fitness evaluations across
  ## workers each generation via parRapply.  Pre-load MASS on each worker so
  ## that ginv() is available inside the Assess_fitness closure.
  if (!is.null(cluster)) {
    parallel::clusterEvalQ(cluster, requireNamespace("MASS", quietly = TRUE))
  }

  out <- kofnGA::kofnGA(
    n           = N,
    k           = n,
    OF          = ObjFun,
    popsize     = ga.popsize,
    keepbest    = ceiling(ga.popsize * ga.propelit),
    tourneysize = max(ceiling(ga.popsize * ga.proptourney), 2),
    ngen        = ga.ngen,
    mutprob     = ga.mutrate,
    initpop     = ga.initpop,
    verbose     = verbose,
    cluster     = cluster
  )

  time.elapsed <- as.numeric(difftime(Sys.time(), ptm, units="secs"))
  out$time.elapsed <- c(preproc=time.elapsed0, GA=time.elapsed)

  return(out)
}

#' Performs inference on two-phase studies data via semiparametric maximum likelihood.
#'
#'
#' @param formula  regression formula, note that if it does not contain the
#'   missing-by-design variable, \code{miscov}, it will return results under
#'   the null hypothesis. Hypothesis testing corresponds to the score
#'   statistic. Otherwise, estimates and hypothesis testing occur under the
#'   alternative hypothesis leading to Wald statistics. All the elements in
#'   \code{formula} except \code{miscov} must be present in \code{data0} and
#'   \code{data1}.
#' @param miscov right hand side formula with the missing-by-design
#'   covariate(s), i.e. the potential causal locus (loci). Must be present in
#'   \code{data1} but absent in \code{data0}.
#' @param auxvar  right hand side formula with the auxiliary variable(s),
#'   i.e. the GWAS SNP from phase 1. Must be present in \code{data0} and
#'   \code{data1}.
#' @param family member of the exponential family (see \code{\link{family}};
#'   \sQuote{quasi} models not available). Default \code{gaussian()}.
#' @param data a data frame containing data from both phases. Phase 2 data is assumed to be the result of the \code{complete.cases} with both \code{miscov} and \code{auxvar}. Missing values in variables other than \code{miscov} are not allowed and may lead to unexpected errors.
#' @param start.values a named list with initial values for the regression
#'   parameters (\code{betas}) and/or the joint distribution between
#'   \code{miscov} and \code{auxvar} (\code{q}). Defaults to \code{NULL}.
#' @param control a list of parameters for controlling the fitting process, i.e., number or EM iterations, tolerance and whether a trace is requested, it is passed to \code{\link{twoPhaseSPML.control}}.
#'
#' @details
#' The function implements the semiparametric maximum likelihood (SPML)
#' estimator for two-phase genetic association studies described by
#' Espin-Garcia et al. The EM algorithm alternates between an E-step that
#' computes observation weights based on the current estimate of the joint
#' distribution of the missing covariate and the auxiliary variable, and an
#' M-step that updates the regression coefficients and the non-parametric
#' distribution via iteratively re-weighted least squares and a closed-form
#' update, respectively.
#'
#' When \code{miscov} is absent from \code{formula}, inference is performed
#' under \eqn{H_0} and score statistics are returned. When \code{miscov} is
#' included in \code{formula}, inference is performed under \eqn{H_a} and
#' Wald statistics are returned.
#'
#' @return An object of class \code{"twoPhaseSPML"}, which is a list
#'   containing:
#'   \describe{
#'     \item{theta}{named numeric vector of estimated regression coefficients.}
#'     \item{qGZ}{data frame with the estimated joint distribution of
#'       \code{miscov} and \code{auxvar}.}
#'     \item{dispersion}{estimated dispersion parameter (variance for
#'       Gaussian; 1 for other families).}
#'     \item{iter}{integer, number of EM iterations performed.}
#'     \item{ll}{log-likelihood at convergence.}
#'     \item{df}{degrees of freedom for the test statistic.}
#'     \item{Sobs}{(null model only) observed score test statistic.}
#'     \item{Sexp}{(null model only) expected score test statistic.}
#'     \item{var_theta}{(alternative model only) named numeric vector of
#'       marginal variances for \code{theta}.}
#'     \item{Wobs}{(alternative model only) observed Wald test statistic.}
#'     \item{Wexp}{(alternative model only) expected Wald test statistic.}
#'     \item{qG}{data frame with the estimated marginal distribution of
#'       \code{miscov}.}
#'     \item{formula}{the \code{formula} used.}
#'     \item{miscov}{the \code{miscov} formula used.}
#'     \item{family}{the \code{family} object used.}
#'     \item{null.model}{logical, \code{TRUE} when the model was fitted under
#'       \eqn{H_0}.}
#'   }
#'
#'
#' @examples
#'
#' data.table::setDTthreads(1)
#' data <- DataGeneration_TPD(N=200)
#' set.seed(1)
#' R <- rep(0, nrow(data))
#' R[sample(nrow(data), 175)] <- 1   # random phase 2 subsample of 500
#' data[R == 0, c("G1")] <- NA
#'
#' # Null model (score test)
#' res_Ho <- twoPhaseSPML(
#'   formula = Y ~ Z,
#'   miscov  = ~ G1,
#'   auxvar  = ~ Z,
#'   data   = data
#' )
#' print(res_Ho)
#'
#' # Alternative model (Wald test)
#' res_Ha <- twoPhaseSPML(
#'   formula = Y ~ Z + G1,
#'   miscov  = ~ G1,
#'   auxvar  = ~ Z,
#'   data   = data
#' )
#' print(res_Ha)
#'
#' @export
#' @import stats
#' @import data.table
#' @importFrom MASS ginv
#' @importFrom Matrix fac2sparse t crossprod colSums
twoPhaseSPML <- function(formula, miscov, auxvar, family = gaussian, data, start.values = NULL, control=list())
{
  if (is.character(family))
    family <- get(family, mode = "function", envir = parent.frame())
  if (is.function(family))
    family <- family()
  if (is.null(family$family)) {
    message(deparse(family))
    stop("'family' not recognized")
  }

  terms_dat <- unique(c(all.vars(formula),
                         attr(terms(miscov), "term.labels"),
                         attr(terms(auxvar), "term.labels")))

  Y_ <- as.character(formula[[2]])
  G_  <- all.vars(miscov)
  Z_  <- all.vars(auxvar)
  terms_data0 <- if (all(G_ %in% all.vars(formula))) {
    c(all.vars(formula)[-which(all.vars(formula) %in% G_)], Z_)
  } else {
    c(all.vars(formula), Z_)
  }
  terms_data0 <- unique(terms_data0)
  
  ### construct data0 and data1 from data, where data0 is a data frame with the complement of the phase 2 data, must contain all variables in \code{formula} and \code{auxvar} but \emph{not} \code{miscov}; data1 is a data frame with the phase 2 data, must contain all variables
  Ridx <- as.numeric(complete.cases(data[c(G_,Z_)]))
  data0 <- data[Ridx==0, terms_data0]
  data1 <- data[Ridx==1, ]
  
  n1 <- nrow(data1); n0 <- nrow(data0); n <- n1 + n0
  nG_ <- length(G_)
  uniqZ <- unique(data1[, Z_]); uniqZ <- uniqZ[order(uniqZ)]

  uniqG <- unique(data1[, G_])
  dat0 <- cbind(id_ = rep(seq(n1 + 1, n, by = 1), each = NROW(uniqG)),
                data0[rep(seq_len(n0), each = NROW(uniqG)), ],
                matrix(rep(t(uniqG), n0), ncol = NCOL(uniqG), byrow = TRUE))
  names(dat0)[(ncol(dat0) - NCOL(uniqG) + 1):ncol(dat0)] <- G_
  qform <- as.formula(paste0("wg_ ~", paste(G_, collapse = "+"), "+", Z_))

  dat <- rbind(cbind(id_ = 1:n1, data1[, terms_dat]),
               dat0[, c("id_", terms_dat)])

  # Initialise the non-parametric terms: p_{g,z}
  uniqGZ <- unique(data1[, c(G_, Z_)])
  q0 <- cbind(uniqGZ, q = 1 / nrow(uniqGZ))
  Zuniq <- unique(c(data1[, Z_], data0[, Z_]))
  q <- cbind(matrix(rep(t(uniqG), length(Zuniq)),
                    ncol = NCOL(uniqG), byrow = TRUE),
             Z = rep(Zuniq, each = NROW(uniqG)))
  colnames(q) <- c(G_, Z_)
  q <- merge(q, q0, all.x = TRUE, sort = FALSE)
  q[is.na(q$q), "q"] <- 0
  nq <- nrow(q)
  if (!is.null(start.values$q)) {
    q1 <- merge(q, start.values$q, by = c(G_, Z_), all.x = TRUE, sort = FALSE)
    q1 <- q1[, c(G_, Z_, "q.y")]
    names(q1)[3] <- "q"
    q <- q1
    # normalization check
    total_q <- sum(q[, "q"], na.rm = TRUE)
    if (abs(total_q - 1) > 1e-8)
      warning("start.values$q does not sum to 1 (sum = ", total_q,"); normalizing.")
    q[, "q"] <- q[, "q"] / total_q
  }

  # Initialise regression parameters
  X <- model.matrix(formula, dat)

  if (is.null(start.values$betas)) {
    theta.o <- glm(formula, data = data1, family = family)$coef
    if(any(is.na(theta.o))) theta.o <- glm(formula, data = dat, family = family)$coef
  } else {
    theta.o <- start.values$betas
  }

  nparms <- length(theta.o)
  
  if (family$family %in% c("gaussian", "Gamma")) {
    # disp <- var(c(data1[, Y_], data0[, Y_]))
    glm_init <- glm(formula, data = dat, family = family)
    theta.o   <- glm_init$coef
    disp      <- sum(glm_init$weights * glm_init$residuals^2) / (n - nparms)
  } else {
    disp <- 1
  }

  ng <- NROW(uniqG)
  n1ones <- rep(1, n1)

  # EM convergence settings
  control <- do.call("twoPhaseSPML.control", control)
  maxiter <- control$EMmaxit
  tol     <- control$EMtol
  iter    <- 0L
  theta   <- theta.o
  theta.o <- theta + 1
  q.o     <- q
  q.o[, "q"] <- q.o[, "q"] - 1
  if (control$trace) message("Initial iteration:", iter, "theta =", theta,
                   "pGZ =", as.numeric(q[, "q"]), "\n")

  while ( iter <= maxiter &
         (max(abs(theta - theta.o)) > tol |
          max(abs(q[, "q"] - q.o[, "q"])) > tol) ) {

    theta.o <- theta; q.o <- q; iter <- iter + 1

    # E-step
    dat$wg_ <- c(n1ones,
                 fnc_wgts(formula, family, dat0, theta, q, disp, G_, Z_, uniqZ))

    # M-step
    glmfit <- suppressWarnings(
      irls_svdnewton(X, dat[, Y_], family = family,
                     weights = dat$wg_, rank_deficiency = "select columns")
    )
    theta <- glmfit$coef

    if (family$family %in% c("gaussian", "Gamma")) {
      # disp <- sum(dat$wg_ * glmfit$residuals^2) / (n - nparms)
      mu_fit  <- family$linkinv(X %*% theta)
      pearson <- (dat[, Y_] - mu_fit) / sqrt(family$variance(mu_fit))
      disp    <- sum(dat$wg_ * pearson^2) / (n - nparms)
    }
    q.it <- data.frame(xtabs(qform, data = dat) / n)

    ordq <- match(apply(q[, c(G_, Z_)], 1, paste, collapse = "-"),
                  apply(q.it[, c(G_, Z_)], 1, paste, collapse = "-"))
    q[, "q"] <- q.it[ordq, "Freq"]

    if (control$trace) message("Iteration:", iter, "theta =", theta,
                     "pGZ =", as.numeric(q[, "q"]), "\n")
  }
  # Final wg calculation at convergence (or in case a dry run is requested)
  dat$wg_ <- c(n1ones,
              fnc_wgts(formula, family, dat0, theta, q, disp, G_, Z_, uniqZ))

  if (control$trace) message("Total iterations:", iter,
                   "theta_end =", theta,
                   "pGZ_end =", as.numeric(q[, "q"]), "\n")

  # Objects for hypothesis testing
  qG <- aggregate(as.formula(paste0("q~", paste(G_, collapse = "+"))), q, FUN = sum)
  qG$k_ <- seq_len(nrow(qG))
  dat$ord_ <- seq_len(nrow(dat))
  dat <- merge(dat, qG[, c(G_, "k_")], by = G_)
  dat <- dat[order(dat$ord_), ]

  Ho <- !all(G_ %in% names(theta))   # TRUE when fitting under H0

  if (Ho) {
    indx_thetas_H0 <- which(!G_ %in% names(theta))
    nG_ <- length(indx_thetas_H0)
    val <- c(theta, rep(0, nG_))
    id  <- c(seq_along(theta), 1 + seq(0.1, 0.9, length.out = nG_))
    theta1 <- val[order(id)]
    names(theta1)[which(names(theta1) == "")] <- G_[indx_thetas_H0]
    formula1 <- update(formula,
                       paste0(".~", paste(G_[indx_thetas_H0], collapse = "+"), "+ ."))
    X1  <- model.matrix(formula1, dat)
    eta1 <- X1 %*% theta1
    W   <- family$mu.eta(eta1)^2 / family$variance(family$linkinv(eta1))
    glmfit1 <- list()
    glmfit1$qr <- qr(X1 * sqrt(as.numeric(dat$wg_ * W)))

    FIM    <- FIM.glm(theta1, qG, disp, formula1, Y_, dat, family, glmfit1)
    expFIM <- expFIM.glm(theta1, qG, disp, formula1, Y_, dat, family, glmfit1)

    res_SobsIM <- Testing_EM_joint(theta1, qG, NULL, G_[indx_thetas_H0], FIM)
    res_SexpIM <- Testing_EM_joint(theta1, qG, NULL, G_[indx_thetas_H0], expFIM)

  } else {
    FIM    <- FIM.glm(theta, qG, disp, formula, Y_, dat, family, glmfit)
    expFIM <- expFIM.glm(theta, qG, disp, formula, Y_, dat, family, glmfit)

    res_VobsIM <- Testing_EM_joint(theta, qG, NULL, G_, FIM)
    res_VexpIM <- Testing_EM_joint(theta, qG, NULL, G_, expFIM)
  }

  ll <- .loglik(theta, q, disp, formula, Y_, G_, Z_, dat, family)

  if (Ho) {
    out <- list(theta      = theta,
                qGZ        = q,
                dispersion = disp,
                Sobs       = res_SobsIM$S,
                Sexp       = res_SexpIM$S,
                df         = res_SexpIM$df,
                qG         = qG,
                iter       = as.integer(iter),
                ll         = ll,
                formula    = formula,
                miscov     = miscov,
                auxvar     = auxvar,
                family     = family,
                null.model = TRUE)
  } else {
    out <- list(theta      = theta,
                qGZ        = q,
                dispersion = disp,
                var_theta  = res_VobsIM$Var,
                Wobs       = res_VobsIM$W,
                Wexp       = res_VexpIM$W,
                df         = res_VexpIM$df,
                qG         = qG,
                iter       = as.integer(iter),
                ll         = ll,
                formula    = formula,
                miscov     = miscov,
                auxvar     = auxvar,
                family     = family,
                null.model = FALSE)
  }

  class(out) <- "twoPhaseSPML"
  out
}


#' Print method for \code{twoPhaseSPML} objects
#'
#' Prints a concise, \code{glm}-style summary of a fitted
#' \code{\link{twoPhaseSPML}} model.
#'
#' @param x an object of class \code{"twoPhaseSPML"}, as returned by
#'   \code{\link{twoPhaseSPML}}.
#' @param digits integer; minimum number of significant digits to print.
#'   Passed to \code{\link{printCoefmat}}. Defaults to
#'   \code{max(3L, getOption("digits") - 3L)}.
#' @param \dots further arguments passed to or from other methods (currently
#'   unused).
#'
#' @details
#' Under the \emph{null} model (\code{miscov} absent from \code{formula}),
#' score statistics (\code{Sobs} and \code{Sexp}) are displayed. Under the
#' \emph{alternative} model, Wald statistics (\code{Wobs} and \code{Wexp})
#' together with standard errors and \eqn{p}-values for the missing-by-design
#' covariate(s) are displayed.  The coefficient table always shows the
#' estimated regression coefficients; standard errors and \eqn{p}-values are
#' shown only when \code{var_theta} is available (alternative model).
#'
#' @return \code{x} is returned invisibly.
#'
#' @seealso \code{\link{twoPhaseSPML}}
#'
#' @examples
#' data.table::setDTthreads(1)
#' data <- DataGeneration_TPD(N=200)
#' set.seed(1)
#' R <- rep(0, nrow(data))
#' R[sample(nrow(data), 175)] <- 1
#' data[R == 0, c("G1")] <- NA
#'
#' res_Ho <- twoPhaseSPML(
#'   formula = Y ~ Z,
#'   miscov  = ~ G1,
#'   auxvar  = ~ Z,
#'   data   = data
#' )
#' print(res_Ho)
#'
#' @export
print.twoPhaseSPML <- function(x, digits = max(3L, getOption("digits") - 3L), ...)
{
  cat("\nCall:\n",
      paste0("twoPhaseSPML(formula = ", deparse(x$formula),
             ", miscov = ", deparse(x$miscov),
             ", auxvar = ", deparse(x$auxvar),
             ", family = ", x$family$family, ")"),
      "\n\n", sep = "")

  # ---- Coefficients table -------------------------
  theta <- x$theta
  G_    <- all.vars(x$miscov)

  if (!x$null.model && !anyNA(x$var_theta)) {
    # Alternative model: show Estimate / SE / z / p
    se    <- sqrt(x$var_theta)
    zval  <- theta / se
    pval  <- 2 * pnorm(abs(zval), lower.tail = FALSE)
    coef_mat <- cbind(Estimate    = theta,
                      `Std. Error` = se,
                      `z value`    = zval,
                      `Pr(>|z|)`   = pval)
    cat("Coefficients:\n")
    printCoefmat(coef_mat, digits = digits, signif.stars = TRUE,
                 na.print = "NA", ...)
  } else {
    # Null model or unavailable SE: show Estimate only
    coef_mat <- cbind(Estimate = theta)
    cat("Coefficients:\n")
    printCoefmat(coef_mat, digits = digits, na.print = "NA", ...)
  }

  # ---- Family / link --------------------
  cat("\nFamily:", x$family$family,
      " Link:", x$family$link, "\n")

  # ---- Dispersion -----------------------
  if (x$family$family %in% c("gaussian", "Gamma")) {
    cat("Dispersion:", format(x$dispersion, digits = digits), "\n")
  }

  # ---- Log-likelihood --------------------
  cat("Log-likelihood:", format(x$ll, digits = digits),
      " (EM iterations:", x$iter, ")\n")

  # ---- Test statistics -------------------
  cat("\nHypothesis test for missing-by-design covariate(s):",
      paste(G_, collapse = ", "), "\n")

  if (x$null.model) {
    cat("  [Null model - Score statistics]\n")
    cat("  Observed score statistic (Sobs):", format(x$Sobs, digits = digits), "\n")
    cat("  Expected score statistic (Sexp):", format(x$Sexp, digits = digits), "\n")
  } else {
    cat("  [Alternative model - Wald statistics]\n")
    cat("  Observed Wald statistic (Wobs):", format(x$Wobs, digits = digits), "\n")
    cat("  Expected Wald statistic (Wexp):", format(x$Wexp, digits = digits), "\n")
  }
  cat("  Degrees of freedom:", x$df, "\n")

  if (!is.null(x$df) && !is.na(x$df)) {
    stat <- if (x$null.model) x$Sobs else x$Wobs
    if (!is.null(stat) && !is.na(stat)) {
      pv <- pchisq(stat, df = x$df, lower.tail = FALSE)
      cat("  p-value (chi-squared):", format.pval(pv, digits = digits), "\n")
    }
  }

  cat("\n")
  invisible(x)
}


#' Auxiliary for Controlling the EM algorithm in twoPhaseSPML.
#'
#'
#' @param EMmaxit integer giving the maximal number of iterations of the EM algorithm.
#' @param EMtol numeric positive convergence tolerance for EM algorithm covnergence.
#' @param trace logical \code{TRUE}, prints progress at each EM iteration. Defaults to \code{FALSE}.
#'
#' @details
#' The function implements the semiparametric maximum likelihood (SPML)
#' 
#'
#'
#' @return A list with components named as the arguments.
#' 
#' @export
twoPhaseSPML.control <- function (EMmaxit = 1000, EMtol = 1e-05, trace = FALSE) 
{
  if (!is.numeric(EMtol) || EMtol <= 0) 
    stop("value of 'epsilon' must be > 0")
  if (!is.numeric(EMmaxit)) 
    stop("maximum number of iterations must be numeric")
  list(EMtol = EMtol, EMmaxit = EMmaxit, trace = trace)
}

## Tests for twoPhaseSPML() and print.twoPhaseSPML()
## -------------------------------------------------------
## These tests use a small, reproducible dataset generated by
## DataGeneration_TPD() so that they run quickly without network access.

# ---- Shared fixture --------------------------------------------------------

make_test_data <- function(seed = 1, n2 = 350, N = 500) {
  set.seed(seed)
  data <- DataGeneration_TPD(N = N, Beta1 = 0.5, LD.r = 0.75,
                              P_g = 0.20, P_z = 0.30)
  R <- rep(0L, nrow(data))
  R[sample(nrow(data), n2)] <- 1L
  data[R == 0L, c("G1")] <- NA
  list(
    data = data
  )
}

# ---- DataGeneration_TPD() --------------------------------------------------

test_that("DataGeneration_TPD returns a data frame with expected columns", {
  set.seed(1)
  d <- DataGeneration_TPD(N = 500)
  expect_s3_class(d, "data.frame")
  expect_true(all(c("Y", "G1", "Z", "S") %in% names(d)))
  expect_equal(nrow(d), 500L)
})

test_that("DataGeneration_TPD is reproducible with set.seed", {
  set.seed(99)
  d1 <- DataGeneration_TPD(N = 300)
  set.seed(99)
  d2 <- DataGeneration_TPD(N = 300)
  expect_equal(d1, d2)
})

# ---- twoPhaseSPML() — return value -----------------------------------------

test_that("twoPhaseSPML returns an object of class 'twoPhaseSPML'", {
  td <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data   = td$data)
  expect_s3_class(res, "twoPhaseSPML")
})

test_that("null model result has expected components", {
  td <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data   = td$data)
  expect_true(res$null.model)
  expect_named(res$theta, c("(Intercept)", "Z"))
  expect_true(!is.null(res$Sobs))
  expect_true(!is.null(res$Sexp))
  expect_null(res$Wobs)   # Wobs absent from null model
  expect_type(res$ll, "double")
  expect_true(is.numeric(res$iter) && res$iter >= 1L)
  expect_s3_class(res$qGZ, "data.frame")
  expect_true("G1" %in% names(res$qGZ))
  expect_true("q" %in% names(res$qGZ))
})

test_that("alternative model result has expected components", {
  td <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z + G1,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data   = td$data)
  expect_false(res$null.model)
  expect_named(res$theta, c("(Intercept)", "Z", "G1"), ignore.order = TRUE)
  expect_true(!is.null(res$Wobs))
  expect_true(!is.null(res$Wexp))
  expect_null(res$Sobs)   # Sobs absent from alternative model
  expect_type(res$var_theta, "double")
  expect_length(res$var_theta, length(res$theta))
})

test_that("stored formula and family are preserved", {
  td <- make_test_data()
  fm <- Y ~ Z + G1
  res <- twoPhaseSPML(formula = fm,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      family  = gaussian,
                      data   = td$data)
  expect_equal(deparse(res$formula), deparse(fm))
  expect_equal(res$family$family, "gaussian")
})

test_that("dispersion is positive for Gaussian family", {
  td <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z + G1,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data   = td$data)
  expect_gt(res$dispersion, 0)
})

test_that("log-likelihood is finite and negative", {
  td <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z + G1,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data   = td$data)
  expect_true(is.finite(res$ll))
  expect_lt(res$ll, 0)
})

test_that("df equals number of miscov variables", {
  td <- make_test_data()
  res_Ho <- twoPhaseSPML(formula = Y ~ Z,
                         miscov  = ~ G1,
                         auxvar  = ~ Z,
                         data   = td$data)
  res_Ha <- twoPhaseSPML(formula = Y ~ Z + G1,
                         miscov  = ~ G1,
                         auxvar  = ~ Z,
                         data  = td$data)
  expect_equal(res_Ho$df, 1L)
  expect_equal(res_Ha$df, 1L)
})

test_that("invalid family raises an error", {
  td <- make_test_data()
  expect_error(
    twoPhaseSPML(formula = Y ~ Z,
                 miscov  = ~ G1,
                 auxvar  = ~ Z,
                 family  = "not_a_family",
                 data = td$data),
    regexp = "not found|family"
  )
})

# ---- print.twoPhaseSPML() --------------------------------------------------

test_that("print.twoPhaseSPML returns x invisibly", {
  td  <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data = td$data)
  out <- withVisible(print(res))
  expect_false(out$visible)
  expect_identical(out$value, res)
})

test_that("print.twoPhaseSPML produces output for null model", {
  td  <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data = td$data)
  out <- capture.output(print(res))
  expect_true(any(grepl("Call", out)))
  expect_true(any(grepl("Coefficients", out)))
  expect_true(any(grepl("Sobs|score", out, ignore.case = TRUE)))
  expect_true(any(grepl("Log-likelihood", out)))
})

test_that("print.twoPhaseSPML produces output for alternative model", {
  td  <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z + G1,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data = td$data)
  out <- capture.output(print(res))
  expect_true(any(grepl("Wobs|Wald", out, ignore.case = TRUE)))
  expect_true(any(grepl("Std\\. Error", out)))
  expect_true(any(grepl("p-value", out, ignore.case = TRUE)))
})

test_that("print.twoPhaseSPML digits argument is respected", {
  td  <- make_test_data()
  res <- twoPhaseSPML(formula = Y ~ Z + G1,
                      miscov  = ~ G1,
                      auxvar  = ~ Z,
                      data = td$data)
  # Should not error with different digits values
  expect_no_error(capture.output(print(res, digits = 2L)))
  expect_no_error(capture.output(print(res, digits = 8L)))
})

# ---- start.values ----------------------------------------------------------

test_that("start.values$betas is accepted without error", {
  td <- make_test_data()
  # Fit null first, use its theta as start
  res0 <- twoPhaseSPML(formula = Y ~ Z,
                       miscov  = ~ G1,
                       auxvar  = ~ Z,
                       data = td$data)
  expect_no_error(
    twoPhaseSPML(formula      = Y ~ Z,
                 miscov       = ~ G1,
                 auxvar       = ~ Z,
                 data        = td$data,
                 start.values = list(betas = res0$theta))
  )
})

# ---- lrtest.twoPhaseSPML() -------------------------------------------------

test_that("lrtest returns correct class and structure", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expect_s3_class(lrt, "lrtest.twoPhaseSPML")
  expect_s3_class(lrt, "data.frame")
  expect_equal(nrow(lrt), 2L)
  expect_true(all(c("#Df", "LogLik", "Df", "Chisq", "Pr(>Chisq)") %in% names(lrt)))
})

test_that("lrtest model order is automatic (smaller model always first)", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)

  lrt_fwd <- lrtest(Ho, Ha)
  lrt_rev <- lrtest(Ha, Ho)  # reversed input order

  expect_equal(lrt_fwd$LogLik, lrt_rev$LogLik)
  expect_equal(lrt_fwd$Chisq,  lrt_rev$Chisq)
})

test_that("lrtest #Df column equals number of regression coefficients", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expect_equal(lrt[["#Df"]][1], length(Ho$theta))
  expect_equal(lrt[["#Df"]][2], length(Ha$theta))
})

test_that("lrtest Df change equals parameter count difference", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expect_equal(lrt[["Df"]][2], length(Ha$theta) - length(Ho$theta))
})

test_that("lrtest Chisq equals -2 * (ll0 - ll1)", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expected_chisq <- -2 * (Ho$ll - Ha$ll)
  expect_equal(lrt[["Chisq"]][2], expected_chisq)
})

test_that("lrtest p-value is in [0, 1]", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  pv <- lrt[["Pr(>Chisq)"]][2]
  expect_gte(pv, 0)
  expect_lte(pv, 1)
})

test_that("lrtest first row has NA for test columns", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expect_true(is.na(lrt[["Df"]][1]))
  expect_true(is.na(lrt[["Chisq"]][1]))
  expect_true(is.na(lrt[["Pr(>Chisq)"]][1]))
})

test_that("lrtest LogLik matches stored ll values", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  expect_equal(lrt[["LogLik"]][1], Ho$ll)
  expect_equal(lrt[["LogLik"]][2], Ha$ll)
})

test_that("print.lrtest.twoPhaseSPML returns x invisibly and produces output", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  lrt <- lrtest(Ho, Ha)

  vis <- withVisible(print(lrt))
  expect_false(vis$visible)
  expect_identical(vis$value, lrt)

  out <- capture.output(print(lrt))
  expect_true(any(grepl("Likelihood ratio", out, ignore.case = TRUE)))
  expect_true(any(grepl("Chisq",            out, ignore.case = TRUE)))
})

test_that("lrtest errors on non-twoPhaseSPML input", {
  td <- make_test_data()
  Ho <- twoPhaseSPML(Y~Z, ~G1, ~Z, data = td$data)
  expect_error(lrtest(Ho, list(ll=0, theta=1:2)), class = "simpleError")
  expect_error(lrtest(lm(1~1), Ho),               class = "simpleError")
})

test_that("lrtest warns on different families", {
  td  <- make_test_data()
  Ho  <- twoPhaseSPML(Y~Z,    ~G1, ~Z, data = td$data)
  Ha  <- twoPhaseSPML(Y~Z+G1, ~G1, ~Z, data = td$data)
  # Manually tamper with family to trigger the check
  Ha_tampered        <- Ha
  Ha_tampered$family <- binomial()
  expect_error(lrtest(Ho, Ha_tampered))
})

#' Likelihood ratio test for \code{twoPhaseSPML} models
#'
#' Compares two nested \code{\link{twoPhaseSPML}} models via a likelihood ratio
#' test (LRT), mirroring the interface of \code{lmtest::lrtest()}.
#'
#' @param object a fitted object of class \code{"twoPhaseSPML"} representing
#'   the \emph{smaller} (null / restricted) model.
#' @param object2 a fitted object of class \code{"twoPhaseSPML"} representing
#'   the \emph{larger} (alternative / unrestricted) model.  The two models
#'   must have been fitted to the same data and with the same \code{miscov},
#'   \code{auxvar}, and \code{family}.  The larger model must have at least
#'   as many parameters as the smaller one.
#' @param \dots further arguments, currently unused.
#'
#' @details
#' The test statistic is
#' \deqn{LRT = -2 \bigl(\ell_0 - \ell_1\bigr),}
#' where \eqn{\ell_0} and \eqn{\ell_1} are the maximised log-likelihoods of
#' the smaller and larger models, respectively.  Under the null hypothesis
#' that the additional parameters are zero, \eqn{LRT} follows a
#' \eqn{\chi^2} distribution with degrees of freedom equal to the difference
#' in the number of estimated regression coefficients between the two models.
#'
#' The function determines model order automatically: whichever of
#' \code{object} and \code{object2} has fewer regression coefficients is
#' treated as the null model, regardless of the order in which they are
#' supplied.
#'
#' @return An object of class \code{"lrtest.twoPhaseSPML"}, which is a
#'   \code{data.frame} with one row per model and the following columns:
#'   \describe{
#'     \item{\code{#Df}}{number of estimated regression coefficients.}
#'     \item{\code{LogLik}}{maximised log-likelihood.}
#'     \item{\code{Df}}{change in degrees of freedom relative to the
#'       previous row (\code{NA} for the first row).}
#'     \item{\code{Chisq}}{LRT statistic (\code{NA} for the first row).}
#'     \item{\code{Pr(>Chisq)}}{two-sided \eqn{p}-value from the
#'       \eqn{\chi^2} distribution (\code{NA} for the first row).}
#'   }
#'   The object also carries a \code{"heading"} attribute (a character vector)
#'   and is printed via \code{\link{print.lrtest.twoPhaseSPML}}.
#'
#' @seealso \code{\link{twoPhaseSPML}}, \code{\link{print.lrtest.twoPhaseSPML}}
#'
#' @examples
#' data.table::setDTthreads(1)
#' data <- DataGeneration_TPD(N=200)
#' set.seed(1)
#' R <- rep(0L, nrow(data))
#' R[sample(nrow(data), 175)] <- 1L
#' data[R == 0L, c("G1")] <- NA
#'
#' # Null model (no G1)
#' res_Ho <- twoPhaseSPML(
#'   formula = Y ~ Z,
#'   miscov  = ~ G1,
#'   auxvar  = ~ Z,
#'   data   = data
#' )
#'
#' # Alternative model (with G1)
#' res_Ha <- twoPhaseSPML(
#'   formula = Y ~ Z + G1,
#'   miscov  = ~ G1,
#'   auxvar  = ~ Z,
#'   data   = data
#' )
#'
#' lrtest(res_Ho, res_Ha)
#'
#' @export
#' @importFrom stats pchisq
lrtest <- function(object, object2, ...) {
  UseMethod("lrtest")
}


#' @rdname lrtest
#' @export
lrtest.twoPhaseSPML <- function(object, object2, ...) {

  # ---- Input validation ----------------------------------------------------
  if (!inherits(object, "twoPhaseSPML"))
    stop("'object' must be of class \"twoPhaseSPML\".")
  if (!inherits(object2, "twoPhaseSPML"))
    stop("'object2' must be of class \"twoPhaseSPML\".")

  if (object$family$family != object2$family$family ||
      object$family$link   != object2$family$link)
    stop("Both models must use the same family and link function.")

  if (deparse(object$miscov) != deparse(object2$miscov))
    warning("The two models have different 'miscov' formulas; ",
            "ensure they are truly nested.")

  # ---- Identify null vs alternative by number of regression parameters -----
  np1 <- length(object$theta)
  np2 <- length(object2$theta)

  if (np1 == np2)
    warning("Both models have the same number of parameters (", np1, "); ",
            "the LRT statistic may not be meaningful.")

  # Smaller model = null, larger model = alternative
  if (np1 <= np2) {
    mod0 <- object;  lab0 <- deparse(object$formula)
    mod1 <- object2; lab1 <- deparse(object2$formula)
  } else {
    mod0 <- object2; lab0 <- deparse(object2$formula)
    mod1 <- object;  lab1 <- deparse(object$formula)
  }

  ll0 <- mod0$ll;  npar0 <- length(mod0$theta)
  ll1 <- mod1$ll;  npar1 <- length(mod1$theta)

  # ---- Test statistic -------------------------------------------------------
  delta_df  <- npar1 - npar0
  lrt_stat  <- -2 * (ll0 - ll1)

  if (lrt_stat < 0) {
    warning("LRT statistic is negative (", round(lrt_stat, 4), "). ",
            "This can occur when the 'larger' model has a lower log-likelihood ",
            "than the 'smaller' model, which may indicate non-convergence or ",
            "numerical instability.")
  }

  pval <- pchisq(lrt_stat, df = delta_df, lower.tail = FALSE)

  # ---- Assemble output data frame ------------------------------------------
  out <- data.frame(
    `#Df`      = c(npar0,  npar1),
    LogLik     = c(ll0,    ll1),
    Df         = c(NA_real_,         as.numeric(delta_df)),
    Chisq      = c(NA_real_,         lrt_stat),
    `Pr(>Chisq)` = c(NA_real_,       pval),
    check.names = FALSE,
    stringsAsFactors = FALSE
  )
  rownames(out) <- c("Model 1", "Model 2")

  attr(out, "heading") <- c(
    "Likelihood ratio test\n",
    paste0("Model 1: ", lab0),
    paste0("Model 2: ", lab1)
  )
  class(out) <- c("lrtest.twoPhaseSPML", "data.frame")
  out
}


#' Print method for \code{lrtest.twoPhaseSPML} objects
#'
#' @param x an object of class \code{"lrtest.twoPhaseSPML"}, as returned by
#'   \code{\link{lrtest.twoPhaseSPML}}.
#' @param digits integer; number of significant digits. Defaults to
#'   \code{max(3L, getOption("digits") - 3L)}.
#' @param \dots further arguments passed to \code{\link{printCoefmat}}.
#'
#' @return \code{x} is returned invisibly.
#'
#' @seealso \code{\link{lrtest.twoPhaseSPML}}
#'
#' @export
print.lrtest.twoPhaseSPML <- function(x,
                                      digits = max(3L, getOption("digits") - 3L),
                                      ...) {
  # Print heading lines
  heading <- attr(x, "heading")
  if (!is.null(heading))
    cat(paste(heading, collapse = "\n"), "\n\n", sep = "")

  # Build a printable matrix
  mat <- as.matrix(x)
  # Format Chisq and p-value rows nicely
  out_mat <- format(x, digits = digits)

  # Use printCoefmat for the bottom row (contains the test result)
  printCoefmat(x,
               digits    = digits,
               na.print  = "",
               has.Pvalue = TRUE,
               ...)

  invisible(x)
}

check_alloc<-function(alloc_exp,Obs_samp){#Check for allocation higher than the observed elements in strata (if that exists keep all subjects in that strata and adjust to keep the sample size fixed)
  while( any(alloc_exp>Obs_samp) ){
    alloc_init<-alloc_exp
    indx1= alloc_exp>Obs_samp
    to_allocate<-sum( alloc_exp[indx1]-Obs_samp[indx1] )
    alloc_exp[indx1]<-Obs_samp[indx1]
    indx2<-!alloc_init>=Obs_samp & alloc_init>0
    alloc_exp[indx2]<- alloc_init[indx2] + floor(to_allocate*alloc_init[indx2]/sum(alloc_init[indx2] ))
  }
  alloc_exp
}


BSS <- function(samp.fracs,n,data,stratf)  {


  ### samp.fracs: vector of sampling fractions with length equal to the number of strata determined by stratf (and in the same order). Must add up to 1.
  ### n: phase 2 sample size
  ### data: phase 1 data, MUST contain stratification factors
  ### stratf: right hand side of a formula with stratification factors (Z and Yst in this case)


  tot_df=data.frame(xtabs(stratf,data=data))
  tot_df$n_strat=round(n*samp.fracs)


  # xtabs(n_stat~Z+S,data=tot_df)
  # xtabs(Freq~Z+S,data=tot_df)


  tot_df$n_strat<-check_alloc(tot_df$n_strat,tot_df$Freq)

  if( any(tot_df$n_strat>tot_df$Freq) ) stop("Sampling fraction greater than 1 for some stratum in the allocation")

  if( sum(round(n*samp.fracs))-sum(tot_df$n_strat)>2 )  {


    to_allocate<-sum(round(n*samp.fracs))-sum(tot_df$n_strat)
    alloc_init<-tot_df$n_strat
    indx <- tot_df$n_strat/tot_df$Freq<1 & !is.na(tot_df$n_strat/tot_df$Freq)
    q<-alloc_init[indx]/sum(alloc_init[indx])
    tot_df$n_strat[indx] <- alloc_init[indx] + ceiling(to_allocate*ifelse(is.nan(q),1/length(alloc_init[indx]),q))


  }


  tot_df$n_strat<-check_alloc(tot_df$n_strat,tot_df$Freq)

  ### Sampling begins
  #p2 <- tot_df$n_strat/tot_df$Freq
  data$R_ <- rep(0, NROW(data))
  stratf_<-all.vars(stratf)
  data$ord_ <-1:NROW(data)
  tot_df$k_<-1:NROW(tot_df)

  df=merge(data,tot_df[,c(stratf_,"k_")],by=stratf_,all.x=TRUE,sort=FALSE)

  for(k in 1:NROW(tot_df)){ #k<-3
    n_k=tot_df[k,"n_strat"]
    s <- sample(which( df$k_==k ), n_k)
    df$R_[s] <- 1
  }

  list(R=df[order(df$ord_),"R_"],eff.samp.fracs=tot_df$n_strat/n)


}


.optMsure<-function(type){
  Dopt<-function(vcov,k){
    if( is.null(k) ){
      return(det(vcov))
    }else{
      return(det(vcov[k,k]))
    }
  }
  Aopt<-function(vcov,k){
    if( is.null(k) ){
      return(sum(diag(vcov)))
    }else{
      return(sum(diag(vcov[k,k])))
    }
  }
  Param<-function(vcov,k){
    vcov[k,k]
  }
  switch(type,'D-opt'=Dopt,'A-opt'=Aopt,"Par-spec"=Param)
}


FIMf <- function(formula,family,dat,beta,p_gz,disp,G_,Z_){
  X <- model.matrix(formula, dat)
  eta = as.vector(if (NCOL(X) == 1L){ X * beta }else X %*% beta)
  mu = family$linkinv(eta)
  ## other quantities of interest
  varmu <- family$variance(mu)
  mu.eta.val <- family$mu.eta(eta)
  if (any(is.na(varmu))) #anyNA(varmu)
    stop("NAs in V(mu)")
  if (any(varmu == 0))
    stop("0s in V(mu)")
  if (any(is.na(mu.eta.val)))
    stop("NAs in d(mu)/d(eta)")
  w <- (mu.eta.val^2)/varmu
  ## Use explicit W_fisher to separate EM weights from GLM curvature (matches FIM.glm)
  mu_fit   <- family$linkinv(as.vector(X %*% beta))
  W_fisher <- dat$wg_ * (family$mu.eta(family$linkfun(mu_fit))^2 /
                         family$variance(mu_fit))

  #Create objects needed in the testing part
  # qG <- aggregate(as.formula(paste0("q~",paste(G_,collapse="+"))),p_gz,FUN=sum); names(qG)[which(names(qG)=="q")]<-"qG"
  qG <- p_gz; names(qG)[which(names(qG)=="q")]<-"qG"

  qG$k_ <- seq_len(nrow(qG))
  dat$ord_ <- seq_len(nrow(dat))

  # dat <- as.data.frame(merge(as.data.table(dat),as.data.table(qG[,c(G_,"qG","k_")]),by=G_, sort = FALSE)) # merge(dat,qG[,c(G_,"qG","k_")],by=G_) ## this one may be sped up
  dat <- as.data.frame(merge(as.data.table(dat),as.data.table(qG[,c(G_,Z_,"qG","k_")]),by=c(G_,Z_), sort = FALSE)) # merge(dat,qG[,c(G_,"qG","k_")],by=G_) ## this one may be sped up

  dat <- dat[order(dat$ord_),]

  Qd1=1/dat$qG
  Qd2=-1/dat$qG^2

  # Qmat <- crossprod(sapply(dat$k_,function(x)1*x==qG$k_), diag(nrow(qG))) ## Alternative way
  # Qmat.d1 <- Qmat*Qd1
  Qmat <- as.matrix(t(fac2sparse(dat$k_))) #class.ind2(dat$k_) ## old way
  Qmat.d1 = Qmat*Qd1
  Qmat.d2 = Qmat*Qd2

  # Qd1 <- 1/qG$qG
  # Qd2 <- -1/qG$qG^2

  # colsmatch <- c(G_,Z_)[c(G_,Z_) %in% names(dat)]
  # sapply(dat[,colsmatch],function(z){qG[,colsmatch] %in% z})

  # Ind <- sapply(dat[,Z_],function(z){qG[,Z_] %in% z})
  #
  # Qmat.d1 <- tcrossprod(model.matrix(~1,dat),Qd1)*t(Ind)
  # colnames(Qmat.d1) <- seq_len(nrow(qG))
  # Qmat.d2 <- tcrossprod(model.matrix(~1,dat),Qd2)*t(Ind)
  # colnames(Qmat.d2) <- 1:nrow(qG)

  Omega=dat$wg_

  ### Expected information
  {
    ## first matrix, or second derivatives (i.e. hessian of log P(Y_i|g_j) + log q_j or jacobian of the gradient)
    I1_0 <- crossprod(X, W_fisher * X) / disp # upper left matrix with the Fisher information for betas
    I2_0 <- colSums(Omega * Qmat.d2)
    I2_0 <- diag(I2_0,NROW(I2_0))

    expI1 <- rbind(cbind(I1_0, matrix(0, nrow=nrow(I1_0), ncol=ncol(I2_0))),
                 cbind(matrix(0, nrow=nrow(I2_0), ncol=ncol(I1_0)), -I2_0))


    ## calculate second matrix, i.e. expected value of cross score functions
    X = X[,,drop=FALSE]
     # w * (Y - mu)/(disp*mu.eta.val)
    Wx2 <- w^2*varmu/(disp*mu.eta.val^2)
    Mx2<- crossprod( X,  Omega*Wx2*X )

    Wxq2 <- 0
    Mxq2 <- crossprod( X,  Omega*Wxq2*Qmat.d1)

    Mq2 <- crossprod( Qmat.d1,  Omega*Qmat.d1 )

    expI2 <- rbind(cbind(Mx2, Mxq2),
                 cbind(t(Mxq2),Mq2))


    ## Third matrix, i.e. outer product of expected score. Note this uses the W's from the second matrix and weight them accordingly
    IIt <- crossprod(fac2sparse(dat$id_)) #tcrossprod(class.ind2.Mat(dat$id_)) #IIt<-tcrossprod(class.ind2(dat$id_)) ## much slower, specially for larga matrices

    Mx3<- as.matrix(crossprod( sign(Wx2)*sqrt(abs(Wx2))*X*Omega, IIt %*% (sqrt(abs(Wx2))*X*Omega) ))
    Mxq3<-as.matrix(crossprod( sign(Wx2)*sqrt(abs(Wxq2))*X*Omega, IIt %*% (sqrt(abs(Wxq2))*Qmat.d1*Omega) ))
    Mq3<- as.matrix(crossprod( Qmat.d1*Omega, IIt %*% (Qmat.d1*Omega) ))

    expI3<-as.matrix(rbind(cbind(Mx3, Mxq3),
                           cbind(t(Mxq3),Mq3)))

    expFIM=expI1-expI2+expI3
    return(list(FIM=expFIM,expI1=expI1,expI2=expI2,expI3=expI3))
  }
}

# another version for obsIMf1 that calculates the observed information by individual faster that
# calling obsIMf1 each time (needed for optimTP.GA), note that by.id=FALSE should render the
# exact same results as obsIMf1
obsIMf1 <- function(formula,family,dat,beta,p_gz,disp,G_,Z_,by.id=FALSE){
  ## Remove cells with zero or negative probability (1/q would be infinite)
  p_gz <- p_gz[p_gz$q > 0, , drop = FALSE]
  Y <- dat[,as.character(formula[[2]])]
  X <- model.matrix(formula, dat)
  eta = as.vector(if (NCOL(X) == 1L){ X * beta }else X %*% beta)
  mu = family$linkinv(eta)
  ## other quantities of interest
  varmu <- family$variance(mu)
  mu.eta.val <- family$mu.eta(eta)

  if (any(is.na(varmu))) #anyNA(varmu)
    stop("NAs in V(mu)")
  if (any(varmu == 0))
    stop("0s in V(mu)")
  if (any(is.na(mu.eta.val)))
    stop("NAs in d(mu)/d(eta)")
  Omega <- dat$wg_
  w <- (mu.eta.val^2)/varmu

  # qG <- aggregate(as.formula(paste0("q~",paste(G_,collapse="+"))),p_gz,FUN=sum); names(qG)[which(names(qG)=="q")]<-"qG"
  qG <- p_gz; names(qG)[which(names(qG)=="q")]<-"qG"

  qG$k_ <- seq_len(nrow(qG))

  # dat$ord_ <- seq_len(nrow(dat))
  # # dat <- as.data.frame(merge(as.data.table(dat),as.data.table(qG[,c(G_,"qG","k_")]),by=G_, sort = FALSE)) # merge(dat,qG[,c(G_,"qG","k_")],by=G_) ## this one may be sped up
  # dat <- as.data.frame(merge(as.data.table(dat),as.data.table(qG[,c(G_,Z_,"qG","k_")]),by=c(G_,Z_), sort = FALSE)) # merge(dat,qG[,c(G_,"qG","k_")],by=G_) ## this one may be sped up
  # dat <- dat[order(dat$ord_),]
  # ## Need the 3 lines below to always have the same number of columns in Qmat irrespective of the subset
  # Qd1 <- 1/dat$qG
  # Qmat0 <- crossprod(sapply(dat$k_,function(x)1*x==qG$k_), diag(nrow(qG)))
  # Qmat.d1bis <- Qmat0*Qd1

  Qmat <- tcrossprod(matrix(rep(1,nrow(dat)),ncol=1),1/qG$qG)
  matchrows <- if( all(c(G_,Z_)%in%names(dat)) & all(c(G_,Z_)%in%names(qG)) ){
    match(interaction(dat[,c(G_,Z_)]), interaction(qG[,c(G_,Z_)]))
  }else if( Z_ %in% names(qG) & Z_ %in% names(dat) ){
    match(dat[,c(Z_)],qG[,c(Z_)])
  }else if( G_ %in% names(qG) & G_ %in% names(dat)){
    match(dat[,c(G_)],qG[,c(G_)])
  }else stop("No matching columns b/t p_gz and dat")

  Qmat.d1 <- Qmat*t(sapply(matchrows,function(s){seq_len(nrow(qG)) == s}))
  # Qmat.d1b <- crossprod(sapply(dat[,Z_],function(z){1*(p_gz[,Z_]==z)/p_gz[,"q"]}), diag(nrow(p_gz)))
  colnames(Qmat.d1) <- seq_len(nrow(qG))

  ## Original way assuming all the values of qG are in dat$k_
  # Qmat <- as.matrix(t(fac2sparse(dat$k_)))   #class.ind2(dat$k_)
  # Qmat.d1 <- Qmat*Qd1

  X <- X[, , drop = FALSE]

  Wxq2 <- w * (Y - mu)/(disp*mu.eta.val)

  U <- cbind(Wxq2*X,Qmat.d1)

  Ind_id <- fac2sparse(dat$id_)
  EU <- Ind_id %*% (Omega*U)

  ## ---- I1: Hessian block (matches FIM.glm) --------------------------------
  ## Beta block: use explicit W_fisher to separate EM weights from GLM curvature
  mu_fit   <- family$linkinv(as.vector(X %*% beta))
  W_fisher <- Omega * (family$mu.eta(family$linkfun(mu_fit))^2 /
                       family$variance(mu_fit))
  I1_beta  <- crossprod(X, W_fisher * X) / disp

  ## q block: Fisher information for q_j = negative expected Hessian
  ##   = -E[d^2 logL / dq_j^2] = +sum_{i: cell_i=j} Omega_i / q_j^2
  ##   Qmat.d1[i,j] = (1/q_j) * 1[cell_i = j], so
  ##   (1/q_j) * Qmat.d1[i,j] = (1/q_j^2) * 1[cell_i=j]  (positive)
  nq <- ncol(Qmat.d1)
  I1_q <- diag(colSums(sweep(Qmat.d1, 2, 1/qG$qG, "*") * Omega), nrow=nq)

  I1 <- rbind(cbind(I1_beta, matrix(0, nrow=nrow(I1_beta), ncol=nq)),
              cbind(matrix(0, nrow=nq, ncol=ncol(I1_beta)), I1_q))

  if( !by.id ){
    ## Second matrix (Louis I2): complete-data score outer product
    obsI2 <- crossprod( sqrt(Omega)*U,  sqrt(Omega)*U )

    ## Third matrix (Louis I3): outer product of per-individual observed-data scores
    obsI3 <- as.matrix(crossprod( EU,  EU ))

    ## Full Louis observed FIM: I1 - I2 + I3
    obsFIM <- I1 - obsI2 + obsI3

  }else{
    ### Row-wise outer product helper
    rwouter <- function(mat){t(apply( mat, 1, function(x) x%o%x))}
    ## Per-individual I2 (one row per individual, vectorised npars^2)
    obsI2 <- as.matrix( Ind_id %*% rwouter( U*sqrt(Omega) ) )
    ## Per-individual I3 (outer product of per-individual score sums)
    obsI3 <- rwouter( EU )
    ## I1 is the aggregate Hessian (same across all individuals)
    ## Return it once; the design optimizer uses the full-cohort FIM from by.id=FALSE
    return(list(I1=I1, I2=obsI2, I3=obsI3))
  }

  return(list(I1=I1, I2=obsI2, I3=obsI3, FIM=obsFIM))
}

obsIMf2 <- function(formula,family,dat,beta,p_gz,disp,G_,Z_,by.id=FALSE){
  #### use enrich package to do this for now
  linkf <- enrich(make.link(family$link), with = "all")
  family <- enrich(family, with = "all")

  Y <- dat[,as.character(formula[[2]])]
  X <- model.matrix(formula, dat)
  eta = as.vector(if (NCOL(X) == 1L){ X * beta }else X %*% beta)
  mu = family$linkinv(eta)
  ## other quantities of interest
  varmu <- family$variance(mu)
  mu.eta.val <- family$mu.eta(eta)
  d2mus <- linkf$d2mu.deta(eta)
  d1variances <- family$d1variance(mu)

  if (any(is.na(varmu))) #anyNA(varmu)
    stop("NAs in V(mu)")
  if (any(varmu == 0))
    stop("0s in V(mu)")
  if (any(is.na(mu.eta.val)))
    stop("NAs in d(mu)/d(eta)")
  Omega <- dat$wg_
  w <- (mu.eta.val^2)/varmu
  qrX <- qr(X*as.numeric(sqrt(Omega * w)))
  w1 <- Omega * (d2mus / varmu - mu.eta.val^2 * d1variances / varmu^2) * (Y - mu)

  # qG <- aggregate(as.formula(paste0("q~",paste(G_,collapse="+"))),p_gz,FUN=sum); names(qG)[which(names(qG)=="q")]<-"qG"
  qG <- p_gz; names(qG)[which(names(qG)=="q")]<-"qG"

  ###
  # Qd1 <- 1/qG$qG
  # Qmat.d1 <- tcrossprod(matrix(rep(1,nrow(dat)),ncol=1),Qd1) #wrong since doesn not discriminate by Z, all values are present

  # Qmat <- as.matrix(t(fac2sparse(dat$k_)))#class.ind2(dat$k_)
  # Qmat <- crossprod(sapply(dat$k_,function(x)1*x==qG$k_), diag(nrow(qG)))
  # Qmat.d1 <- Qmat*Qd1
  # Qmat.d1 <- tcrossprod(model.matrix(~1,dat),Qd1)
  # Qmat <- tcrossprod(matrix(rep(1,nrow(dat)),ncol=1),1/qG$qG)
  # matchrows <- if( all(c(G_,Z_)%in%names(dat)) & all(c(G_,Z_)%in%names(qG)) ){
  #   match(interaction(dat[,c(G_,Z_)]), interaction(qG[,c(G_,Z_)]))
  # }else if( Z_ %in% names(qG) & Z_ %in% names(dat) ){
  #   match(dat[,c(Z_)],qG[,c(Z_)])
  # }else if( G_ %in% names(qG) & G_ %in% names(dat)){
  #   match(dat[,c(G_)],qG[,c(G_)])
  # }else stop("No matching columns b/t p_gz and dat")
  # Qmat.d1 <- Qmat*t(sapply(matchrows,function(s){seq_len(nrow(qG)) == s}))

  Qmat.d1 <- crossprod(sapply(dat[,Z_],function(z){1*(p_gz[,Z_]==z)/p_gz[,"q"]}), diag(nrow(p_gz)))

  colnames(Qmat.d1) <- seq_len(nrow(qG))
  if( !by.id ){
    ## calculate second matrix, i.e. expected value of cross score functions
    X <- X[, , drop = FALSE]

    Wxq2 <- w * (Y - mu)/(disp*mu.eta.val)

    # U <- cbind(Wxq2*X,Qmat.d1)
    #
    # obsI2 <- crossprod( U,  Omega*U )

    Mx2 <- crossprod( Wxq2*X,  Omega*Wxq2*X )
    Mxq2 <- crossprod( X,  Omega*Wxq2*Qmat.d1)

    Mq2 <- crossprod( Qmat.d1,  Omega*Qmat.d1 )
    Mq2 <- diag(diag(Mq2),nrow=nrow(Mq2))

    obsI2 <- rbind(cbind(Mx2, Mxq2),
                   cbind(t(Mxq2),Mq2))

    ## Third matrix, i.e. outer product of expected score.
    IIt <- crossprod(fac2sparse(dat$id_)) #tcrossprod(class.ind2.Mat(dat$id_)) #IIt<-tcrossprod(class.ind2(dat$id_)) ## much slower, specially for large matrices
    signWxq2 <- sign(Wxq2)
    sqrtWxq2 <- sqrt(abs(Wxq2))

    Mx3 <- as.matrix(crossprod( Wxq2*X*Omega, IIt %*% (Wxq2*X*Omega) ))
    Mxq3 <- as.matrix(crossprod( signWxq2*sqrtWxq2*X*Omega, IIt %*% (sqrtWxq2*Qmat.d1*Omega) ))
    Mq3 <- as.matrix(crossprod( Qmat.d1*Omega, IIt %*% (Qmat.d1*Omega) ))

    obsI3 <- as.matrix(rbind(cbind(Mx3, Mxq3),
                             cbind(t(Mxq3),Mq3)))

    # obsI3 <- as.matrix(crossprod( U*Omega, IIt %*% U*Omega) ) ## not symmetric

    return(list(I2=obsI2,I3=obsI3))
  }else{

    X <- X[, , drop = FALSE]

    ### these functions implement row-wise outer product
    rwouter <- function(mat){t(apply( mat, 1, function(x) x%o%x))}
    rwouterxy <- function(mat1,mat2){nx=NCOL(mat1);ny=NCOL(mat2);t(apply( cbind(mat1,mat2), 1, function(x) x[1:nx]%o%x[(nx+1):(nx+ny)]))}

    ## calculate second matrix, i.e. expected value of cross score functions
    Wxq2 <- w * (Y - mu)/(disp*mu.eta.val)
    X_ <- cbind(Wxq2*X*sqrt(Omega),Qmat.d1*sqrt(Omega))
    obsI2 <- do.call(rbind,by(rwouter( X_),list(id_=dat$id_),colSums))

    ## Third matrix, i.e. outer product of expected score. Note this uses the W's from the second matrix and weight them accordingly
    IIt <- crossprod(fac2sparse(dat$id_))

    X_ <- cbind(Wxq2*X*Omega,Qmat.d1*Omega)

    obsI3 <- do.call(rbind,by(rwouterxy(  X_, as.matrix(IIt %*%  X_)),list(id_=dat$id_),colSums))

    return(list(I2=obsI2,I3=obsI3))
  }
}

Dmu_fnc <- function(formula,family,dat,beta,disp){
  #### use enrich package to do this for now
  Y <- dat[,as.character(formula[[2]])]
  X <- model.matrix(formula, dat)
  eta = as.vector(if (NCOL(X) == 1L){ X * beta }else X %*% beta)
  mu = family$linkinv(eta)
  ## other quantities of interest
  varmu <- family$variance(mu)
  mu.eta.val <- family$mu.eta(eta)

  if (any(is.na(varmu))) #anyNA(varmu)
    stop("NAs in V(mu)")
  if (any(varmu == 0))
    stop("0s in V(mu)")
  if (any(is.na(mu.eta.val)))
    stop("NAs in d(mu)/d(eta)")

  w <- (mu.eta.val^2)/varmu

  Dmu <- w * (Y - mu)/(disp*mu.eta.val)

  return(Dmu)
}

weigthedloglik <-  function(theta,nbetas,q,disp,formula,Y_,G_,Z_,dat,family)  {
  betas <- theta[1:nbetas]
  q$qG_ <- theta[(nbetas+1):length(theta)]
  datbis <- merge(dat,q,by=c(G_))
  In <- t(fac2sparse(datbis$id_))#class.ind2.Mat(datbis$id_)
  mu <- family$linkinv(model.matrix(formula,datbis)%*%betas)

  if(family$family=="gaussian"){
    sig<-sqrt(disp)
    ll <- sum(  datbis$wg_*log( colSums( dnorm(datbis[,Y_], mu, sig)*datbis$qG_*In ) ) )

  }else if(family$family=="binomial"){
    ll <- sum( datbis$wg_*log( colSums( dbinom(datbis[,Y_],size=1,prob=mu)*datbis$qG_*In ) ) )
  }else stop("family not one of: gaussian, binomial, Gamma, inverse.gaussian or poisson.")

  return(ll)
}

# dat0=dat; q0=p_gz
# dat0=dat_; q0=p_gz
omega1_g <- function(formula,family,dat0,beta,q0,disp,G_,Z_){

  N <- wg0 <- id_ <- lwg0 <- NULL # as suggested by https://www.r-bloggers.com/2019/08/no-visible-binding-for-global-variable/

  resp<-as.character(formula[[2]]); terms_q<-c(G_,Z_)
  dat0$ord <- seq_len(nrow(dat0))
  q_ <- names(q0)[!names(q0) %in% terms_q]
  if( q_ %in% names(dat0) ){
    namsold<-names(dat0)[which(names(dat0)==q_)]
    names(dat0)[which(names(dat0)==q_)]<-paste0(q_,".old")
  }
  dat0bis<-merge(as.data.table(dat0),as.data.table(q0[,c(terms_q,"q")]),by=terms_q,all.x=TRUE,sort = FALSE)

  X0 = model.matrix(formula,dat0bis)
  eta = as.vector(if (NCOL(X0) == 1L){ X0 * beta }else X0 %*% beta)
  mu <-family$linkinv(eta)

  if(family$family=="gaussian"){
    dat0bis$lwg0 <- dnorm(x=dat0bis[[resp]],mean=mu,sd=sqrt(disp), log=TRUE) + log(dat0bis[["q"]])
  }else if(family$family=="binomial"){
    dat0bis$lwg0 <- dbinom(x=dat0bis[[resp]],size=1,prob=mu, log=TRUE) + log(dat0bis[["q"]])
  }else if(family$family=="poisson"){
    dat0bis$lwg0 <- dpois(x=dat0bis[[resp]],lambda=mu, log=TRUE) + log(dat0bis[["q"]])
  }else if(family$family=="Gamma"){
    dat0bis$lwg0<-dgamma(x=dat0bis[[resp]],shape=1/disp,scale=mu*disp, log=TRUE) + log(dat0bis[["q"]])
  }else stop("family not one of: gaussian, binomial, Gamma or poisson.")

  dat0bis$wg0 <- exp(dat0bis$lwg0)

  dat0bis[, N:=sum(wg0,na.rm = TRUE), by = id_] ## fast way to add a column (N) with the sum of the weigths by id_

  ##    ## Procedure to normalize probabilities under small values taken from http://stats.stackexchange.com/questions/66616/converting-normalizing-very-small-likelihood-values-to-probability
  if( any(dat0bis$N==0) ){
    zerindx <- which(dat0bis$N==0)
    dfwg0 <- dat0bis[zerindx,c("id_","ord","lwg0"), with=FALSE]
    dfwg0[, max:=max(lwg0),by=id_]
    if( !all(dfwg0$ord==dat0bis$ord[zerindx]) ) stop("Something went wrong in the probability normalization process.")
    dat0bis$wg0[zerindx] <- exp(dfwg0$lwg0-dfwg0$max)
    dat0bis[zerindx, N:=sum(wg0), by = id_]
  }

  dat0bis$wg <- dat0bis$wg0/dat0bis$N
  dat0bis <- dat0bis[order(dat0bis$ord),]

  return(list(wg=dat0bis$wg,pYZ=dat0bis$N))
  #return(dat0bis$wg)
}

# function fitnessTP
fitnessTP  <-  function(obsIM.R,Rj,optimMeasure,K.idx=NULL)  {

  FIM <- obsIM.R$IM1 + matrix(colSums(obsIM.R$IM2_id*Rj),ncol=obsIM.R$npars) - matrix(colSums(obsIM.R$IM3_id*Rj),ncol=obsIM.R$npars)

  D <- rbind(diag(obsIM.R$npars-1),0)
  F1 <-  crossprod(D, FIM %*% D)

  ### Function that determines the optimality measure (D-, A- or parameter-specfic for now)
  ### if optimMeasure=="Par-spec" then K.idx MUST be not null
  if( optimMeasure=="Par-spec" & is.null(K.idx) ) stop("For a parameter-specific criterion K.idx must be provided.")

  ## asign these functions to the execution environment
  optMsuref=.optMsure(optimMeasure)

  if( obsIM.R$Null ){
    Ind <- obsIM.R$Ind

    invF1_noInd <- ginv(F1[!Ind,!Ind])
    V1 <- F1[Ind,Ind] - F1[Ind,!Ind] %*% invF1_noInd %*% F1[!Ind,Ind]
    invV1 <- solve( V1 )

    return( optMsuref(invV1,K.idx) )

  }else {
    vcov = solve(F1)

    return(optMsuref(vcov,K.idx))
  }

}

# function obsIM.R
obsIM.R  <-  function(formula,miscov,auxvar,family,data,beta,p_gz,disp=NULL)  {
  ### Indicators (0,1) for phase 2 belonging, same length as nrow(data)
  if (is.character(family))
    family <- get(family, mode = "function", envir = parent.frame())
  if (is.function(family))
    family <- family()
  if (is.null(family$family)) {
    message(deparse(family))
    stop("'family' not recognized")
  }
  terms_dat <- unique( c(all.vars(formula), attr(terms(miscov), "term.labels") , attr(terms(auxvar), "term.labels")) )
  Y_<-as.character(formula[[2]])
  G_<-all.vars(miscov)
  Z_<-all.vars(auxvar)

  ## remove enties with zero or negative probabilities
  p_gz <- p_gz[p_gz$q>0,]

  ### check if names in G_ and Z_ are in the joint ditribution data.frame p_gz
  g.vals<-unique(p_gz[,G_]);
  N <- NROW(data)

  if( is.null(disp) ){
    if( family$family %in% c("gaussian","Gamma") ){
      disp = var(data[,Y_])
    }else disp=1
  }

  dat_ <- cbind(id_=rep(seq(1,N,by=1), each=NROW(g.vals)),data[rep(seq_len(N), each=NROW(g.vals)), ],matrix(rep(t(g.vals),N), ncol=NCOL(g.vals), byrow=TRUE))
  names(dat_)[(NCOL(dat_)-NCOL(g.vals)+1):ncol(dat_)] <- G_
  dat_ <- droplevels(dat_)

  wgs <- omega1_g(formula,family,dat_,beta,p_gz,disp,G_,Z_)

  dat_$wg_ <- wgs$wg
  dat_ <- dat_[!is.na(dat_$wg_),]

  ## Baseline FIM: I3/N (empirical score outer product; valid large-N approximation)
  obsIM_all <- obsIMf1(formula,family,dat_,beta,p_gz,disp,G_,Z_,by.id = FALSE)
  IM1       <- obsIM_all$I3 / N

  ## Per-individual I2 and I3 for phase-2 improvement
  IM23_id <- obsIMf1(formula,family,dat_,beta,p_gz,disp,G_,Z_,by.id = TRUE)

  IM2_id <- IM23_id$I2 / N
  IM3_id <- IM23_id$I3 / N

  npars <- NCOL(IM1)

  Ind <- rep(FALSE,npars-1)
  Betas_ids <- which(grepl(paste0(paste0("^",G_),collapse="|"),colnames(model.matrix(formula,dat_[1,]))))
  Ind[Betas_ids] <- TRUE

  Null <- all(beta[Ind]==0)

  return( list(IM1=IM1,IM2_id=IM2_id,IM3_id=IM3_id,npars=npars, Ind=Ind, Null=Null) )
}


## p_Z <- data.frame(xtabs(~Z,dat_sim)/nrow(dat_sim)); LD.r <- 0.75; maf_G <- 0.2
### This function calculates p_gz given a distribution for Z (using, say xtabs), a MAF for G and a LD value (R).

# This function calculates p_gz given a distribution for Z (using, say xtabs), a MAF for G and a LD value (R).
# INTERNAL FUNCTION
p_gz_func <- function(p_Z, maf_G, LD.r, G_, Z_)  {
  ### I'm gonna assume here that G has 3 level genotypes (GG, Gg and gg), I will look further for generalizations later. Also, this only takes into account a single G. Further extensions for multiple G's are warranted.
  nZ <- NROW(p_Z)
  P_d <- maf_G; ## MAF of G
  P_a <- if(nZ==1){
    p_Z
  } else ifelse(nZ==3, p_Z[p_Z[,1]==2,"Freq"] + p_Z[p_Z[,1]==1,"Freq"]/2, p_Z[p_Z[,1]==1,"Freq"]) ## MAF of Z
  P_D <- 1 - P_d; P_A <- 1 - P_a ;

  Rsq <- LD.r^2
  LD <- LD.r * sqrt(P_D*P_d*P_A*P_a)
  # Dpr <- LD /( (LD<0)*min(P_D*P_A,P_d*P_a) + (LD>=0)*min(P_d*P_A,P_D*P_a) )
  #
  # if( Dpr > 1 | Dpr < -1 ) stop("This combination of MAF for G and genotype for Z seems incompatible. D' > 1 or D'< -1.")
  Dpr <- LD /( (LD<0)*max(-P_D*P_A,-P_d*P_a) + (LD>=0)*min(P_D*P_a,P_d*P_A) )
  if( Dpr > 1 | Dpr < 0 ) stop("This combination of MAF for G and genotype for Z seems incompatible. D' > 1 or D'< 0.")

  ## Haplotype frequencies (two SNPs)
  h.freqs <- rep(0, 4)
  h.freqs[1] <- LD + (1-P_d)*(1-P_a) ## DA
  h.freqs[2] <- 1 - P_d - h.freqs[1] ## Da
  h.freqs[3] <- 1 - P_a - h.freqs[1] ## dA
  h.freqs[4] <- LD + P_d*P_a ##  da
  names(h.freqs) <- c()

  ## Genotype frequencies (two SNPs)
  cross <- h.freqs %*% t(h.freqs)
  genot <- matrix(rep(0,3*3), nrow=3, ncol=3)
  genot[1,1] <- cross[1,1]
  genot[1,2] <- cross[1,2] + cross[2,1]
  genot[1,3] <- cross[2,2]
  genot[2,1] <- cross[1,3] + cross[3,1]
  genot[2,2] <- cross[1,4] + cross[2,3] + cross[3,2] + cross[4,1]
  genot[2,3] <- cross[2,4] + cross[4,2]
  genot[3,1] <- cross[3,3]
  genot[3,2] <- cross[3,4] + cross[4,3]
  genot[3,3] <- cross[4,4]
  rownames(genot) <- 0:2
  colnames(genot) <- 0:2
  crosstable <- data.frame(genot,check.names = FALSE)
  p_gz <- data.frame(rows = rownames(crosstable), stack(crosstable))
  names(p_gz)[names(p_gz)=="rows"] <- G_
  names(p_gz)[names(p_gz)=="ind"] <- Z_
  names(p_gz)[names(p_gz)=="values"] <- "q"
  p_gz[,G_] <- as.numeric(as.character(p_gz[,G_]))
  p_gz[,Z_] <- as.numeric(as.character(p_gz[,Z_]))
  return(p_gz[,c(G_,Z_,"q")])
}


### taken from gtools to avoid dependency
perms <- function (n, r, v = 1:n, set = TRUE, repeats.allowed = FALSE){
  if (mode(n) != "numeric" || length(n) != 1 || n < 1 || (n%%1) !=
      0)
    stop("bad value of n")
  if (mode(r) != "numeric" || length(r) != 1 || r < 1 || (r%%1) !=
      0)
    stop("bad value of r")
  if (!is.atomic(v) || length(v) < n)
    stop("v is either non-atomic or too short")
  if ((r > n) & repeats.allowed == FALSE)
    stop("r > n and repeats.allowed=FALSE")
  if (set) {
    v <- unique(sort(v))
    if (length(v) < n)
      stop("too few different elements")
  }
  v0 <- vector(mode(v), 0)
  if (repeats.allowed)
    sub <- function(n, r, v) {
      if (r == 1)
        matrix(v, n, 1)
      else if (n == 1)
        matrix(v, 1, r)
      else {
        inner <- Recall(n, r - 1, v)
        cbind(rep(v, rep(nrow(inner), n)), matrix(t(inner),
                                                  ncol = ncol(inner), nrow = nrow(inner) * n,
                                                  byrow = TRUE))
      }
    }
  else sub <- function(n, r, v) {
    if (r == 1)
      matrix(v, n, 1)
    else if (n == 1)
      matrix(v, 1, r)
    else {
      X <- NULL
      for (i in 1:n) X <- rbind(X, cbind(v[i], Recall(n -
                                                        1, r - 1, v[-i])))
      X
    }
  }
  sub(n, r, v[1:n])
}


# This function calculates multilocus genotype frequencies (p_gz) in which each individual locus is biallelic given allele frequencies and linkage disequilibrium among loci. Adapted from the original by Liam J. Revell 2018
## Osvaldo Espin-Garcia, 2021
# @param nloci number of locus involved or a vector with SNP names.
# @param p  (Major) allele frequencies of the dominant (in this case, merely uppercase) allele at each locus.
# @param LD Linkage disequilibrium info across pairs or loci. Flat LD matrix with as many rows as unique pairs of nloci. It must contain 3 columns: locus1, locus2 and r value (Pearson correlation).
# INTERNAL FUNCTION
multilocus.Pgz <- function(nloci=2,p=NULL,LD=NULL){
  if(is.null(p)) p <- rep(0.5, nloci)
  if(is.null(LD)) LD <- cbind(t(combn(1:nloci,2)), 0)
  if(length(p)!=nloci) nloci<-length(p)
  allelesfreqs <- cbind(p,1-p)
  COMBN <- perms(n=2,r=nloci,set=TRUE,repeats.allowed=TRUE)
  ALLELESp <- LETTERS[1:nloci]
  ALLELESq <- letters[1:nloci]
  ALLELESnams <- vector()
  FREQ <- rep(1,nrow(COMBN))
  for( i in 1:nrow(COMBN) ){
    for( j in 1:nloci ){
      FREQ[i] <- FREQ[i]*allelesfreqs[j,COMBN[i,j]]

      atype <- if(COMBN[i,j]==1) ALLELESp[j]
      else if(COMBN[i,j]==2) ALLELESq[j]

      ALLELESnams[i] <- if(j==1) atype else paste(ALLELESnams[i],atype,sep="")
    }
  }
  ### Add/Subtract LD
  FREQ2 <- FREQ
  pairs <- t(combn(1:nloci,2))
  if( nrow(pairs)!=nrow(LD) ){
    stop("The number of rows in LD matrix should correspond to the number of possible combinations: nrows(LD):",nrow(LD),"!=", nrow(pairs), " (num. possible combinations).")
  }
  for( k in 1:nrow(pairs) ){
    LD.pair <- LD[pairs[k,1]==LD[,1] & pairs[k,2]==LD[,2], 3]
    kthfreqs <- allelesfreqs[t(pairs[k,]),]
    LD.pair <- LD.pair * sqrt(prod(kthfreqs))

    Dpr <- LD.pair / ( (LD.pair<0)*max(-prod(kthfreqs[,1]),-prod(kthfreqs[,2])) + (LD.pair>=0)*min(prod(diag(kthfreqs)),prod(1-diag(kthfreqs))) )
    if( Dpr > 1 | Dpr < 0 ) message("One combination of MAFs and LD seems incompatible. i.e. D' > 1 or D'< 0. The frequencies will be adjusted/normalized in the results accordingly. Zero frequencies can be expected. SNP Pair: ", paste0(pairs[k,],collapse=", "))

    ind <- COMBN[,pairs[k,1]]==COMBN[,pairs[k,2]]
    FREQ2 <- FREQ2 + ind*LD.pair - (1-ind)*LD.pair
  }

  ### Generate Genotypes
  cross <- FREQ2 %*% t(FREQ2)
  colnames(cross) <- rownames(cross) <- ALLELESnams
  crossnames <- matrix(do.call(paste0, expand.grid(ALLELESnams,ALLELESnams)), ncol=length(ALLELESnams))
  crossnames_sorted <- apply(crossnames,c(1,2), function(x) sapply(lapply(strsplit(x, NULL), sort), paste, collapse=""))

  GENOTYPE <- vector()
  GENO <- perms(n=3,r=nloci,set=TRUE,repeats.allowed=TRUE)
  qval <- rep(0, nrow(GENO))
  for( i in 1:nrow(GENO) ){
    for( j in 1:nloci ){

      gtype <- if(GENO[i,j]==1) paste(rep(ALLELESp[j],2),collapse="")
      else if(GENO[i,j]==2) paste(c(ALLELESp[j],ALLELESq[j]),collapse="")
      else if(GENO[i,j]==3) paste(rep(ALLELESq[j],2),collapse="")

      GENOTYPE[i]<-if(j==1) gtype else paste(GENOTYPE[i],gtype,sep="")
    }
    ### match strings
    indMat <- which(crossnames_sorted %in% sapply(lapply(strsplit(GENOTYPE[i], NULL), sort), paste, collapse=""))
    # cat(indMat,"\n")
    qval[i] <- qval[i] + sum(cross[indMat])
  }

  ### Ensure values are between 0 and 1
  if( any(qval<0 | qval>1) ) message("At least one computed frequency was found to be <0 or >1. The frequencies will be adjusted/normalized accordingly. Zero frequencies can be expected.")
  qval <- pmin(pmax(qval,0),1)
  qval <- qval/sum(qval)

  res <- cbind(GENO-1, q=qval)
  res <- data.frame(res)
  cols <- if( nloci==2 ){
    c("Z","G")
  }else c("Z",paste0("G",1:(nloci-1)))
  colnames(res)[1:nloci] <- cols
  return(res)
}

#' This code a function that generates a sample Y, Z, G given some initial parameters.
#'
#'
#' @param Beta0  intercept  Default 2
#' @param Beta1  genetic effect   Default 0.5
#' @param Disp  dispersion parameter, e.g., variance of the error term when family is \code{gaussian()}, the reciprocal of the shape parameter when family is \code{Gamma()} Default: 1
#' @param N Phase 1 sample size, i.e. GWAS data (Default: 5000)
#' @param LD.r linkage disequilibrium (r) between G and Z  (Default: 0.75)
#' @param P_g  minor allele frequency for G, the causal SNP
#' @param P_z  minor allele frequency for Z, the GWAS SNP
#' @param family member of the exponential family (see \code{\link{family}};\sQuote{quasi} models not available). Default \code{gaussian()}.
#' @param tao quantile value to define the stratification of the response variable, Y (only applicable when family is \code{gaussian()}) (default: 2/5)
#' 
#' @return  A dataframe with complete data Y, G, Z, S, where G and Z come from the same haplotype determined by P_g, P_z and LD.r; Y is generated from E[Y|G] = family$linkinv(Beta0 + Beta1 x G). S is a 3 level variable determined by Y, Beta1, Disp and tao (only returned when family is gaussian). the function iterates across generated datasets until Z and Y are associated at a suggestive genome wide threshold of p<=1e-05.
#' 
#' 
#' @examples
#'
#' data = DataGeneration_TPD()
#'
#' @export
DataGeneration_TPD <- function(Beta0 = 2, Beta1 = 0.5, Disp = 1, N = 5000, LD.r = 0.75, P_g = 0.2, P_z = 0.3, family = gaussian(), tao = 2/5)  {
  
  message("Beta0=",Beta0, " Beta1=", Beta1, "\n");
  family_str = family$family
  # P_g<-0.2; P_z<-0.3 #MAF of G and Z respectively
  # tao = quantile to define the strata in Y
  
  P_G <- 1 - P_g; P_Z <- 1 - P_z ;
  
  Rsq <- LD.r^2
  LD <- LD.r * sqrt(P_G*P_g*P_Z*P_z)
  # Dpr <- LD /( (LD<0)*min(P_G*P_Z,P_g*P_z) + (LD>=0)*min(P_z*P_Z,P_G*P_g) )
  # Dpr <- LD /( (LD<0)*min(P_G*P_Z,P_g*P_z) + (LD>=0)*min(P_G*P_z,P_g*P_Z) ) ## corrected on Nov 2018 following Wikipedia and https://doi.org/10.1186/1471-2105-8-428
  Dpr <- LD /( (LD<0)*max(-P_G*P_Z,-P_g*P_z) + (LD>=0)*min(P_G*P_z,P_g*P_Z) ) ## corrected on Apr 2019 following Wikipedia and https://doi.org/10.1186/1471-2105-8-428
  
  #Values
  #cat(Rsq, LD, Dpr,"\n")
  
  ## Calculate haplotype frequencies (two SNPs)
  h.freqs <- rep(0, 4)
  h.freqs[1] <- LD + (1-P_g)*(1-P_z)
  h.freqs[2] <- 1- P_g - h.freqs[1]
  h.freqs[3] <- 1 - P_z - h.freqs[1]
  h.freqs[4] <- P_g - h.freqs[3]
  
  # This data generation can be generalized to an arbitrary number of columns as follows:
  # sampled_data <- df[sample(nrow(df), size = N, replace = TRUE, prob = df$prob_weight), ]
  # where df is data.frame with realizations of a joint distribution, coming from, say p_gz_func or multilocus.Pgz.

  pval<-1
  k<-0
  while( pval>=1e-05 ){
    h1 <- sample(1:4, size=N, prob=h.freqs, replace=TRUE)
    h2 <- sample(1:4, size=N, prob=h.freqs, replace=TRUE)
    G <- as.numeric(h1==3 | h1==4) + as.numeric(h2==3 | h2==4)
    Z <- as.numeric(h1==2 | h1==4) + as.numeric(h2==2 | h2==4)
    #Simulate the data with G the seq SNP and Z the tag SNP
    eta <- Beta0 + G*Beta1
    if (family_str == "binomial") {
      Y   <- rbinom(N, 1, family$linkinv(eta))
    } else if (family_str == "poisson"){  # poisson
      Y   <- rpois(N, family$linkinv(eta))
    } else if (family_str == "gaussian"){
      Y   <- rnorm(N, mean=family$linkinv(eta), sd = sqrt(Disp))
      Yst <- sapply(Y,function(x){
        if( x<qnorm(tao,mean=Beta0,sd=sqrt(Disp)) ){return(1)
        }else if( x>qnorm(1-tao,mean=Beta0,sd=sqrt(Disp)) ){return(3)
        }else return(2)})
    } else if (family_str == "Gamma"){
      Y   <- rgamma(N, scale = family$linkinv(eta)*Disp, shape = 1/Disp)
    }else stop("family argument is not one of binomial, poisson, gaussian or Gamma.")
    auxfit.comp <- glm(Y~Z, family = family)   
    pval <- pchisq(as.numeric(coef(auxfit.comp)[2]^2/diag(vcov(auxfit.comp))[2]),1, lower.tail = FALSE)
    k <- k+1
    if( pval<1e-05 ){       
      G0 <- sample(0:2, size=N, prob=c(P_G^2,2*P_G*P_g,P_g^2), replace=TRUE)
      dat_sim_it <-  if( family_str == "gaussian" ){
        data.frame(wait_it=k,Y=Y,G1=G,Z=Z,G0=G0,S=Yst)
        }else data.frame(wait_it=k,Y=Y,G1=G,Z=Z,G0=G0)
    }
  }
  return(dat_sim_it)
}

---
title: "Introduction to twoPhaseGAS"
author: "Osvaldo Espin-Garcia, Apostolos Dimitromanolakis, Shelley Bull"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to twoPhaseGAS}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width  = 7,
  fig.height = 4
)
```

## Overview

**twoPhaseGAS** provides tools for the design and analysis of two-phase
genetic association studies (GAS) in which a subset of the full cohort is
selected for targeted sequencing (phase 2), while the remaining individuals
contribute only GWAS-level data (phase 1).

The package offers:

1. **Data simulation** via `DataGeneration_TPD()`.
2. **Phase 2 design optimisation** via `twoPhaseDesign()` / `twoPhase()`.
3. **Joint analysis** of phase 1 and phase 2 data via `twoPhaseSPML()`, which
   implements a semiparametric maximum likelihood (SPML) estimator using the
   EM algorithm.

---

## 1. Simulating two-phase data

```{r simulate}
library(twoPhaseGAS)

data.table::setDTthreads(1)

set.seed(42)
data <- DataGeneration_TPD(
  Beta0  = 2,    # intercept
  Beta1  = 0.5,  # genetic effect (G on Y)
  Disp = 1,      # error variance for default gaussian family
  N      = 3000, # phase 1 cohort size
  LD.r   = 0.75, # LD (r) between causal SNP G and GWAS SNP Z
  P_g    = 0.20, # MAF of G
  P_z    = 0.30  # MAF of Z
)
head(data)
```

The simulated data frame contains:

| Column | Description |
|--------|-------------|
| `Y`    | Continuous outcome |
| `G1`   | Causal sequence variant (absent from phase 1 GWAS) |
| `Z`    | GWAS SNP (available for everyone) |
| `S`    | Stratum variable derived from residuals |

---

## 2. Selecting a phase 2 subsample

Here we use simple random sampling; `twoPhaseDesign()` can be used for
optimised designs.

```{r subsample}
set.seed(1)
n2   <- 500                              # phase 2 size
R    <- rep(0L, nrow(data))
R[sample(nrow(data), n2)] <- 1L         # 1 = selected for phase 2

data[R == 0, c("G1")]  <- NA # Make all phase 1 data complement G1 values  missing

cat("Phase 1 complement:", sum(R == 0), "individuals\n")
cat("Phase 2:           ", sum(R == 1), "individuals\n")
```

---

## 3. Analysis under H₀ (score test)

When the missing-by-design covariate (`miscov`) is **absent** from
`formula`, `twoPhaseSPML()` fits the null model and returns score statistics.

```{r null-model}
res_Ho <- twoPhaseSPML(
  formula = Y ~ Z,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

print(res_Ho)
```

The `print` method (modelled on `glm`) displays:

- the fitted call,
- estimated regression coefficients,
- family and link,
- log-likelihood and number of EM iterations, and
- the observed and expected score statistics with a chi-squared *p*-value.

---

## 4. Analysis under Hₐ (Wald test)

Including `miscov` in `formula` switches to the alternative model.  The EM
algorithm now estimates the effect of `G1` jointly with the regression
parameters, and Wald statistics are reported.

```{r alt-model}
res_Ha <- twoPhaseSPML(
  formula = Y ~ Z + G1,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

print(res_Ha)
```

---

## 5. Accessing results programmatically

The returned object is a list of class `"twoPhaseSPML"`.

```{r access}
# Regression coefficients
res_Ha$theta

# Log-likelihood
res_Ha$ll

# Estimated joint distribution of G1 and Z
head(res_Ha$qGZ)

# Wald statistic and degrees of freedom
cat("Wobs =", res_Ha$Wobs, "  df =", res_Ha$df, "\n")
cat("p-value =", pchisq(res_Ha$Wobs, df = res_Ha$df, lower.tail = FALSE), "\n")
```

---

## 6. Using `start.values` for warm starts

When analysing many variants, passing converged estimates from a nearby model
as `start.values` can reduce computation time.

```{r warm-start, eval = FALSE}
# Use null-model estimates as starting point for the alternative model
res_warm <- twoPhaseSPML(
  formula     = Y ~ Z + G1,
  miscov      = ~ G1,
  auxvar      = ~ Z,
  data       = data,
  start.values = list(betas = res_Ho$theta,
                      q     = res_Ho$qGZ)
)
```

---

## Session information

```{r session}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lrtest.twoPhaseSPML.R
\name{lrtest}
\alias{lrtest}
\alias{lrtest.twoPhaseSPML}
\title{Likelihood ratio test for \code{twoPhaseSPML} models}
\usage{
lrtest(object, object2, ...)

\method{lrtest}{twoPhaseSPML}(object, object2, ...)
}
\arguments{
\item{object}{a fitted object of class \code{"twoPhaseSPML"} representing
the \emph{smaller} (null / restricted) model.}

\item{object2}{a fitted object of class \code{"twoPhaseSPML"} representing
the \emph{larger} (alternative / unrestricted) model.  The two models
must have been fitted to the same data and with the same \code{miscov},
\code{auxvar}, and \code{family}.  The larger model must have at least
as many parameters as the smaller one.}

\item{\dots}{further arguments, currently unused.}
}
\value{
An object of class \code{"lrtest.twoPhaseSPML"}, which is a
  \code{data.frame} with one row per model and the following columns:
  \describe{
    \item{\code{#Df}}{number of estimated regression coefficients.}
    \item{\code{LogLik}}{maximised log-likelihood.}
    \item{\code{Df}}{change in degrees of freedom relative to the
      previous row (\code{NA} for the first row).}
    \item{\code{Chisq}}{LRT statistic (\code{NA} for the first row).}
    \item{\code{Pr(>Chisq)}}{two-sided \eqn{p}-value from the
      \eqn{\chi^2} distribution (\code{NA} for the first row).}
  }
  The object also carries a \code{"heading"} attribute (a character vector)
  and is printed via \code{\link{print.lrtest.twoPhaseSPML}}.
}
\description{
Compares two nested \code{\link{twoPhaseSPML}} models via a likelihood ratio
test (LRT), mirroring the interface of \code{lmtest::lrtest()}.
}
\details{
The test statistic is
\deqn{LRT = -2 \bigl(\ell_0 - \ell_1\bigr),}
where \eqn{\ell_0} and \eqn{\ell_1} are the maximised log-likelihoods of
the smaller and larger models, respectively.  Under the null hypothesis
that the additional parameters are zero, \eqn{LRT} follows a
\eqn{\chi^2} distribution with degrees of freedom equal to the difference
in the number of estimated regression coefficients between the two models.

The function determines model order automatically: whichever of
\code{object} and \code{object2} has fewer regression coefficients is
treated as the null model, regardless of the order in which they are
supplied.
}
\examples{
data.table::setDTthreads(1)
data <- DataGeneration_TPD(N=200)
set.seed(1)
R <- rep(0L, nrow(data))
R[sample(nrow(data), 175)] <- 1L
data[R == 0L, c("G1")] <- NA

# Null model (no G1)
res_Ho <- twoPhaseSPML(
  formula = Y ~ Z,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

# Alternative model (with G1)
res_Ha <- twoPhaseSPML(
  formula = Y ~ Z + G1,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)

lrtest(res_Ho, res_Ha)

}
\seealso{
\code{\link{twoPhaseSPML}}, \code{\link{print.lrtest.twoPhaseSPML}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataGeneration.R
\name{DataGeneration_TPD}
\alias{DataGeneration_TPD}
\title{This code a function that generates a sample Y, Z, G given some initial parameters.}
\usage{
DataGeneration_TPD(
  Beta0 = 2,
  Beta1 = 0.5,
  Disp = 1,
  N = 5000,
  LD.r = 0.75,
  P_g = 0.2,
  P_z = 0.3,
  family = gaussian(),
  tao = 2/5
)
}
\arguments{
\item{Beta0}{intercept  Default 2}

\item{Beta1}{genetic effect   Default 0.5}

\item{Disp}{dispersion parameter, e.g., variance of the error term when family is \code{gaussian()}, the reciprocal of the shape parameter when family is \code{Gamma()} Default: 1}

\item{N}{Phase 1 sample size, i.e. GWAS data (Default: 5000)}

\item{LD.r}{linkage disequilibrium (r) between G and Z  (Default: 0.75)}

\item{P_g}{minor allele frequency for G, the causal SNP}

\item{P_z}{minor allele frequency for Z, the GWAS SNP}

\item{family}{member of the exponential family (see \code{\link{family}};\sQuote{quasi} models not available). Default \code{gaussian()}.}

\item{tao}{quantile value to define the stratification of the response variable, Y (only applicable when family is \code{gaussian()}) (default: 2/5)}
}
\value{
A dataframe with complete data Y, G, Z, S, where G and Z come from the same haplotype determined by P_g, P_z and LD.r; Y is generated from E[Y|G] = family$linkinv(Beta0 + Beta1 x G). S is a 3 level variable determined by Y, Beta1, Disp and tao (only returned when family is gaussian). the function iterates across generated datasets until Z and Y are associated at a suggestive genome wide threshold of p<=1e-05.
}
\description{
This code a function that generates a sample Y, Z, G given some initial parameters.
}
\examples{

data = DataGeneration_TPD()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhaseSPMLglm_.R
\name{twoPhaseSPML}
\alias{twoPhaseSPML}
\title{Performs inference on two-phase studies data via semiparametric maximum likelihood.}
\usage{
twoPhaseSPML(
  formula,
  miscov,
  auxvar,
  family = gaussian,
  data,
  start.values = NULL,
  control = list()
)
}
\arguments{
\item{formula}{regression formula, note that if it does not contain the
missing-by-design variable, \code{miscov}, it will return results under
the null hypothesis. Hypothesis testing corresponds to the score
statistic. Otherwise, estimates and hypothesis testing occur under the
alternative hypothesis leading to Wald statistics. All the elements in
\code{formula} except \code{miscov} must be present in \code{data0} and
\code{data1}.}

\item{miscov}{right hand side formula with the missing-by-design
covariate(s), i.e. the potential causal locus (loci). Must be present in
\code{data1} but absent in \code{data0}.}

\item{auxvar}{right hand side formula with the auxiliary variable(s),
i.e. the GWAS SNP from phase 1. Must be present in \code{data0} and
\code{data1}.}

\item{family}{member of the exponential family (see \code{\link{family}};
\sQuote{quasi} models not available). Default \code{gaussian()}.}

\item{data}{a data frame containing data from both phases. Phase 2 data is assumed to be the result of the \code{complete.cases} with both \code{miscov} and \code{auxvar}. Missing values in variables other than \code{miscov} are not allowed and may lead to unexpected errors.}

\item{start.values}{a named list with initial values for the regression
parameters (\code{betas}) and/or the joint distribution between
\code{miscov} and \code{auxvar} (\code{q}). Defaults to \code{NULL}.}

\item{control}{a list of parameters for controlling the fitting process, i.e., number or EM iterations, tolerance and whether a trace is requested, it is passed to \code{\link{twoPhaseSPML.control}}.}
}
\value{
An object of class \code{"twoPhaseSPML"}, which is a list
  containing:
  \describe{
    \item{theta}{named numeric vector of estimated regression coefficients.}
    \item{qGZ}{data frame with the estimated joint distribution of
      \code{miscov} and \code{auxvar}.}
    \item{dispersion}{estimated dispersion parameter (variance for
      Gaussian; 1 for other families).}
    \item{iter}{integer, number of EM iterations performed.}
    \item{ll}{log-likelihood at convergence.}
    \item{df}{degrees of freedom for the test statistic.}
    \item{Sobs}{(null model only) observed score test statistic.}
    \item{Sexp}{(null model only) expected score test statistic.}
    \item{var_theta}{(alternative model only) named numeric vector of
      marginal variances for \code{theta}.}
    \item{Wobs}{(alternative model only) observed Wald test statistic.}
    \item{Wexp}{(alternative model only) expected Wald test statistic.}
    \item{qG}{data frame with the estimated marginal distribution of
      \code{miscov}.}
    \item{formula}{the \code{formula} used.}
    \item{miscov}{the \code{miscov} formula used.}
    \item{family}{the \code{family} object used.}
    \item{null.model}{logical, \code{TRUE} when the model was fitted under
      \eqn{H_0}.}
  }
}
\description{
Performs inference on two-phase studies data via semiparametric maximum likelihood.
}
\details{
The function implements the semiparametric maximum likelihood (SPML)
estimator for two-phase genetic association studies described by
Espin-Garcia et al. The EM algorithm alternates between an E-step that
computes observation weights based on the current estimate of the joint
distribution of the missing covariate and the auxiliary variable, and an
M-step that updates the regression coefficients and the non-parametric
distribution via iteratively re-weighted least squares and a closed-form
update, respectively.

When \code{miscov} is absent from \code{formula}, inference is performed
under \eqn{H_0} and score statistics are returned. When \code{miscov} is
included in \code{formula}, inference is performed under \eqn{H_a} and
Wald statistics are returned.
}
\examples{

data.table::setDTthreads(1)
data <- DataGeneration_TPD(N=200)
set.seed(1)
R <- rep(0, nrow(data))
R[sample(nrow(data), 175)] <- 1   # random phase 2 subsample of 500
data[R == 0, c("G1")] <- NA

# Null model (score test)
res_Ho <- twoPhaseSPML(
  formula = Y ~ Z,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)
print(res_Ho)

# Alternative model (Wald test)
res_Ha <- twoPhaseSPML(
  formula = Y ~ Z + G1,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)
print(res_Ha)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimJTC_.R
\name{optimTP.GA}
\alias{optimTP.GA}
\title{Genetic-algorithm optimal phase 2 design for a two-phase genetic association study.}
\usage{
optimTP.GA(
  formula,
  miscov,
  auxvar,
  family,
  n,
  data,
  beta,
  p_gz,
  disp = NULL,
  ga.popsize,
  ga.propelit,
  ga.proptourney,
  ga.ngen,
  ga.mutrate,
  ga.initpop = NULL,
  optimMeasure,
  K.idx = NULL,
  seed = 1,
  verbose = 0,
  cluster = NULL
)
}
\arguments{
\item{formula}{A two-sided \code{\link{formula}} specifying the regression
model of interest, e.g. \code{Y ~ G + Z}.  All variables except those in
\code{miscov} must be present in \code{data}.}

\item{miscov}{Right-hand-side formula identifying the missing-by-design
covariate(s) (the sequencing variant), e.g. \code{~ G}.}

\item{auxvar}{Right-hand-side formula identifying the auxiliary variable
(the GWAS tag SNP), e.g. \code{~ Z}.  Must be present in \code{data}.}

\item{family}{GLM family object (see \code{\link{family}}).  Default:
\code{gaussian()}.}

\item{n}{Integer.  Target phase 2 sample size.}

\item{data}{A \code{data.frame} containing the phase 1 data.  Must include
all variables in \code{formula} (except those in \code{miscov}) and
\code{auxvar}.}

\item{beta}{Numeric vector of assumed regression coefficients, one per
term in \code{formula} (including the intercept).  Setting
\code{beta[G] = 0} triggers the score-test (null-model) criterion; any
non-zero value triggers the Wald (alternative-model) criterion.}

\item{p_gz}{A \code{data.frame} with the assumed joint distribution of
(\code{miscov}, \code{auxvar}).  Must contain columns for each variable
in \code{miscov} and \code{auxvar} and a column \code{q} of
non-negative probability masses summing to 1.  Rows with \code{q <= 0}
are silently removed.}

\item{disp}{Numeric dispersion parameter (variance for Gaussian/Gamma;
fixed at 1 for binomial/Poisson).  If \code{NULL} (default),
estimated as \code{var(Y)}.}

\item{ga.popsize}{Integer.  GA population size (number of candidate
solutions maintained per generation).}

\item{ga.propelit}{Numeric in \eqn{(0, 1)}.  Proportion of the population
that is elite (carried forward unchanged to the next generation).
The elite count is \code{ceiling(ga.popsize * ga.propelit)}.}

\item{ga.proptourney}{Numeric in \eqn{(0, 1)}.  Proportion of the
population used in each tournament selection.  The tournament size is
\code{max(ceiling(ga.popsize * ga.proptourney), 2)}.}

\item{ga.ngen}{Integer.  Number of GA generations to run.}

\item{ga.mutrate}{Numeric in \eqn{[0, 1]}.  Per-index mutation probability.
At each mutation step, each of the \code{n} selected indices is
independently replaced by a randomly chosen non-selected index with
probability \code{ga.mutrate}.}

\item{ga.initpop}{Optional \code{ga.popsize}-by-\code{n} integer matrix of
starting solutions (each row is a candidate subset of size \code{n}).
If \code{NULL} (default), the initial population is generated randomly.
When \code{cluster} is provided, the rows are split evenly across
workers.}

\item{optimMeasure}{Character string specifying the optimality criterion.
One of \code{"Par-spec"} (parameter-specific), \code{"A-opt"}
(A-optimality), or \code{"D-opt"} (D-optimality).}

\item{K.idx}{Integer index (or vector of indices) identifying the
parameter(s) of primary interest.  Required when
\code{optimMeasure = "Par-spec"}.}

\item{seed}{Integer random seed, set before GA initialisation.  Default: 1.}

\item{verbose}{Integer.  If \code{> 0}, the best objective value is printed
every \code{verbose} generations.  Default: \code{0} (silent).}

\item{cluster}{An optional parallel cluster object created by
\code{\link[parallel]{makeCluster}}.  When provided,
\code{length(cluster)} independent GA searches are run in parallel
(each worker receives an equal share of \code{ga.initpop} rows, or a
random initial population if \code{ga.initpop = NULL}), and the best
solution across all workers is returned.  The caller is responsible for
starting and stopping the cluster.  Default: \code{NULL} (sequential
single-run GA).}
}
\value{
An object of class \code{"kofnGA"} (see
  \code{\link[kofnGA]{kofnGA}}) with the following elements:
  \describe{
    \item{bestsol}{Integer vector of length \code{n} giving the row indices
      (in \code{data}) of the selected phase 2 individuals.}
    \item{bestobj}{Numeric value of the optimality criterion at
      \code{bestsol}.}
    \item{pop}{Final population matrix (\code{ga.popsize} rows,
      \code{n} columns), sorted by fitness.}
    \item{obj}{Objective values for each row of \code{pop}.}
    \item{old}{List with per-generation \code{best}, \code{obj}, and
      \code{avg} records.}
    \item{time.elapsed}{Named numeric vector with the pre-processing
      (\code{preproc}) and GA (\code{GA}) times in seconds.}
  }
}
\description{
Uses a genetic algorithm (GA) to search over individual-level phase 2
selection indicators \eqn{R_i \in \{0,1\}} for the subset of size
\code{n} that maximises the anticipated Fisher information for the
regression parameters of interest.  Wraps \code{\link[kofnGA]{kofnGA}}.
When a parallel cluster is supplied, multiple independent GA runs are
executed in parallel and the best solution across all runs is returned.
}
\details{
The fitness function evaluates the anticipated Fisher information
matrix (FIM) for a candidate selection indicator \eqn{R}:
\deqn{\mathrm{FIM}(R) = I_3/N + \sum_{i: R_i=1}(I_{2,i} - I_{3,i})/N,}
where \eqn{I_3} is the empirical score outer-product estimator (the
numerically stable approximation to the full Louis (1982) observed FIM
valid for large \eqn{N}, since \eqn{I_1 \approx I_2} in the genotype
expansion context), and \eqn{I_{2,i}}, \eqn{I_{3,i}} are the
per-individual complete-data score outer product and observed-data
score outer product, respectively.  Under \eqn{H_0} (\code{beta[G] = 0})
the score-test Schur complement is optimised; under \eqn{H_a} the full
inverse FIM is optimised.
}
\examples{
\donttest{
data <- DataGeneration_TPD(N = 500)
data <- data[, c("Y", "Z")]  # phase-1 data: no G column
p_gz <- data.frame(Z = c(0, 0, 0, 1, 1, 1, 2, 2, 2), # theoretical joint distribution
        G1 = c(0, 1, 2, 0, 1, 2, 0, 1, 2), # between G and Z (LD = 0.75, MAF_G = 0.2, MAF_Z = 0.3)
        q = c(0.486, 0.004, 0, 0.143, 0.276, 0.001, 0.011, 0.04, 0.039)) 

res_ga <- optimTP.GA(
  formula      = Y ~ Z + G1,
  miscov       = ~ G1,
  auxvar       = ~ Z,
  family       = gaussian(),
  n            = 100,
  data         = data,
  beta         = c(2, 0.5, 0.1),
  p_gz         = p_gz,
  ga.popsize   = 40,
  ga.propelit  = 0.1,
  ga.proptourney = 0.1,
  ga.ngen      = 50,
  ga.mutrate   = 0.01,
  optimMeasure = "Par-spec",
  K.idx        = 2L
)
cat("Best objective:", res_ga$bestobj, "\n")
cat("Selected indices (first 10):", head(res_ga$bestsol, 10), "\n")

## Parallel processing (2 workers)
cl <- parallel::makeCluster(2L)
res_par <- optimTP.GA(
  formula      = Y ~ Z + G1,
  miscov       = ~ G1,
  auxvar       = ~ Z,
  family       = gaussian(),
  n            = 100,
  data         = data,
  beta         = c(2, 0.5, 0.1),
  p_gz         = p_gz,
  ga.popsize   = 40,
  ga.propelit  = 0.1,
  ga.proptourney = 0.1,
  ga.ngen      = 50,
  ga.mutrate   = 0.01,
  optimMeasure = "Par-spec",
  K.idx        = 2L,
  cluster      = cl
)
parallel::stopCluster(cl)
cat("Best objective (parallel):", res_par$bestobj, "\n")
}

}
\seealso{
\code{\link[kofnGA]{kofnGA}}, \code{\link{twoPhaseDesign}},
  \code{\link{optimTP.LM}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhase.R
\name{twoPhaseDesign}
\alias{twoPhaseDesign}
\title{Compute sample allocations for a two phase study design.}
\usage{
twoPhaseDesign(
  beta,
  maf_G,
  LD,
  data,
  n2,
  design_formula = Y ~ G + Z,
  family = gaussian(),
  S,
  perc_Y = c(1/5, 4/5),
  p_gz,
  design = c("RDS", "LM", "GA", "PPS", "BAL", "COM", "TZL"),
  ndraws = 10,
  optimCriterion = c("Par-spec", "A-opt", "D-opt"),
  overallMethod = c("med-max", "cumm")
)
}
\arguments{
\item{beta}{Vector of betas (length of 3, corresponding to intercept, effect size for G, effect size for Z).}

\item{maf_G}{minor allele frequency for G (numeric, ranging from 0 to 1). Numeric or vector of possible values.}

\item{LD}{Correlation r betweek G and Z (numeric, ranging from -1 to 1, r value). Numeric or vector of possible values.}

\item{data}{Data frame with Y and Z variables.}

\item{n2}{Phase 2 sample size.}

\item{design_formula}{Formula for the regression model, default is Y ~ G + Z, where Z is the GWAS SNP, G is the sequence variant. Y is outcome. Rename the variables in your data.frame to match Y and Z, G is the seq-SNP not present in the data.frame.}

\item{family}{Distrubution of the outcome. Default: gaussian(). Familes available for glm can be used here. See help(stats::family) for examples.}

\item{S}{stratification of the outcome Y. Optional. Should be a numeric vector with strata categories (e.g. 1 1 2 2 3 3 ). If present, its length must be equal to the number of rows in data. Default: NULL. Needed when Y does not render itself into strata, e.g. Gaussian, Poisson, Gamma.}

\item{perc_Y}{vector of percentiles in increasing order for which the outcome Y will be stratified. Default: c(1/5,4/5). Only used when S is NA. Note that setting up S is strongly suggested.}

\item{p_gz}{data frame with the joint distribution between G and Z. See examples for the right format. Default: NULL. If present, values of maf_G and LD are disregarded in the analysis.}

\item{design}{string for the design to use for phase 2 sample selection. One of residual-dependent sampling ("RDS"), optimal as defined by Tao, Zheng and Lin (2019) ("TZL"), optimal via Lagrange multipliers ("LM"), optimal via genetic algorithm ("GA"), probability proportional to size ("PPS"), balanced ("BAL") or combined ("COM") allocations. Default: "RDS". See details for a more explanations.}

\item{ndraws}{integer that determines the number of draws to examine when design is one of "pps", "bal" or "comb" and the design parameter combinations is greater than 1. Default: 10}

\item{optimCriterion}{string denoting the optimality criterion used during the optimization. One of "Par-spec" (default), "A-opt" or "D-opt". For parameter-specific, A-optimality or D-optimality, respectively.}

\item{overallMethod}{string denoting the method to select the overall design when multiple design parameters are given. One of "med-max" (default) or "cumm" for median-maximum and cummulative frequencies, respectively.

Note that in this version strata are always defined in terms of Z and S, i.e. a joint design, future implementations may relax this by allowing for only S or only Z (marginal designs, outcome- or covariate-dependent, respectively)}
}
\value{
sth
}
\description{
Compute sample allocations for a two phase study design.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhase.R
\name{twoPhaseHeuristic}
\alias{twoPhaseHeuristic}
\title{Select samples for phase 2 under heuristic designs.}
\usage{
twoPhaseHeuristic(
  design = c("pps", "bal", "comb"),
  ndraws = 1,
  data = NA,
  n2 = NA,
  family = gaussian(),
  S = NULL,
  perc_Y = c(1/5, 4/5)
)
}
\arguments{
\item{design}{Heuristic design to use for phase 2 sample selection. One of probability proportional to size ("pps"), balanced ("bal") or combined ("comb") allocations. Default: "pps".}

\item{ndraws}{Number of draws of the heuristic design to generate Default: 1.}

\item{data}{Data frame with Y and Z variables.}

\item{n2}{Phase 2 sample size.}

\item{family}{Distrubution of the outcome. Default: gaussian(). Familes available for glm can be used here. See help(stats::family) for examples.}

\item{S}{stratification of the outcome Y. Optional. Should be a numeric vector with strata categories (e.g. 1 1 2 2 3 3 ). If present, its length must be equal to the number of rows in data. Default: NULL. Needed when Y does not render itself into strata, e.g. Gaussian, Poisson, Gamma.}

\item{perc_Y}{vector of percentiles in increasing order for which the outcome Y will be stratified. Default: c(1/5,4/5). Only used when S is NA. Note that setting up S is strongly suggested.

Note that in this version strata are always defined in terms of Z and S, i.e. a joint design, future implementations may relax this by allowing for only S or only Z (marginal designs, outcome- or covariate-dependent, respectively)}
}
\value{
sth
}
\description{
Select samples for phase 2 under heuristic designs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhase.R
\name{twoPhase}
\alias{twoPhase}
\title{Internal function to optimize over a range of maf, LD and betas.}
\usage{
twoPhase(
  beta = c(1, 1, 1),
  maf_G = 0.1,
  LD = 0.3,
  data = NA,
  n2 = NA,
  design_formula = Y ~ G + Z,
  family = gaussian(),
  useGeneticAlgorithm = FALSE
)
}
\arguments{
\item{beta}{Vector of betas (length of 3, corresponding to intercept, effect size for G, effect size for Z).}

\item{maf_G}{Maf for G (numeric, ranging from 0 to 1).}

\item{LD}{Correlation r betweek G and Z (numeric, ranging from -1 to 1, r value).}

\item{data}{Data frame with Y and Z variables.}

\item{n2}{Phase 2 sample size}

\item{design_formula}{Formula for the regression model, default is Y ~ G + Z. Z is the GWAS SNP, G is the sequence variant. Y is outcome.}

\item{family}{Distrubution of the outcome (default: gaussian()  ). Familes available for glm can be used here. See help(stats::family) for examples.}

\item{useGeneticAlgorithm}{If TRUE, use genetic algorithm in addition to Lagrange multiplier approach (slower). Default: FALSE}
}
\value{
sth
}
\description{
Internal function to optimize over a range of maf, LD and betas.
}
\examples{

data = twoPhase()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/optimJTC_.R
\name{optimTP.LM}
\alias{optimTP.LM}
\title{Lagrange-multiplier optimal phase 2 design for a two-phase genetic association study.}
\usage{
optimTP.LM(
  formula,
  miscov,
  auxvar,
  strata,
  family,
  n,
  data,
  beta,
  p_gz,
  disp = NULL,
  optimMeasure,
  K.idx = NULL,
  min.nk = NULL,
  logical.sub = NULL
)
}
\arguments{
\item{formula}{A two-sided \code{\link{formula}} specifying the regression
model of interest, e.g.\code{Y ~ G + Z}.  All variables except
\code{miscov} must be present in \code{data}.}

\item{miscov}{Right-hand-side formula identifying the missing-by-design
covariate(s) (the sequencing variant), e.g.\code{~ G}.}

\item{auxvar}{Right-hand-side formula identifying the auxiliary variable
(the GWAS tag SNP), e.g.\code{~ Z}.  Must be present in \code{data}.}

\item{strata}{Right-hand-side formula defining the stratification variables
used to cross-tabulate \code{data}, e.g.\code{~ Z + S}.  All variables
must be present in \code{data}.}

\item{family}{GLM family object (see \code{\link{family}}).  Passed
directly to the information-matrix calculations.  Default:
\code{gaussian()}.}

\item{n}{Integer.  Target phase 2 sample size (number of individuals to
be sequenced).}

\item{data}{A \code{data.frame} containing the phase 1 data.  Must include
all variables in \code{formula} (except those in \code{miscov}),
\code{auxvar}, and \code{strata}.}

\item{beta}{Named or unnamed numeric vector of assumed regression
coefficients, with one element per term in \code{formula} (including
the intercept).  These define the operating point at which the
Fisher information is evaluated.  Setting \code{beta[G] = 0} triggers
the score-test (null-model) criterion; any non-zero value triggers the
Wald (alternative-model) criterion.}

\item{p_gz}{A \code{data.frame} with the assumed joint distribution of
(\code{miscov}, \code{auxvar}).  Must contain one column per variable
in \code{miscov} and \code{auxvar} plus a column \code{q} holding
non-negative probability masses that sum to 1.  Rows with
\code{q <= 0} are silently removed.  Typically generated by
\code{\link{twoPhaseDesign}} from \code{maf_G} and \code{LD}, or
supplied directly by the user.}

\item{disp}{Numeric dispersion parameter.  For Gaussian and Gamma
families this is the variance (\eqn{\sigma^2}) or dispersion
(\eqn{\phi}), respectively; for binomial and Poisson it is fixed at
1.  If \code{NULL} (default) it is estimated as \code{var(Y)}.}

\item{optimMeasure}{Character string specifying the optimality criterion.
One of \code{"Par-spec"} (parameter-specific: minimise the asymptotic
variance of \code{beta[K.idx]}), \code{"A-opt"} (A-optimality:
minimise \code{trace(FIM^{-1})}), or \code{"D-opt"} (D-optimality:
maximise \code{log det(FIM)}).}

\item{K.idx}{Integer index (or vector of indices) into \code{beta}
identifying the parameter(s) of primary interest for
\code{optimMeasure = "Par-spec"}.  Ignored for A- and D-optimality.
Required when \code{optimMeasure = "Par-spec"}.}

\item{min.nk}{Optional integer scalar or vector of length equal to the
number of strata.  Specifies the minimum number of individuals to
sample from each stratum.  If a scalar, the same minimum is applied to
all strata.  Default: \code{NULL} (no minimum, i.e. zero).}

\item{logical.sub}{Optional logical vector of length \code{nrow(data)}.
If supplied, only the rows where \code{logical.sub = TRUE} are used in
the information calculations (the full \code{data} is still returned).
Useful for sensitivity analyses on subsets.  Default: \code{NULL}.}
}
\value{
A \code{data.frame} with one row per stratum (as defined by
  \code{strata}) and the following columns from
  \code{xtabs(strata, data)}:
  \describe{
    \item{(strata variables)}{The stratification variable values.}
    \item{Freq}{Stratum size \eqn{N_k}.}
    \item{prR_cond_optim}{Optimal conditional sampling probability
      \eqn{n_k / n} (proportional allocation weight).}
    \item{convergence}{Convergence code from the solver (\code{0} = success;
      codes from \code{nloptr::cobyla} or \code{dfoptim::hjkb}).}
  }
}
\description{
Computes stratum-specific sampling fractions that maximise the anticipated
Fisher information for the regression parameters of interest subject to a
fixed total phase 2 sample size.  The optimisation is performed by the
COBYLA algorithm (\code{nloptr}) with a derivative-free fallback
(\code{dfoptim::hjkb}) if COBYLA fails to converge.
}
\details{
The function computes, for each stratum defined by \code{strata}, the
stratum-specific sampling fraction \eqn{\pi_k = n_k / N_k} that maximises
the chosen information criterion.  The optimisation is constrained so that
(i) the total selected sample equals \code{n}, (ii) no stratum is
over-sampled (\eqn{n_k \le N_k}), and (iii) the \code{min.nk} lower
bounds are respected.

The baseline Fisher information uses the empirical score outer-product
estimator \eqn{I_3/N} (see \code{\link{optimTP.GA}} for details).
}
\examples{
\donttest{
data <- DataGeneration_TPD(N = 1000)
data <- data[, c("Y", "Z")]  # phase-1 data: no G column
p_gz <- data.frame(Z = c(0, 0, 0, 1, 1, 1, 2, 2, 2), # theoretical joint distribution
        G1 = c(0, 1, 2, 0, 1, 2, 0, 1, 2), # between G and Z (LD = 0.75, MAF_G = 0.2, MAF_Z = 0.3)
        q = c(0.486, 0.004, 0, 0.143, 0.276, 0.001, 0.011, 0.04, 0.039)) 
        
data$S <- as.numeric(cut(data$Y,
            breaks = quantile(data$Y, probs = c(0, 0.4, 0.8, 1)),
            include.lowest = TRUE))

res_lm <- optimTP.LM(
  formula     = Y ~ Z + G1,
  miscov      = ~ G1,
  auxvar      = ~ Z,
  strata      = ~ Z + S,
  family      = gaussian(),
  n           = 200,
  data        = data,
  beta        = c(2, 0.5, 0.1),
  p_gz        = p_gz,
  optimMeasure = "Par-spec",
  K.idx       = 2L
)
head(res_lm)
}

}
\seealso{
\code{\link{twoPhaseDesign}}, \code{\link{optimTP.GA}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lrtest.twoPhaseSPML.R
\name{print.lrtest.twoPhaseSPML}
\alias{print.lrtest.twoPhaseSPML}
\title{Print method for \code{lrtest.twoPhaseSPML} objects}
\usage{
\method{print}{lrtest.twoPhaseSPML}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{an object of class \code{"lrtest.twoPhaseSPML"}, as returned by
\code{\link{lrtest.twoPhaseSPML}}.}

\item{digits}{integer; number of significant digits. Defaults to
\code{max(3L, getOption("digits") - 3L)}.}

\item{\dots}{further arguments passed to \code{\link{printCoefmat}}.}
}
\value{
\code{x} is returned invisibly.
}
\description{
Print method for \code{lrtest.twoPhaseSPML} objects
}
\seealso{
\code{\link{lrtest.twoPhaseSPML}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhaseSPMLglm_.R
\name{print.twoPhaseSPML}
\alias{print.twoPhaseSPML}
\title{Print method for \code{twoPhaseSPML} objects}
\usage{
\method{print}{twoPhaseSPML}(x, digits = max(3L, getOption("digits") - 3L), ...)
}
\arguments{
\item{x}{an object of class \code{"twoPhaseSPML"}, as returned by
\code{\link{twoPhaseSPML}}.}

\item{digits}{integer; minimum number of significant digits to print.
Passed to \code{\link{printCoefmat}}. Defaults to
\code{max(3L, getOption("digits") - 3L)}.}

\item{\dots}{further arguments passed to or from other methods (currently
unused).}
}
\value{
\code{x} is returned invisibly.
}
\description{
Prints a concise, \code{glm}-style summary of a fitted
\code{\link{twoPhaseSPML}} model.
}
\details{
Under the \emph{null} model (\code{miscov} absent from \code{formula}),
score statistics (\code{Sobs} and \code{Sexp}) are displayed. Under the
\emph{alternative} model, Wald statistics (\code{Wobs} and \code{Wexp})
together with standard errors and \eqn{p}-values for the missing-by-design
covariate(s) are displayed.  The coefficient table always shows the
estimated regression coefficients; standard errors and \eqn{p}-values are
shown only when \code{var_theta} is available (alternative model).
}
\examples{
data.table::setDTthreads(1)
data <- DataGeneration_TPD(N=200)
set.seed(1)
R <- rep(0, nrow(data))
R[sample(nrow(data), 175)] <- 1
data[R == 0, c("G1")] <- NA

res_Ho <- twoPhaseSPML(
  formula = Y ~ Z,
  miscov  = ~ G1,
  auxvar  = ~ Z,
  data   = data
)
print(res_Ho)

}
\seealso{
\code{\link{twoPhaseSPML}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/twoPhaseSPMLglm_.R
\name{twoPhaseSPML.control}
\alias{twoPhaseSPML.control}
\title{Auxiliary for Controlling the EM algorithm in twoPhaseSPML.}
\usage{
twoPhaseSPML.control(EMmaxit = 1000, EMtol = 1e-05, trace = FALSE)
}
\arguments{
\item{EMmaxit}{integer giving the maximal number of iterations of the EM algorithm.}

\item{EMtol}{numeric positive convergence tolerance for EM algorithm covnergence.}

\item{trace}{logical \code{TRUE}, prints progress at each EM iteration. Defaults to \code{FALSE}.}
}
\value{
A list with components named as the arguments.
}
\description{
Auxiliary for Controlling the EM algorithm in twoPhaseSPML.
}
\details{
The function implements the semiparametric maximum likelihood (SPML)
}
